Package: rerddap
Title: General Purpose Client for 'ERDDAP™' Servers
Description: General purpose R client for 'ERDDAP™' servers. Includes
    functions to search for 'datasets', get summary information on
    'datasets', and fetch 'datasets', in either 'csv' or 'netCDF' format.
    'ERDDAP™' information: 
    <https://upwell.pfeg.noaa.gov/erddap/information.html>.
Version: 1.3.0
Date: 2026-06-26
License: MIT + file LICENSE
Authors@R: c(
    person("Scott", "Chamberlain", role = "aut"),
    person("Ben", "Tupper", role = "ctb"),
    person("Salvador Jesús Fernández", "Bejarano", role = "ctb"),
    person("Roy", "Mendelssohn", role = c("cre", "ctb"),
    email = "roy.mendelssohn@noaa.gov")
    )
URL: https://docs.ropensci.org/rerddap/,
        https://github.com/ropensci/rerddap
BugReports: https://github.com/ropensci/rerddap/issues
LazyData: true
Depends: R (>= 4.00)
Encoding: UTF-8
Language: en-US
Imports: crul (>= 0.7.4), dplyr (>= 0.5.0), data.table (>= 1.12.0),
        digest, hoardr (>= 0.5.2), jsonlite (>= 1.6), lubridate,
        methods, nanoparquet, ncdf4 (>= 1.16), tibble, utils, xml2 (>=
        1.2.0)
Suggests: knitr, plotly, rmarkdown
VignetteBuilder: knitr
X-schema.org-applicationCategory: Climate
X-schema.org-keywords: earth, science, climate, precipitation,
        temperature, storm, buoy, NOAA
X-schema.org-isPartOf: https://ropensci.org
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0.9000
NeedsCompilation: no
Packaged: 2026-07-02 18:20:51 UTC; rmendels
Author: Scott Chamberlain [aut],
  Ben Tupper [ctb],
  Salvador Jesús Fernández Bejarano [ctb],
  Roy Mendelssohn [cre, ctb]
Maintainer: Roy Mendelssohn <roy.mendelssohn@noaa.gov>
Repository: CRAN
Date/Publication: 2026-07-03 03:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.info,character)
S3method(as.info,info)
S3method(browse,character)
S3method(browse,default)
S3method(browse,griddap_csv)
S3method(browse,griddap_nc)
S3method(browse,info)
S3method(browse,tabledap)
S3method(cache_delete,character)
S3method(cache_delete,griddap_csv)
S3method(cache_delete,griddap_nc)
S3method(cache_delete,list)
S3method(cache_delete,tabledap)
S3method(cache_details,character)
S3method(cache_details,default)
S3method(cache_details,griddap_csv)
S3method(cache_details,griddap_nc)
S3method(cache_details,list)
S3method(cache_details,tabledap)
S3method(print,ed_search)
S3method(print,ed_search_adv)
S3method(print,griddap_csv)
S3method(print,griddap_nc)
S3method(print,griddap_parquet)
S3method(print,info)
S3method(print,rerddap_cache)
S3method(print,rerddap_cache_info)
S3method(print,tabledap)
export(as.info)
export(browse)
export(cache_delete)
export(cache_delete_all)
export(cache_details)
export(cache_info)
export(cache_list)
export(cache_setup)
export(convert_time)
export(convert_units)
export(disk)
export(ed_datasets)
export(ed_search)
export(ed_search_adv)
export(eurl)
export(fipscounty)
export(global_search)
export(griddap)
export(info)
export(key_words)
export(memory)
export(servers)
export(tabledap)
export(version)
importFrom(crul,HttpClient)
importFrom(data.table,rbindlist)
importFrom(digest,digest)
importFrom(dplyr,as_data_frame)
importFrom(dplyr,bind_rows)
importFrom(hoardr,hoard)
importFrom(jsonlite,fromJSON)
importFrom(ncdf4,nc_close)
importFrom(ncdf4,nc_open)
importFrom(ncdf4,ncvar_get)
importFrom(utils,URLencode)
importFrom(utils,head)
importFrom(utils,read.csv)
importFrom(utils,read.delim)
importFrom(xml2,read_html)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_find_first)
importFrom(xml2,xml_text)

## ----initialize, echo = FALSE-------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE, 
  comment = "#>", 
  cache.path = "cache/",
  fig.path = "man/figures/"
)
library(rerddap)
data("colors")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# tabledap <- function(x, ..., fields=NULL, distinct=FALSE, orderby=NULL,
#   orderbymax=NULL, orderbymin=NULL, orderbyminmax=NULL, units=NULL, fmt = 'csv',
#   url = eurl(), store = disk(), callopts=list())

## ----eval = FALSE-------------------------------------------------------------
# install.packages("rerddap")

## ----eval = FALSE-------------------------------------------------------------
# remotes::install_github("ropensci/rerddap")

## -----------------------------------------------------------------------------
library("rerddap")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# library("akima")
# library("dplyr")
# library("ggplot2")
# library("mapdata")
# library("ncdf4")
# library("plot3D")

## ----eval = FALSE-------------------------------------------------------------
# ?rerddap

## ----eval = FALSE-------------------------------------------------------------
# whichSST <- ed_search(query = "SST")

## ----eval = FALSE-------------------------------------------------------------
# whichSST <- ed_search(query = "SST MODIS")

## -----------------------------------------------------------------------------
servers()

## -----------------------------------------------------------------------------
100*100*4*8*365

## ----eval = FALSE, echo = TRUE------------------------------------------------
# browse('jplMURSST41')
# browse('siocalcofiHydroCast')

## ----eval = FALSE, echo = TRUE------------------------------------------------
# info('jplMURSST41')
# info('siocalcofiHydroCast')

## ----eval = FALSE, echo = TRUE------------------------------------------------
# latitude = c(22., 51.)
# longitude = c(-140., -105)
# time = c("2017-01-01", "2017-01-02")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# sstInfo <- info('jplMURSST41')
# murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c("2017-01-01", "2017-01-02"), fields = 'analysed_sst')
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c("2017-01-01", "2017-01-02"), stride = c(1,1,5), fields = 'analysed_sst')
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# 'time>=2010-01-01'
# 'time<=2010-12-31'
# 'scientific_name="Sardinops sagax"'
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# CPSinfo <- info('FRDCPSTrawlLHHaulCatch')
# sardines <- tabledap(CPSinfo, fields = c('latitude',  'longitude', 'time', 'scientific_name', 'subsample_count'), 'time>=2010-01-01', 'time<=2010-12-31', 'scientific_name="Sardinops sagax"' )
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# CPSinfo <- info('FRDCPSTrawlLHHaulCatch')
# oncorhynchus <- tabledap(CPSinfo,
#                          fields = c('latitude',  'longitude', 'time', 'scientific_name',
#                          'subsample_count'),
#                          'scientific_name=~"Oncorhynchus.*"'
#                          )
# head(oncorhynchus)
# 
# <ERDDAP™ tabledap> FRDCPSTrawlLHHaulCatch
#    Path: [/var/folders/46/jyz1mm5x5bvbf59b5g8f7f580000gn/T//RtmpLpXHMI/R/rerddap/0ce4939b104016aafebee121efeeba00.csv]
#    Last updated: [2024-10-29 11:14:10.975094]
#    File size:    [0.02 mb]
# # A tibble: 6 × 5
#   latitude longitude time                scientific_name      subsample_count
#      <dbl>     <dbl> <dttm>              <chr>                          <int>
# 1     43.0     -125. 2003-07-09 06:08:00 Oncorhynchus kisutch               1
# 2     43.0     -125. 2003-07-09 07:08:00 Oncorhynchus kisutch               1
# 3     43       -125. 2003-07-09 08:07:00 Oncorhynchus                       1
# 4     46.0     -124. 2003-07-12 08:52:00 Oncorhynchus                      10
# 5     46.0     -124. 2003-07-12 10:24:00 Oncorhynchus                      10
# 6     46.0     -124. 2003-07-12 12:09:00 Oncorhynchus                       2
# 
# 

## ----MUR, eval = FALSE, echo = TRUE-------------------------------------------
# require("ggplot2")
# require("mapdata")
# require("rerddap")
# sstInfo <- info('jplMURSST41')
# # get latest daily sst
# murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c('last','last'), fields = 'analysed_sst')
# mycolor <- colors$temperature
# w <- map_data("worldHires", ylim = c(22., 51.), xlim = c(-140, -105))
# ggplot(data = murSST$data, aes(x = longitude, y = latitude, fill = analysed_sst)) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     geom_raster(interpolate = FALSE) +
#     scale_fill_gradientn(colours = mycolor, na.value = NA) +
#     theme_bw() + ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = c(-140, -105),  ylim = c(22., 51.)) + ggtitle("Latest MUR SST")

## ----VIIRS, echo = TRUE, eval = FALSE-----------------------------------------
# require("ggplot2")
# require("mapdata")
# require("rerddap")
# sstInfo <- info('erdVHsstaWS3day', url = "https://coastwatch.pfeg.noaa.gov/erddap/")
# # get latest 3-day composite sst
# viirsSST <- griddap(sstInfo, latitude = c(41., 31.), longitude = c(-128., -115), time = c('last','last'), fields = 'sst')
# # remap latitiudes and longitudes to even grid
# myLats <- unique(viirsSST$data$latitude)
# myLons <- unique(viirsSST$data$longitude)
# myLats <- seq(range(myLats)[1], range(myLats)[2], length.out = length(myLats))
# myLons <- seq(range(myLons)[1], range(myLons)[2], length.out = length(myLons))
# # melt these out to full grid
# mapFrame <- expand.grid(x = myLons, y = myLats)
# mapFrame$y <- rev(mapFrame$y)
# # form a frame with the new values and the data
# tempFrame <- data.frame(sst = viirsSST$data$sst, lat = mapFrame$y, lon = mapFrame$x)
# mycolor <- colors$temperature
# w <- map_data("worldHires", ylim = c(30., 42.), xlim = c(-128, -114))
# ggplot(data = tempFrame, aes(x = lon, y = lat, fill = sst)) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     geom_raster(interpolate = FALSE) +
#     scale_fill_gradientn(colours = mycolor, na.value = NA) +
#     theme_bw() + ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = c(-128, -114),  ylim = c(30., 42.)) + ggtitle("Latest VIIRS 3-day SST")
# 

## ----echo = TRUE, eval = FALSE------------------------------------------------
# require("ggplot2")
# require("rerddap")
# viirsSST1 <- griddap(sstInfo, latitude = c(36., 36.),
#                      longitude = c(-126., -126.),
#                      time = c('2015-01-01','2015-12-31'), fields = 'sst')
# tempTime <- as.Date(viirsSST1$data$time, origin = '1970-01-01', tz = "GMT")
# tempFrame <- data.frame(time = tempTime, sst = viirsSST1$data$sst)

## ----echo = TRUE, eval = FALSE------------------------------------------------
# ggplot(tempFrame, aes(time, sst)) +
#   geom_line() +
#   theme_bw() +
#   ylab("sst") +
#   ggtitle("VIIRS SST at (36N, 126W)")

## ----VHNCHla, eval = FALSE, echo = TRUE---------------------------------------
# require("ggplot2")
# require("mapdata")
# require("rerddap")
# chlaInfo <- info('erdVHNchla3day')
# viirsCHLA <- griddap(chlaInfo, latitude = c(41., 31.),
#                      longitude = c(-128., -115), time = c('last','last'),
#                      fields = 'chla')

## ----eval = FALSE, echo = TRUE------------------------------------------------
# mycolor <- colors$chlorophyll
# w <- map_data("worldHires", ylim = c(30., 42.), xlim = c(-128, -114))
# ggplot(data = viirsCHLA$data, aes(x = longitude, y = latitude, fill = log(chla))) +
#   geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   geom_raster(interpolate = FALSE) +
#   scale_fill_gradientn(colours = mycolor, na.value = NA) +
#   theme_bw() + ylab("latitude") + xlab("longitude") +
#   coord_fixed(1.3, xlim = c(-128, -114),  ylim = c(30., 42.)) +
#   ggtitle("Latest VIIRS 3-day Chla")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# require("rerddap")
# dataInfo <- rerddap::info('hawaii_d90f_20ee_c4cb')
# xpos <- c(135.25, 240.25)
# ypos <- c(20.25, 60.25)
# zpos <- c(70.02, 70.02)
# tpos <- c('2010-12-15', '2010-12-15')
# soda70 <- griddap(dataInfo,  longitude = xpos, latitude = ypos,
#                   time = tpos, depth = zpos, fields = 'temp' )
# 

## ----soda70Plot, eval = FALSE, echo = TRUE------------------------------------
# require("ggplot2")
# require("mapdata")
# xlim <- c(135, 240)
# ylim <- c(20, 60)
# my.col <- colors$temperature
# ## Must do a kludge to remove countries that wrap and mess up the plot
# w1 <- map("world2Hires", xlim = c(135, 240), ylim = c(20, 60), fill = TRUE, plot = FALSE)
# remove <- c("UK:Great Britain", "France", "Spain", "Algeria", "Mali", "Burkina Faso", "Ghana", "Togo")
# w <- map_data("world2Hires", regions = w1$names[!(w1$names %in% remove)], ylim = ylim, xlim = xlim)
# myplot <- ggplot() +
#     geom_raster(data = soda70$data, aes(x = longitude, y = latitude, fill = temp), interpolate = FALSE) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(-3,30), name = "temperature") +
#     ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#     ggtitle(paste("70m temperature ", soda70$data$time[1]))
# myplot

## ----NAtlSSS, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# urlBase <- "https://erddap.marine.ie/erddap/"
# parameter <- "sea_surface_salinity"
# sssTimes <- c("last", "last")
# sssLats <- c(48.00625, 57.50625)
# sssLons <- c(-17.99375, -1.00625)
# dataInfo <- rerddap::info("IMI_NEATL", url = urlBase)
# NAtlSSS <- griddap(dataInfo, longitude = sssLons, latitude = sssLats, time = sssTimes, fields = parameter, url = urlBase)

## ----NAtlSSSplot, eval = FALSE, echo = TRUE-----------------------------------
# require("ggplot2")
# require("mapdata")
# xlim <- c(-17.99375, -1.00625)
# ylim <- c(48.00625, 57.50625)
# my.col <- colors$salinity
# w <- map_data("worldHires", ylim = ylim, xlim = xlim)
# myplot <- ggplot() +
#     geom_raster(data = NAtlSSS$data, aes(x = longitude, y = latitude, fill = sea_surface_salinity), interpolate = FALSE) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(34, 36), name = "salinity") +
#     ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#     ggtitle(paste("salinity", NAtlSSS$data$time[1]))
# myplot

## ----IFREMER, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# urlBase <- "https://www.ifremer.fr/erddap/"
# parameter <- "Salinity"
# ifrTimes <- c("2010-05-16", "2010-05-16")
# ifrLats <- c(30., 50.)
# ifrLons <- c(-140., -110.)
# ifrDepth <- c(75., 75.)
# dataInfo <- rerddap::info("SDC_GLO_CLIM_TS_V2_2", url = urlBase)
# ifrPSAL <- griddap(dataInfo, longitude = ifrLons, latitude = ifrLats, time = ifrTimes, depth = ifrDepth,  fields = parameter, url = urlBase)

## ----ifrPSALplot, eval = FALSE, echo = TRUE-----------------------------------
# ## ggplot2 has trouble with unequal y's
#  require("akima")
#  require("dplyr")
#  require("ggplot2")
#  require("mapdata")
#   xlim <- c(-140, -110)
#   ylim <- c(30, 51)
#   my.col <- colors$salinity
#   myplot <- ggplot() +
#     geom_raster(data = ifrPSAL$data, aes(x = longitude, y = latitude, fill = Salinity), interpolate = FALSE) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(32, 35), name = "salinity") +
#     ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) + ggtitle(paste("salinity at 75 meters",ifrPSAL$data$time[1] ))
#  myplot

## ----eval = FALSE, echo = TRUE------------------------------------------------
# require("rerddap")
# hydroInfo <- info('siocalcofiHydroCast')

## ----calCOFI, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# calcofi.df <- tabledap(hydroInfo, fields = c('cst_cnt',  'date', 'year', 'month', 'julian_date', 'julian_day', 'rpt_line', 'rpt_sta', 'cruz_num', 'intchl', 'intc14', 'time'), 'time>=1984-01-01T00:00:00Z', 'time<=2014-04-17T05:35:00Z')

## ----calCOFInum, eval = FALSE, echo = TRUE------------------------------------
# if (packageVersion('rerddap') < '1.1.0'){
#     calcofi.df$cruz_num <- as.numeric(calcofi.df$cruz_num)
#     calcofi.df$intc14 <- as.numeric(calcofi.df$intc14)
#     calcofi.df$time <- as.Date(calcofi.df$time, origin = '1970-01-01', tz = "GMT")
# }

## ----calCOFIPlot, eval = FALSE,  echo = TRUE----------------------------------
# require("dplyr")
# 
# # calculate cruise means
# by_cruznum <- group_by(calcofi.df, cruz_num)
# tempData <- select(by_cruznum, year, month, cruz_num, intchl, intc14)
# CruiseMeans <- summarize(by_cruznum, cruisechl = mean(intchl, na.rm = TRUE), cruisepp = mean(intc14, na.rm = TRUE), year = median(year, na.rm = TRUE), month = median(month, na.rm = TRUE))
# tempTimes <- paste0(CruiseMeans$year,'-',CruiseMeans$month,'-1')
# cruisetimes <- as.Date(tempTimes, origin = '1970-01-01', tz = "GMT")
# CruiseMeans$cruisetimes <- cruisetimes
# # calculate monthly "climatologies"
# byMonth <- group_by(CruiseMeans, month)
# climate <- summarize(byMonth, ppClimate = mean(cruisepp, na.rm = TRUE), chlaClimate = mean(cruisechl, na.rm = TRUE))
# # calculate anomalies
# CruiseMeans$chlanom <- CruiseMeans$cruisechl - climate$chlaClimate[CruiseMeans$month]
# CruiseMeans$ppanom <- CruiseMeans$cruisepp - climate$ppClimate[CruiseMeans$month]
# # calculate mean yearly anomaly
# byYear <- select(CruiseMeans, year)
# tempData <- select(CruiseMeans, year, chlanom, ppanom )
# byYear <- group_by(tempData, year)
# yearlyAnom <- summarize(byYear, ppYrAnom = mean(ppanom, na.rm = TRUE), chlYrAnom = mean(chlanom, na.rm = TRUE))
# yearlyAnom$year <- ISOdate(yearlyAnom$year, 01, 01, hour = 0)
# ggplot(yearlyAnom, aes(year, chlYrAnom)) + geom_line() +
#   theme_bw() + ggtitle('yearly chla anom')
# ggplot(yearlyAnom, aes(year, ppYrAnom)) + geom_line() +
#   theme_bw() + ggtitle('yearly pp anom')

## ----calCOFI-parquet, eval = FALSE, echo = TRUE-------------------------------
# require("rerddap")
# calcofi.df <- tabledap(hydroInfo, fields = c('cst_cnt',  'date', 'year', 'month', 'julian_date', 'julian_day', 'rpt_line', 'rpt_sta', 'cruz_num', 'intchl', 'intc14', 'time'), 'time>=1984-01-01T00:00:00Z', 'time<=2014-04-17T05:35:00Z', fmt = 'parquet')

## ----CPS Query, eval = FALSE, echo = TRUE-------------------------------------
# require("rerddap")
# CPSquery <- ed_search(query = 'CPS Trawl')
# CPSquery$alldata[[1]]$summary
# CPSquery$alldata[[1]]$tabledap
# CPSquery$alldata[[1]]$dataset_id

## ----CPSINfo, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# CPSinfo <- info('FRDCPSTrawlLHHaulCatch')

## ----eval = FALSE, echo = TRUE------------------------------------------------
# require("dplyr")
# require("rerddap")
# sardines <- tabledap(CPSinfo, fields = c('latitude',  'longitude', 'time', 'scientific_name', 'subsample_count'), 'time>=2010-01-01', 'time<=2012-01-01', 'scientific_name="Sardinops sagax"' )
# if (packageVersion('rerddap') < '1.1.0'){
#     sardines$time <- as.Date(sardines$time, origin = '1970-01-01', tz = "GMT")
#     sardines$latitude <- as.numeric(sardines$latitude)
#     sardines$longitude <- as.numeric(sardines$longitude)
# }
# sardines$time <- as.Date(sardines$time, origin = '1970-01-01', tz = "GMT")
# sardines$latitude <- as.numeric(sardines$latitude)
# sardines$longitude <- as.numeric(sardines$longitude)
# sardine2010 <- filter(sardines, time < as.Date('2010-12-01'))

## ----eval = FALSE, echo = TRUE------------------------------------------------
# # get the dataset info
# sstInfo <- info('erdMWsstdmday')
# # get 201004 monthly sst
# sst201004 <- griddap('erdMWsstdmday', latitude = c(22., 51.), longitude = c(220., 255), time = c('2010-04-16','2010-04-16'), fields = 'sst')
# # get 201104 monthly sst
# sst201104 <- griddap('erdMWsstdmday', latitude = c(22., 51.), longitude = c(220., 255), time = c('2011-04-16','2011-04-16'), fields = 'sst')

## ----CPSPlot, eval = FALSE, echo = TRUE---------------------------------------
# # get polygons of coast for this area
# w <- map_data("worldHires", ylim = c(22., 51.), xlim = c(220 - 360, 250 - 360))
# # plot 201004 sst on the map
# sardine2010 <- filter(sardines, time < as.Date('2010-12-01', origin = '1970-01-01', tz = "GMT"))
# sardine2011 <- filter(sardines, time > as.Date('2010-12-01', origin = '1970-01-01', tz = "GMT"))
# mycolor <- colors$temperature
# p1 <- ggplot() +
#   geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   geom_raster(data = sst201004$data, aes(x = (longitude - 360), y = latitude, fill = sst), interpolate = FALSE) +
#   scale_fill_gradientn(colours = mycolor, na.value = NA, limits = c(5,30)) +
#   theme_bw() + ylab("latitude") + xlab("longitude") +
#   coord_fixed(1.3, xlim = c(220 - 360, 250 - 360),  ylim = c(22., 51.))
# 
# # plot 201104 sst on the map
# p2 <- ggplot() +
#   geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   geom_raster(data = sst201104$data, aes(x = (longitude - 360), y = latitude, fill = sst), interpolate = FALSE) +
#   geom_point(data = sardine2011, aes(x = longitude, y = latitude, colour = subsample_count)) +
#   scale_fill_gradientn(colours = mycolor, na.value = NA, limits = c(5,30)) +
#   theme_bw() + ylab("latitude") + xlab("longitude") +
#   coord_fixed(1.3, xlim = c(220 - 360, 250 - 360),  ylim = c(22., 51.))
# p1 + geom_point(data = sardine2010, aes(x = longitude, y = latitude, colour = subsample_count)) + scale_colour_gradient(space = "Lab", na.value = NA, limits = c(0,80))
# 
# p2 +   geom_point(data = sardine2011, aes(x = longitude, y = latitude, colour = subsample_count)) + scale_colour_gradient(space = "Lab", na.value = NA, limits = c(0,80))

## ----eval = FALSE,  echo = TRUE-----------------------------------------------
# sardinops <- tabledap(CPSinfo, fields = c('longitude', 'latitude', 'time'),  'scientific_name="Sardinops sagax"')
# if (packageVersion('rerddap') < '1.1.0') {
#     sardinops$time <- as.Date(sardinops$time, origin = '1970-01-01', tz = "GMT")
#     sardinops$year <- as.factor(format(sardinops$time, '%Y'))
#     sardinops$latitude <- as.numeric(sardinops$latitude)
#     sardinops$longitude <- as.numeric(sardinops$longitude)
# }
# sardinops$year <- as.factor(format(sardinops$time, '%Y'))
# 

## ----sardinesPlot, eval = FALSE, echo = TRUE----------------------------------
# xlim <- c(-135, -110)
# ylim <- c(30, 51)
# coast <- map_data("worldHires", ylim = ylim, xlim = xlim)
# ggplot() +
#     geom_point(data = sardinops, aes(x = longitude, y = latitude, colour = year)) +
#     geom_polygon(data = coast, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#     ggtitle("Location of sardines by year in EPM Trawls")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# # get location and station ID of NDBC buoys in a region
# BuoysInfo <- info('cwwcNDBCMet')
# locationBuoys <- tabledap(BuoysInfo, distinct = TRUE, fields = c("station", "longitude", "latitude"), "longitude>=-124", "longitude<=-121", "latitude>=37", "latitude<=47")
# if (packageVersion('rerddap') < '1.1.0') {
#    locationBuoys$latitude <- as.numeric(locationBuoys$latitude)
#    locationBuoys$longitude <- as.numeric(locationBuoys$longitude)
# }

## ----NDBC, eval = FALSE,  echo = TRUE-----------------------------------------
# xlim <- c(-130, -110)
# ylim <- c(35, 50)
# coast <- map_data("worldHires", ylim = ylim, xlim = xlim)
# ggplot() +
#    geom_point(data = locationBuoys, aes(x = longitude , y = latitude, colour = factor(station) )) +
#    geom_polygon(data = coast, aes(x = long, y = lat, group = group), fill = "grey80") +
#    theme_bw() + ylab("latitude") + xlab("longitude") +
#    coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#    ggtitle("Location of buoys in given region")
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# buoyData <- tabledap(BuoysInfo, fields = c("time", "wspd"), 'station="46012"', 'time>=2012-01-01', 'time<=2013-01-01')
# if (packageVersion('rerddap') < '1.1.0') {
#     buoyData$wspd <- as.numeric(buoyData$wspd)
#     buoyData$time <- as.Date(buoyData$time, origin = '1970-01-01', tz = "GMT")
# }

## ----NDBCTS, eval = FALSE, echo = TRUE----------------------------------------
# ggplot(buoyData, aes(time, wspd)) +
#   geom_line() +
#   theme_bw() +
#   ylab("wind speed") +
#   ggtitle("Wind Speed in 2012 from buoy 46012 ")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# urlBase <- "https://data.ioos.us/gliders/erddap/"
# gliderInfo <- info("sp064-20161214T1913",  url = urlBase)
# glider <- tabledap(gliderInfo, fields = c("longitude", "latitude", "depth", "salinity"), 'time>=2016-12-14', 'time<=2016-12-23', url = urlBase)
# if (packageVersion('rerddap') < '1.1.0') {
#     glider$longitude <- as.numeric(glider$longitude)
#     glider$latitude <- as.numeric(glider$latitude)
#     glider$depth <- as.numeric(glider$depth)
# }

## ----glider, eval = FALSE, echo = TRUE----------------------------------------
# require("plot3D")
# scatter3D(x = glider$longitude , y = glider$latitude , z = -glider$depth, colvar = glider$salinity,              col = colors$salinity, phi = 40, theta = 25, bty = "g", type = "p",
#            ticktype = "detailed", pch = 10, clim = c(33.2,34.31), clab = 'Salinity',
#            xlab = "longitude", ylab = "latitude", zlab = "depth",
#            cex = c(0.5, 1, 1.5))

## ----echo = TRUE, eval = FALSE------------------------------------------------
# atnURL <- 'https://oceanview.pfeg.noaa.gov/erddap/'
# atnInfo <- info('gtoppAT', url = atnURL)
# atnData <- tabledap(atnInfo, fields = c("time", "longitude", "latitude"), 'toppID="1807001"', url = atnURL)
# atnData$latitude <- as.numeric(atnData$latitude)
# atnData$longitude <- as.numeric(atnData$longitude)
# ncdcSST = array(NA_real_, dim = length(atnData$time))
# ncdcSSTInfo = info('ncdcOisst2Agg')
# time_bound <- c(as.character(atnData$time[1]), as.character(atnData$time[66]))
# for (i in 1:length(atnData$time)) {
#   extract <- griddap(ncdcSSTInfo,
#                      fields = 'sst',
#                      latitude = c(atnData$latitude[i], atnData$latitude[i]),
#                      longitude = c(atnData$longitude[i], atnData$longitude[i]),
#                      time = time_bound
#                      )
# ncdcSST[i] <- extract$data$sst
# }

## ----eval = FALSE,  echo = TRUE-----------------------------------------------
# ylim <- c(32.5, 34)
# xlim <- c(-119, -116.5)
# mycolor <- colors$temperature
# w <- map_data("worldHires", ylim = ylim, xlim = xlim)
# alldata <- data.frame(sst = ncdcSST, longitude = atnData$longitude - 360, latitude = atnData$latitude)
# z <- ggplot(alldata, aes(x = longitude, y = latitude)) +
#    geom_point(aes(colour = sst), size = .5)
# z + geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   theme_bw() +
#   scale_colour_gradientn(colours = mycolor, limits = c(16.9, 17.3), "SST") +
#   coord_fixed(1.3, xlim = xlim, ylim = ylim) + ggtitle("SST Along Track")
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# urlBase <- 'https://coastwatch.pfeg.noaa.gov/erddap/'
# nphInfo <- info('erdNph', url = urlBase)
# nphData <- tabledap(nphInfo, fields = c("year", "maxSLP" ), 'month=2', 'year>=1987', url = urlBase)
# nphData$maxSLP <- as.numeric(nphData$maxSLP)
# urlBase <- 'https://oceanview.pfeg.noaa.gov/erddap/'
# cohoInfo <- info('cciea_SM_CA_CO_ABND', url = urlBase)
# cohoData <- tabledap(cohoInfo, fields = c("abundance_anomaly", "time"),  url = urlBase)
# if (packageVersion('rerddap') < '1.1.0') {
#     cohoData$abundance_anomaly <- as.numeric(cohoData$abundance_anomaly)
# }
# alldata <- data.frame(coho = cohoData$abundance_anomaly[1:27], maxSLP = nphData$maxSLP, year = nphData$year)

## ----eval = FALSE, echo = TRUE------------------------------------------------
# ggplot(alldata) + geom_line(aes(x = year, y = coho), colour = 'blue') + theme_bw() + ggtitle("coho abundance anomaly")
# ggplot(alldata) + geom_line(aes(x = year, y = maxSLP), colour = 'red') + theme_bw() + ggtitle("MaxSLP")

## ----echo = TRUE, eval = FALSE------------------------------------------------
# 
# bbox_set <- function(plotobj, landmask = TRUE, xlim = NULL, ylim = NULL, interactive = FALSE)
# 

## ----echo = TRUE, eval = FALSE------------------------------------------------
# library(plotly)
# 
# myURL <- "https://coastwatch.pfeg.noaa.gov/erddap/"
# 
# sstInfo <- rerddap::info('erdVHsstaWS3day', url = myURL)
# # get latest 3-day composite sst
# viirsSST <- rerddap::griddap(sstInfo,
#                              latitude = c(41., 31.),
#                              longitude = c(-128., -115),
#                              time = c('last','last'),
#                              fields = 'sst')
# 
# w <- map("worldHires", xlim = c(-140., -114), ylim = c(30., 42.), fill = TRUE, plot = FALSE)
# my_plot <- plotdap::add_griddap(plotdap::plotdap(mapData = w), viirsSST, ~sst, fill = "thermal" )
# 
# my_plot1 <- plotdap::bbox_set(my_plot, xlim = c(-128., -115), ylim = c(31, 41.), interactive = TRUE)
# 
# # must have '"tooltip = "text"' to work properly.
# plotly::ggplotly(my_plot1, tooltip = "text")
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# ?cache_delete
# ?cache_delete_all
# ?cache_details
# ?cache_list

## -----------------------------------------------------------------------------
require("ncdf4")
exampleFile <- system.file("extdata", "MWsstd1day.nc", package = "rerddap")
sstFile <- nc_open(exampleFile)

## -----------------------------------------------------------------------------
names(sstFile$dim)

## -----------------------------------------------------------------------------
sstFile$dim$longitude$vals

## -----------------------------------------------------------------------------
names(sstFile$var)

## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(1, 1, 1, 1), count = c(1, 1, -1, -1))

## -----------------------------------------------------------------------------
latMin <- which(sstFile$dim$latitude$vals == 30.0)
latMax <- which(sstFile$dim$latitude$vals == 30.5)
lonMin <- which(sstFile$dim$longitude$vals == 210.0)
lonMax <- which(sstFile$dim$longitude$vals == 210.5)

## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMin, latMin, 1, 1), count = c( (lonMax - lonMin + 1), (latMax - latMin + 1), 1, 1 ))

## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMax, latMin, 1, 1), count = c(1, 1, 1, -1 ))

## -----------------------------------------------------------------------------
latMin <- max(which(sstFile$dim$latitude$vals <= 30.1))
latMax <- min(which(sstFile$dim$latitude$vals >= 30.3))
lonMin <- max(which(sstFile$dim$longitude$vals <= 210.1))
lonMax <- min(which(sstFile$dim$longitude$vals >= 210.3))


## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMin, latMin, 1, 1), count = c( (lonMax - lonMin + 1), (latMax - latMin + 1), 1, 1 ))

#' @title global_search
#' @description Search for ERDDAP™ tabledap or griddap datasets from a list
#' of ERDDAP™ servers based on search terms.
#' @param query (character) Search terms
#' @param server_list (list of character) List of ERDDAP™ servers to search
#' @param which_service (character) One of tabledep or griddap.
#' @return If successful a dataframe wih columns:
#'  \itemize{
#'     \item title - the dataset title
#'     \item dataset_id - the datasetid on that ERDDAP™ server
#'     \item url - base url of dataset ERDDAP™ server
#'  }
#'  if urls are valid,  no match is found,  will return no match found
#'  else returns error message
#' @details Uses the 'reddap' function ed_search() to search over
#' the list of servers
#' @examples
#' # get list of servers know by
#' # https://irishmarineinstitute.github.io/awesome-erddap
#' # e_servers <- servers()$url
#' # select a couple to search
#' # e_servers <- e_servers[c(1, 40)]
#' # to meet CRAN time limits will only search 1 place
#' e_servers <- "https://coastwatch.pfeg.noaa.gov/erddap/"
#' test_query <- 'C-HARM v1 2-Day Forecast'
#' query_results <- global_search(test_query, e_servers, "griddap")
#' @seealso
#'  \code{\link[crul]{HttpClient}}
#' @rdname global_search
#' @export
global_search <- function(query, server_list, which_service) {
  #  check that input is of correct type
  check_arg(query, "character")
  check_arg(server_list, "character")
  check_arg(which_service, "character")
  which_service <- match.arg(which_service, c("tabledap","griddap"), FALSE)
  search_url <- list()
  for (iserv in seq(1, length(server_list))) {
    # set which server
    my_serv <- server_list[iserv]
    good_url <- TRUE
    # will do two tests on the server
    # First is if erddap is not in URL
    test_url <- grepl('erddap', my_serv, fixed = TRUE)
    if (!test_url) {   # erddap not in url
      print(paste('server URL ', my_serv), ' does not contain "erddap"')
      print('will be removed from server list')
      good_url <- FALSE
    }
    #  test if URL reachable
    if (good_url) {
      test_url <- crul::HttpClient$new(my_serv)
      test_url_head <- try(test_url$head(), silent = TRUE)
      if (suppressWarnings(class(test_url_head)[1] == 'try-error'))  {
        print(paste(' server URL', my_serv, ' not responding'))
        good_url <- FALSE
      }
    }
    # good URL, responding, make query
    if (good_url){
      temp_list <- try(ed_search(query = query, url = my_serv,
                                 which = which_service),
                       silent = TRUE)
      class_error <- suppressWarnings(class(temp_list)[1]  == 'try-error')
      if(!class_error) {
        temp_list <- temp_list$info
        temp_names <- names(temp_list)
        temp_list$url <- rep(my_serv, nrow(temp_list))
        names(temp_list) <- c(temp_names, 'url')
        search_url <- rbind(search_url, temp_list)
      }
    }
  }
  if(length(search_url) == 0) {
    print('no search results found')
    return('no search results found')
  } else {
    return(search_url)
  }
}

#' Get ERDDAP(TM) gridded data
#'
#' @export
#' @template griddap_params
#' @template griddap_egs
griddap <- function(datasetx, ..., fields = 'all', stride = 1, fmt = "nc",
  url = eurl(), store = disk(), read = TRUE, callopts = list()) {
  x <- datasetx
  calls <- names(sapply(match.call(), deparse))[-1]
  calls_vec <- "ncdf" %in% calls
  if (any(calls_vec)) {
    stop(
      "The parameter ncdf has been removed. We use ncdf4 package now internally",
         call. = FALSE)
  }

  dimargs <- list(...)
  if (length(dimargs) == 0) stop("no dimension arguments passed, see ?griddap")
  if (inherits(x, "info")) {
    url <- x$base_url
    message("info() output passed to x; setting base url to: ", url)
  } else {
    x <- as.info(x, url)
  }
  if (attr(x, "type") != "griddap")
    stop("datasetid '", attr(x, "datasetid"), "' not of type griddap")
  check_dims(dimargs, x)
  if (!is.null(dimargs$time)) {
    check_time_range(dimargs, x)
  }
  check_lat_text(dimargs)
  check_lon_text(dimargs)
  dimargs <- fix_dims(dimargs, .info = x)
  check_lon_data_range(dimargs, x)
  check_lat_data_range(dimargs, x)
  d <- attr(x, "datasetid")
  var <- field_handler(fields, x$variables$variable_name)
  dims <- dimvars(x)
  store <- toggle_store(fmt, store)
  if (all(var == "none")) {
    args <- paste0(sapply(dims, function(y) {
      parse_args(x, y, stride, dimargs, wname = TRUE)
    }), collapse = ",")
  } else {
    pargs <- sapply(dims, function(y) parse_args(x, y, stride, dimargs))
    args <- paste0(lapply(var, function(y) {
      paste0(y, paste0(pargs, collapse = ""))
    }), collapse = ",")
  }
 # fmt <- match.arg(fmt, c("nc", "csv"))
  fmt <- match.arg(fmt, c("nc", "csv", "parquet"))
  if (fmt == "parquet") {
    url_version <- version(url)
    url_version <- as.numeric(sub(".*=", "", url_version))
    if (url_version < 2.25) {
      print(paste0("Selected ERDDAP is version ", url_version))
      stop("ERDDAP version >= 2.25 is required for parquet - program stops")
    }
  }
  lenURL <- nchar(url)
  if (substr(url, lenURL, lenURL) != '/') {
    url <- paste0(url, '/')
  }
#  resp <- erd_up_GET(url = sprintf("%sgriddap/%s.%s", url, d, fmt), dset = d,
#                     args = args, store = store, fmt = fmt, callopts)
  erddap_ext <- if (fmt == "parquet") "parquetWMeta" else fmt
  resp <- erd_up_GET(url = sprintf("%sgriddap/%s.%s", url, d, erddap_ext),
                     dset = d, args = args, store = store, fmt = fmt, callopts)
  loc <- if (store$store == "disk") resp else "memory"
#  outclasses <- switch(fmt,
#                      nc = c("griddap_nc", "nc", "list"),
#                       csv = c("griddap_csv", "csv", "data.frame"))
  outclasses <- switch(fmt,
                       nc      = c("griddap_nc",  "nc",  "list"),
                       csv     = c("griddap_csv", "csv", "data.frame"),
                       parquet = c("griddap_nc",  "nc",  "list"))
  read <- toggle_read(read, store)
#  structure(
#    read_all(resp, fmt, read),
#    class = outclasses,
#    datasetid = d,
#    path = loc,
#    # url = url_build(sprintf("%sgriddap/%s.%s", url, d, fmt), args)
#   url = url_build(sprintf("%sgriddap/%s.%s", url, d, erddap_ext), args)
#  )
#}
  parsed <- read_all(resp, fmt, read)
  if (fmt == "parquet") {
    parsed <- list(
      summary = parquet_summary(x, parsed, var),
      data    = parsed
    )
  }
  structure(
    parsed,
    class = outclasses,
    datasetid = d,
    path = loc,
    url = url_build(sprintf("%sgriddap/%s.%s", url, d, erddap_ext), args)
  )
}


toggle_read <- function(x, store) {
  if (store$store == "memory") {
    return(TRUE)
  } else {
    return(x)
  }
}

#toggle_store <- function(fmt, store) {
#  if (fmt == "nc") {
#    if (store$store == "memory") {
#      disk()
#    } else {
#      store
#    }
#  } else {
#    store
#  }
#}

toggle_store <- function(fmt, store) {
  if (fmt %in% c("nc", "parquet")) {
    if (store$store == "memory") {
      disk()
    } else {
      store
    }
  } else {
    store
  }
}



#' @export
print.griddap_csv <- function(x, ...) {
  finfo <- file_info(attr(x, "path"))
  cat(sprintf("<ERDDAP(TM) griddap> %s", attr(x, "datasetid")), sep = "\n")
  path <- attr(x, "path")
  path2 <- if (file.exists(path)) path else "<beware: file deleted>"
  cat(sprintf("   Path: [%s]", path2), sep = "\n")
  if (attr(x, "path") != "memory") {
    cat(sprintf("   Last updated: [%s]", finfo$mtime), sep = "\n")
    cat(sprintf("   File size:    [%s mb]", finfo$size), sep = "\n")
  }
  cat(sprintf("   Dimensions:   [%s X %s]\n", NROW(x), NCOL(x)), sep = "\n")
  print(tibble::as_tibble(x))
}

#' @export
print.griddap_parquet <- function(x, ...) {
  finfo <- file_info(attr(x, "path"))
  cat(sprintf("<ERDDAP(TM) griddap> %s", attr(x, "datasetid")), sep = "\n")
  path <- attr(x, "path")
  path2 <- if (file.exists(path)) path else "<beware: file deleted>"
  cat(sprintf("   Path: [%s]", path2), sep = "\n")
  if (attr(x, "path") != "memory") {
    cat(sprintf("   Last updated: [%s]", finfo$mtime), sep = "\n")
    cat(sprintf("   File size:    [%s mb]", finfo$size), sep = "\n")
  }
  cat(sprintf("   Dimensions: [%s X %s]\n", NROW(x), NCOL(x)), sep = "\n")
  print(tibble::as_tibble(x))
}


#' @export
print.griddap_nc <- function(x, ...) {
  finfo <- file_info(attr(x, "path"))
  cat(sprintf("<ERDDAP(TM) griddap> %s", attr(x, "datasetid")), sep = "\n")
  path <- attr(x, "path")
  path2 <- if (file.exists(path)) path else "<beware: file deleted>"
  cat(sprintf("   Path: [%s]", path2), sep = "\n")
  if (attr(x, "path") != "memory") {
    cat(sprintf("   Last updated: [%s]", finfo$mtime), sep = "\n")
    cat(sprintf("   File size:    [%s mb]", finfo$size), sep = "\n")
  }
  cat(sprintf("   Dimensions (dims/vars):   [%s X %s]", x$summary$ndims, x$summary$nvars), sep = "\n")
  cat(sprintf("   Dim names: %s", paste0(names(x$summary$dim), collapse = ", ")), sep = "\n")
  cat(sprintf("   Variable names: %s", paste0(unname(sapply(x$summary$var, "[[", "longname")), collapse = ", ")), sep = "\n")
  cat(sprintf("   data.frame (rows/columns):   [%s X %s]", dim(x$data)[1], dim(x$data)[2]), sep = "\n\n")
  print(tibble::as_tibble(x$data))
}

parquet_summary <- function(x, data, var) {
  # x   : the info() object (has x$alldata with all variable/dim attributes)
  # data: data.frame from nanoparquet::read_parquet()
  # var : character vector of data variable names (from field_handler)
  
  dims <- dimvars(x)
  
  # build dim list — one entry per dimension variable
  dim_list <- lapply(dims, function(nm) {
    att       <- x$alldata[[nm]]
    units_val <- att$value[att$attribute_name == "units"]
    cal_val   <- att$value[att$attribute_name == "calendar"]
    structure(
      list(
        name          = nm,
        len           = length(unique(data[[nm]])),
        units         = if (length(units_val) > 0) units_val else NA_character_,
        calendar      = if (length(cal_val)   > 0) cal_val   else NULL,
        vals          = sort(unique(data[[nm]])),
        create_dimvar = TRUE
      ),
      class = "ncdim4"
    )
  })
  names(dim_list) <- dims
  
  # build var list — one entry per requested data variable
  var_list <- lapply(var, function(nm) {
    att          <- x$alldata[[nm]]
    units_val    <- att$value[att$attribute_name == "units"]
    long_val     <- att$value[att$attribute_name == "long_name"]
    missval_val  <- att$value[att$attribute_name == "missing_value"]
    fillval_val  <- att$value[att$attribute_name == "_FillValue"]
    prec_val     <- att$data_type[att$row_type == "variable"]
    missval <- if (length(missval_val) > 0) as.numeric(missval_val)
    else if (length(fillval_val) > 0) as.numeric(fillval_val)
    else NA_real_
    structure(
      list(
        name     = nm,
        ndims    = length(dims),
        units    = if (length(units_val) > 0) units_val else NA_character_,
        longname = if (length(long_val)  > 0) long_val  else nm,
        prec     = if (length(prec_val)  > 0) prec_val  else "float",
        missval  = missval,
        size     = vapply(dims, function(d) length(unique(data[[d]])),
                          integer(1))
      ),
      class = "ncvar4"
    )
  })
  names(var_list) <- var
  
  list(
    filename   = NA_character_,
    ndims      = length(dims),
    nvars      = length(var),
    natts      = nrow(x$alldata$NC_GLOBAL),
    dim        = dim_list,
    var        = var_list,
    unlimdimid = -1L
  )
}


field_handler <- function(x, y){
  x <- match.arg(x, c(y, "none", "all"), TRUE)
  if (length(x) == 1 && x == "all") {
    y
  } else if (all(x %in% y) || x == "none") {
    x
  }
}

check_dims <- function(dimargs, .info) {
  if (any(lengths(dimargs )!= 2)) {
    print("All coordinate bounds must be of length 2, even if same value")
    print("Present values are:") 
    print(dimargs)
    stop("rerddap halted", call. = FALSE)
  }
  if (!all(names(dimargs) %in% dimvars(.info))) {
    stop(sprintf("Some input dimensions (%s) don't match those in dataset (%s)",
                 paste0(names(dimargs), collapse = ", "),
                 paste0(dimvars(.info), collapse = ", ")), call. = FALSE)
  }
}

check_lon_text <- function(dimargs) {
  if (!is.null(dimargs$longitude)) {
    if (any(sapply(dimargs$longitude, class) == "character")) {
      txt <- dimargs$longitude[sapply(dimargs$longitude, class) == "character"]
      if (!all(grepl("last", txt))) stop("Only text values allowed are 'last' & variants on that", call. = FALSE)
    }
  }
}

check_lat_text <- function(dimargs) {
  if (!is.null(dimargs$latitude)) {
    if (any(sapply(dimargs$latitude, class) == "character")) {
      txt <- dimargs$latitude[sapply(dimargs$latitude, class) == "character"]
      if (!all(grepl("last", txt))) stop("Only text values allowed are 'last' & variants on that", call. = FALSE)
    }
  }
}

is_lon_text <- function(dimargs) {
  if (!is.null(dimargs$longitude)) {
    any(sapply(dimargs$longitude, class) == "character")
  } else {
    FALSE
  }
}

is_lat_text <- function(dimargs) {
  if (!is.null(dimargs$latitude)) {
    any(sapply(dimargs$latitude, class) == "character")
  } else {
    FALSE
  }
}

check_time_range <- function(dimargs, x) {
#  if(!class(dimargs$time) == 'character'){
  if(!is.character(dimargs$time)){
    print('time must be given as character strings')
    print('you are passing ', paste0(class(dimargs$time)))
    stop('rerddap halted', call. = FALSE)
  }
  global <- x$alldata$NC_GLOBAL
  tt <- global[ global$attribute_name %in%c('time_coverage_end','time_coverage_start'), "value", ]
  tt <- rev(tt)
  if (!('last' %in% dimargs$time)){
      if((dimargs$time[1] < tt[1]) | (dimargs$time[2] > tt[2])) {
        print('time bounds are out of range')
        print('You gave: ') 
        print(dimargs$time)
        print("Dataset times are: ")
        print(tt)
        stop('rerddap halted', call. = FALSE)
      }
  }
}


check_lon_data_range <- function(dimargs, .info) {
  if (!is.null(dimargs$longitude)) {
    val <- .info$alldata$longitude[ .info$alldata$longitude$attribute_name == "actual_range", "value"]
    val2 <- as.numeric(strtrim(strsplit(val, ",")[[1]]))
    if (!is_lon_text(dimargs)) {
      if (max(dimargs$longitude) > max(val2) || min(dimargs$longitude) < min(val2)) {
        stop(sprintf("One or both longitude values (%s) outside data range (%s)",
                     paste0(dimargs$longitude, collapse = ", "),
                     paste0(val2, collapse = ", ")), call. = FALSE)
      }
    }
  }
}

check_lat_data_range <- function(dimargs, .info) {
  if (!is.null(dimargs$latitude)) {
    val <- .info$alldata$latitude[ .info$alldata$latitude$attribute_name == "actual_range", "value"]
    val2 <- as.numeric(strtrim(strsplit(val, ",")[[1]]))
    if (!is_lat_text(dimargs)) {
      if (max(dimargs$latitude) > max(val2) || min(dimargs$latitude) < min(val2)) {
        stop(sprintf("One or both latitude values (%s) outside data range (%s)",
                     paste0(dimargs$latitude, collapse = ", "),
                     paste0(val2, collapse = ", ")), call. = FALSE)
      }
    }
  }
}

fix_dims <- function(dimargs, .info) {
  for (i in seq_along(dimargs)) {
    tmp <- dimargs[[i]]
    nm <- names(dimargs[i])
    tmp <- grep("last+", tmp, value = TRUE, invert = TRUE)
    if (nm == "time") {
      tmp <- as.Date(tmp)
    }
    
    val <- .info$alldata[[nm]][ .info$alldata[[nm]]$attribute_name == "actual_range", "value"]
    val2 <- as.numeric(strtrim(strsplit(val, ",")[[1]]))
    if (length(tmp) != 0) {
      if (which.min(val2) != which.min(tmp)) {
        dimargs[[i]] <- rev(dimargs[[i]])
      }
    }
    
    ## new
    # if (nm %in% c('latitude', 'longitude')) {
    if (nm != 'time') {
      z <- unlist(strsplit(.info$alldata[[nm]]$value[1], ","))
      spacing <- as.numeric(unlist(strsplit(z[3], "=")[[1]])[2])
      if ((!is.na(spacing)) & (spacing < 0)) {
        if (!(dimargs[[i]][1] > dimargs[[i]][2])) {
          dimargs[[i]] <- rev(dimargs[[i]])
        }
      }
    }
  }
  dimargs
}

parse_args <- function(.info, dim, s, dimargs, wname = FALSE){
  tmp <- if (dim %in% names(dimargs)) {
    dimargs[[dim]]
  } else {
    if (dim == "time") {
      times <- c(getvar(.info, "time_coverage_start"), getvar(.info, "time_coverage_end"))
      sprintf('[(%s):%s:(%s)]', times[1], s, times[2])
    } else {
      actrange <- foo(.info$alldata[[dim]], "actual_range")
      gsub("\\s+", "", strsplit(actrange, ",")[[1]])
    }
  }
  tmp <- format(tmp, scientific = FALSE)
  if (length(s) > 1) {
     if (!length(s) == length(dimvars(.info))) stop("Your stride vector must equal length of dimension variables", call. = FALSE)
    names(s) <- dimvars(.info)
    if (!wname) {
      sprintf('[(%s):%s:(%s)]', tmp[1], s[[dim]], tmp[2])
    } else {
      sprintf('%s[(%s):%s:(%s)]', dim, tmp[1], s[[dim]], tmp[2])
    }
  } else {
    if (!wname) {
      if (length(tmp) == 1) {
          tmp
      } else {
        sprintf('[(%s):%s:(%s)]', tmp[1], s, tmp[2])
      }
    } else {
      if (length(tmp) == 1) {
           tmp
      } else {
        sprintf('%s[(%s):%s:(%s)]', dim, tmp[1], s, tmp[2])
       }
    }
  }
}

getvar <- function(x, y){
  x$alldata$NC_GLOBAL[ x$alldata$NC_GLOBAL$attribute_name == y, "value"]
}

getvars <- function(x){
  vars <- names(x$alldata)
  vars[ !vars %in% c("NC_GLOBAL", "time", x$variables$variable_name) ]
}

getallvars <- function(x){
  vars <- names(x$alldata)
  vars[ !vars %in% "NC_GLOBAL" ]
}

dimvars <- function(x){
  vars <- names(x$alldata)
  vars[ !vars %in% c("NC_GLOBAL", x$variables$variable_name) ]
}

erd_up_GET <- function(url, dset, args, store, fmt, callopts) {
  if (length(args) > 0) url <- sprintf("%s?%s", url, args)
  url1 <- url
  url1 <- gsub('\\[', '%5B', url1)
  url1 <- gsub('\\]', '%5D', url1)
  cli <- crul::HttpClient$new(url = url1, opts = callopts)
  if (store$store == "disk") {
    # store on disk
    key <- gen_key(url, args, fmt)
    if ( file.exists(file.path(store$path, key)) ) {
      file.path(store$path, key)
    } else {
      dir.create(store$path, showWarnings = FALSE, recursive = TRUE)
      if (!store$overwrite) {
        stop('overwrite was `FALSE`, see ?disk')
      }
      # res <- cli$get(disk = file.path(store$path, key))
      response <- tryCatch(
        {
          res <- cli$get(disk = file.path(store$path, key)) # Attempt to fetch
        },
        error = function(e) {
          message("Curl request failed to get file: ", e$message)
          quit(save = "no", status = 1)  # Gracefully exit R session
        }
      )
      # delete file if error, and stop message
      err_handle(res, store, key)
      # return file path
      res$content
    }
  } else {
    # read into memory, bypass disk storage
    # res <- cli$get()
    response <- tryCatch(
      {
        res <- cli$get()  # Attempt to fetch
      },
      error = function(e) {
        message("Curl request failed to get grid data: ", e$message)
        quit(save = "no", status = 1)  # Gracefully exit R session
      }
    )
    # if error stop message
    err_handle(res, store, key)
    # return response object
    res
  }
}

writepath <- function(path, d, fmt) file.path(path, paste0(d, ".", fmt))

gen_key <- function(url, args, fmt) {
  ky <- paste0(url, "?", args)
  paste0(digest::digest(ky), ".", fmt)
}

# libfile <- function() file.path(path.expand("~/.rerddap"), ".library")

# hash_file <- function(x) {
#   if (!file.exists(x)) {
#     cat("\n", file = x)
#   }
# }

# write_key <- function(path, key) {
#   hash_file(path)
#   cat("- ", key, file = path, append = TRUE)
# }

file_info <- function(x) {
  tmp <- file.info(x)
  row.names(tmp) <- NULL
  tmp2 <- tmp[,c('mtime', 'size')]
  tmp2$size <- round(tmp2$size/1000000L, 2)
  tmp2
}

strextract <- function(str, pattern) regmatches(str, regexpr(pattern, str))

strtrim <- function(str) gsub("^\\s+|\\s+$", "", str)

#' Convert a CF Standard Name to/from a GCMD Science Keyword
#'
#' @export
#' @param cf character; A cf standard name
#' http://cfconventions.org/Data/cf-standard-names/27/build/cf-standard-name-table.html
#' @param gcmd character; A GCMD science keyword
#' http://gcmd.gsfc.nasa.gov/learn/keyword_list.html
#' @param url A URL for an ERDDAP™ server. Default:
#' https://upwell.pfeg.noaa.gov/erddap/. See [eurl()] for 
#' more information
#' @param ... Curl options passed on to [crul::verb-GET]
#' @examples  \dontrun{
#' key_words(cf = "air_pressure")
#' cat(key_words(cf = "air_pressure"))
#'
#' # a different ERDDAP™ server
#' # key_words(cf = "air_pressure", url = servers()$url[6])
#' }

key_words <- function(cf = NULL, gcmd = NULL, url = eurl(), ...){
  either_or_keywords(cf, gcmd)
  args <- rc(list(cf = cf, gcmd = gcmd))
  cli <- crul::HttpClient$new(url = file.path(pu(url), 'convert/keywords.txt'), 
    opts = list(...))
  # res <- cli$get(query = args)
  response <- tryCatch(
    {
      res <- cli$get(query = args)  # Attempt to fetch
    },
    error = function(e) {
      message("Curl request failed on finding keyword: ", e$message)
      quit(save = "no", status = 1)  # Gracefully exit R session
    }
  )
  
  res$raise_for_status()
  res$parse("UTF-8")
}

either_or_keywords <- function(cf, gcmd) {
  if (!xor(!is.null(cf), !is.null(gcmd))) {
    stop("Provide either cf or gcmd, not both", call. = FALSE)
  }
}

# Function to get UTM zone from a single longitude and latitude pair
# originally from David LeBauer I think
# @param lon Longitude, in decimal degree style
# @param lat Latitude, in decimal degree style
long2utm <- function(lon, lat) {
  if (56 <= lat & lat < 64) {
    if (0 <= lon & lon < 3) { 31 } else
      if (3 <= lon & lon < 12) { 32 } else { NULL }
  } else {
    if (72 <= lat) {
      if (0 <= lon & lon < 9) { 31 } else
        if (9 <= lon & lon < 21) { 33 } else
          if (21 <= lon & lon < 33) { 35 } else
            if (33 <= lon & lon < 42) { 37 } else { NULL }
    }
    (floor((lon + 180)/6) %% 60) + 1
  }
}

rc <- function(l) Filter(Negate(is.null), l)

read_csv <- function(x){
  tmp <- read.csv(x, header = FALSE, sep = ",", stringsAsFactors = FALSE,
                  skip = 3)
  nmz <- names(read.csv(x, header = TRUE, sep = ",", stringsAsFactors = FALSE,
                        skip = 1, nrows = 1))
  names(tmp) <- tolower(nmz)
  tmp
}

read_data <- function(x, nrows = -1){
  if (inherits(x, "HttpResponse")) {
    x <- x$parse("UTF-8")
    tmp <- read.csv(text = x, header = FALSE, sep = ",",
                    stringsAsFactors = FALSE, skip = 2, nrows = nrows)
    nmz <- names(read.csv(text = x, header = TRUE, sep = ",",
                          stringsAsFactors = FALSE, nrows = 1))
  } else {
    tmp <- read.csv(x, header = FALSE, sep = ",", stringsAsFactors = FALSE,
                    skip = 2, nrows = nrows)
    nmz <- names(read.csv(x, header = TRUE, sep = ",",
                          stringsAsFactors = FALSE, nrows = 1))
  }
  stats::setNames(tmp, tolower(nmz))
}

#read_all <- function(x, fmt, read) {
#  switch(fmt,
#         csv = {
#           if (read) {
#            read_data(x)
#           } else {
#             read_data(x, 10)
#          }
#        },
#         nc = {
#           if (read) {
#             ncdf4_get(x)
#           } else {
#             ncdf_summary(x)
#           }
#         }
#  )
#}

read_all <- function(x, fmt, read) {
  switch(fmt,
         csv = {
           if (read) {
             read_data(x)
           } else {
             read_data(x, 10)
           }
         },
         nc = {
           if (read) {
             ncdf4_get(x)
           } else {
             ncdf_summary(x)
           }
         },
         parquet = {
           out <- nanoparquet::read_parquet(x)
           if ("time" %in% names(out)) {
             out$time <- format(out$time, "%Y-%m-%dT%H:%M:%SZ")
           }
           out
           }  
         )
}


read_table <- function(x, fmt){
  if (inherits(x, "HttpResponse")) {
    txt <- gsub('\n$', '', x$parse("UTF-8"))
    read.csv(text = txt, sep = ",", stringsAsFactors = FALSE,
             blank.lines.skip = FALSE)[-1, , drop = FALSE]
  } else {
    if (fmt =='csv') {
      read.delim(x, sep = ",", stringsAsFactors = FALSE,
                 blank.lines.skip = FALSE)[-1, , drop = FALSE]
    }  else {
      temp_data <- nanoparquet::read_parquet(x)
    }
  }
}

replace_value_with_na <- function(x, fillValue) {
  if (is.numeric(x)) {
    test_value <- as.numeric(fillValue)
    x[x == test_value] <- NA
  }
  return(x)
}


pu <- function(x) sub("/$|//$", "", x)

strect <- function (str, pattern) regmatches(str, regexpr(pattern, str))


err_handle_old <- function(x, store, key) {
  if (x$status_code > 201) {
    tt <- if (store$store == "disk") x$content else x$parse("UTF-8")
    html <- tryCatch(read_html(tt), error = function(e) e)
    if (inherits(html, "error")) {
      mssg <- tt
    } else {
      mssg <- xml_text(xml_find_all(html, '//body//p[text()[contains(., "Error") or contains(., "error")]]')) %||% ""
      mssg <- sub("Message\\s", "", mssg)
      if (!nzchar(mssg)) {
        mssg <- strect(xml_text(xml_find_first(html, '//body')), "Query error.+")
      }
    }
    if (store$store != "memory") unlink(file.path(store$path, key))
    stop(paste0(mssg, collapse = "\n\n"), call. = FALSE)
  }
}

err_handle <- function(x, store, key) {
  if (x$status_code > 201) {
    tt <- if (store$store == "disk") x$content else x$parse("UTF-8")
    html <- tryCatch(read_html(tt), error = function(e) e)
    
    if (inherits(html, "error")) {
      # read_html() failed -> just use whatever tt is
      mssg <- tt
    } else {
      # Try to find ANY element under <body> containing "Error"/"error"
      nodes <- xml_find_all(
        html,
        '//body//*[contains(., "Error") or contains(., "error")]'
      )
      
      if (length(nodes) > 0) {
        mssg <- xml_text(nodes)
        mssg <- paste(mssg, collapse = "\n")
      } else {
        # Fallback: work with entire body text
        body_txt <- xml_text(xml_find_first(html, "//body"))
        
        # ERDDAP-style "Error { ... message="..." ... }"
        if (grepl("message=", body_txt, fixed = TRUE)) {
          # Allow internal quotes in the message
          mssg <- sub('.*message="(.*)"\\s*;.*', '\\1', body_txt)
        } else {
          mssg <- body_txt
        }
      }
      
      # Your original cleanup and regex fallback
      mssg <- sub("Message\\s", "", mssg)
      
      if (!nzchar(mssg)) {
        mssg <- strect(xml_text(xml_find_first(html, "//body")), "Query error.+")
      }
    }
    
    if (store$store != "memory") unlink(file.path(store$path, key))
    
    # This will always raise an error in R
    stop(paste0(mssg, collapse = "\n\n"), call. = FALSE)
  }
}



err_handle2 <- function(x) {
  if (x$status_code > 201) {
    tt <- x$parse("UTF-8")
    mssg <- xml_text(xml_find_all(read_html(tt), "//h1"))
    stop(paste0(mssg, collapse = "\n\n"), call. = FALSE)
  }
}

erddap_GET <- function(url, args = NULL, ...) {
  cli <- crul::HttpClient$new(url = url, opts = list(...))
  # tt <- cli$get(query = args)
  response <- tryCatch(
    {
      tt <- cli$get(query = args)  # Attempt to fetch
    },
    error = function(e) {
      message("Curl request failed: ", e$message)
      quit(save = "no", status = 1)  # Gracefully exit R session
    }
  )
  
  err_handle2(tt)
  stopifnot(tt$response_headers$`content-type` == 'application/json;charset=UTF-8')
  out <- tt$parse("UTF-8")
  jsonlite::fromJSON(out, FALSE)
}

url_build <- function(url, args = NULL) {
  if (is.null(args)) return(url)
  paste0(url, "?", args)
}

`%||%` <- function (x, y) {
  if (is.null(x) || length(x) == 0) y else x
}

#' Get ERDDAP™ tabledap data.
#'
#' @export
#'
#' @param x Anything coercable to an object of class info. So the output of
#' a call to [info()], or a datasetid, which will internally be passed
#' through [info()]
#' @param ... Any number of key-value pairs in quotes as query constraints.
#' See Details & examples
#' @param fields Columns to return, as a character vector
#' @param distinct If `TRUE` ERDDAP™ will sort all of the rows in the results
#' table (starting with the first requested variable, then using the second
#' requested variable if the first variable has a tie, ...), then remove all
#' non-unique rows of data. In many situations, ERDDAP™ can return distinct
#' values quickly and efficiently. But in some cases, ERDDAP™ must look through
#' all rows of the source dataset.
#' @param orderby If used, ERDDAP™ will sort all of the rows in the results
#' table (starting with the first variable, then using the second variable
#' if the first variable has a tie, ...). Normally, the rows of data in the
#' response table are in the order they arrived from the data source. orderBy
#' allows you to request that the results table be sorted in a specific way.
#' For example, use `orderby=c("stationID,time")` to get the results
#' sorted by stationID, then time. The orderby variables MUST be included in
#' the list of requested variables in the fields parameter.
#' @param orderbymax Give a vector of one or more fields, that must be included
#' in the fields parameter as well. Gives back data given constraints. ERDDAP™
#' will sort all of the rows in the results table (starting with the first
#' variable, then using the second variable if the first variable has a
#' tie, ...) and then just keeps the rows where the value of the last sort
#' variable is highest (for each combination of other values).
#' @param orderbymin Same as `orderbymax` parameter, except returns
#' minimum value.
#' @param orderbyminmax Same as `orderbymax` parameter, except returns
#' two rows for every combination of the n-1 variables: one row with the
#' minimum value, and one row with the maximum value.
#' @param fmt whether download should be as '.csv' (default) or '.parquet'
#' @param units One of 'udunits' (units will be described via the UDUNITS
#' standard (e.g.,degrees_C)) or 'ucum' (units will be described via the
#' UCUM standard (e.g., Cel)).
#' @param url A URL for an ERDDAP™ server.
#' Default: https://upwell.pfeg.noaa.gov/erddap/ - See [eurl()] for 
#' more information
#' @param store One of `disk` (default) or `memory`. You can pass
#' options to `disk`
#' @param callopts Curl options passed on to [crul::verb-GET] (must be
#' named parameters)
#'
#' @return An object of class \code{tabledap}. This class is a thin wrapper
#' around a data.frame, so the data you get back is a data.frame with metadata
#' attached as attributes (datasetid, path (path where the csv is stored on
#' your machine), url (url for the request))
#'
#' @details
#' For key-value pair query constraints, the valid operators are =,
#' != (not equals), =~ (a regular expression test), <, <=, >, and >= . For
#' regular expressions you need to add a regular expression. For others, nothing
#' more is needed. Construct the entry like \code{'time>=2001-07-07'} with the
#' parameter on the left, value on the right, and the operator in the middle,
#' all within a set of quotes. Since ERDDAP accepts values other than \code{=},
#' we can't simply do \code{time = '2001-07-07'} as we normally would.
#'
#' Server-side functionality: Some tasks are done server side. You don't have
#' to worry about what that means. They are provided via parameters in this
#' function. See \code{distinct}, \code{orderby}, \code{orderbymax},
#' \code{orderbymin}, \code{orderbyminmax}, and \code{units}.
#'
#' Data is cached based on all parameters you use to get a dataset, including
#' base url, query parameters. If you make the same exact call in the same or
#' a different R session, as long you don't clear the cache, the function only
#' reads data from disk, and does not have to request the data from the web
#' again.
#'
#' If you run into an error like "HTTP Status 500 - There was a (temporary?)
#' problem. Wait a minute, then try again.". it's likely they are hitting
#' up against a size limit, and they should reduce the amount of data they
#' are requesting either via space, time, or variables. Pass in
#' \code{config = verbose()} to the request, and paste the URL into your
#' browser to see if the output is garbled to examine if there's a problem
#' with servers or this package
#'
#' @references https://upwell.pfeg.noaa.gov/erddap/index.html
#' @examples \dontrun{
#' # Just passing the datasetid without fields gives all columns back
#' tabledap('erdCinpKfmBT')
#'
#' # Pass time constraints
#' tabledap('erdCinpKfmBT', 'time>=2006-08-24')
#'
#' # Pass in fields (i.e., columns to retrieve) & time constraints
#' tabledap('erdCinpKfmBT',
#'   fields = c('longitude', 'latitude', 'Aplysia_californica_Mean_Density'),
#'   'time>=2006-08-24'
#' )
#'
#' # Get info on a datasetid, then get data given information learned
#' info('erdCalCOFIlrvsiz')$variables
#' tabledap('erdCalCOFIlrvsiz', fields=c('latitude','longitude','larvae_size',
#'    'itis_tsn'), 'time>=2011-10-25', 'time<=2011-10-31')
#'
#' # An example workflow
#' ## Search for data
#' (out <- ed_search(query='fish', which = 'table'))
#' ## Using a datasetid, search for information on a datasetid
#' id <- out$alldata[[1]]$dataset_id
#' vars <- info(id)$variables
#' ## Get data from the dataset
#' vars$variable_name[1:3]
#' tabledap(id, fields = vars$variable_name[1:3])
#'
#' # Time constraint
#' ## Limit by time with date only
#' (info <- info('erdCinpKfmBT'))
#' tabledap(info, fields = c(
#'   'latitude','longitude','Haliotis_fulgens_Mean_Density'),
#'   'time>=2001-07-14')
#'
#' # Use distinct parameter - compare to distinct = FALSE
#' tabledap('sg114_3',
#'    fields=c('longitude','latitude','trajectory'),
#'    'time>=2008-12-05', distinct = TRUE)
#'
#' # Use units parameter
#' ## In this example, values are the same, but sometimes they can be different
#' ## given the units value passed
#' tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
#'    'time>=2007-09-19', 'time<=2007-09-21', units='udunits')
#' tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
#'    'time>=2007-09-19', 'time<=2007-09-21', units='ucum')
#'
#' # Use orderby parameter
#' tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
#'    'time>=2007-09-19', 'time<=2007-09-21', orderby='temperature')
#' # Use orderbymax parameter
#' tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
#'    'time>=2007-09-19', 'time<=2007-09-21', orderbymax='temperature')
#' # Use orderbymin parameter
#' tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
#'    'time>=2007-09-19', 'time<=2007-09-21', orderbymin='temperature')
#' # Use orderbyminmax parameter
#' tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
#'    'time>=2007-09-19', 'time<=2007-09-21', orderbyminmax='temperature')
#' # Use orderbymin parameter with multiple values
#' tabledap('erdCinpKfmT',
#'    fields=c('longitude','latitude','time','depth','temperature'),
#'    'time>=2007-06-10', 'time<=2007-09-21',
#'    orderbymax=c('depth','temperature')
#' )
#'
#' # Integrate with taxize
#' out <- tabledap('erdCalCOFIlrvcntHBtoHI',
#'    fields = c('latitude','longitude','scientific_name','itis_tsn'),
#'    'time>=2007-06-10', 'time<=2007-09-21'
#' )
#' tsns <- unique(out$itis_tsn[1:100])
#' library("taxize")
#' classif <- classification(tsns, db = "itis")
#' head(rbind(classif)); tail(rbind(classif))
#'
#' # Write to memory (within R), or to disk
#' (out <- info('erdCinpKfmBT'))
#' ## disk, by default (to prevent bogging down system w/ large datasets)
#' ## the 2nd call is much faster as it's mostly just the time of reading
#' ## in the table from disk
#' system.time( tabledap('erdCinpKfmBT', store = disk()) )
#' system.time( tabledap('erdCinpKfmBT', store = disk()) )
#' ## memory
#' tabledap('erdCinpKfmBT', store = memory())
#'
#' # use a different ERDDAP™ server
#' ## NOAA IOOS NERACOOS
#' url <- "http://www.neracoos.org/erddap/"
#' tabledap("E01_optics_hist", url = url)
#' }

tabledap <- function(x, ..., fields=NULL, distinct=FALSE, orderby=NULL,
  orderbymax=NULL, orderbymin=NULL, orderbyminmax=NULL, units=NULL, fmt = 'csv',
  url = eurl(), store = disk(), callopts=list()) {
  

  if (inherits(x, "info")) {
    url <- x$base_url
    message("info() output passed to x; setting base url to: ", url)
  }
  x <- as.info(x, url)
  
  # if fmt is parquet,  check the ERDDAP version
  
  if (fmt == 'parquet') {
    url_version <- version(url)
    url_version <- as.numeric(sub(".*=", "", url_version))
    if (url_version < 2.25)  {
      print(paste0('Selected ERDDAP is version ', url_version))
      stop('ERDDAP version greater than 2.25 is required for parquet - program stops')
    }
  }
  
  fields <- paste(fields, collapse = ",")
  lenURL <- nchar(url)
  if (substr(url, lenURL, lenURL) != '/') {
    url <- paste0(url, '/')
  }
  if (fmt == 'csv') {
    url <- sprintf(paste0(url, "tabledap/%s.csv?%s"), attr(x, "datasetid"),
                   fields)
  } else if (fmt == 'parquet') {
    url <- sprintf(paste0(url, "tabledap/%s.parquetWMeta?%s"), attr(x, "datasetid"),
                   fields)
  } else {
     print(paste0('format given is ', fmt))
     stop('fmt must be either csv or parquet')
  }
  args <- list(...)
  distinct <- if (distinct) 'distinct()' else NULL
  units <- if (!is.null(units)) {
    makevar(toupper(units), 'units("%s")')
  } else {
    units
  }
  orderby <- makevar(orderby, 'orderBy("%s")')
  orderbymax <- makevar(orderbymax, 'orderByMax("%s")')
  orderbymin <- makevar(orderbymin, 'orderByMin("%s")')
  orderbyminmax <- makevar(orderbyminmax, 'orderByMinMax("%s")')
  moreargs <- rc(list(distinct, orderby, orderbymax, orderbymin,
                      orderbyminmax, units))
  args <- c(args, moreargs)
  args <- lapply(args, URLencode, reserved = TRUE)
  args <- paste0(args, collapse = "&")
  if (!nchar(args[[1]]) == 0) {
    url <- paste0(url, '&', args)
  }
  resp <- erd_tab_GET(url, dset = attr(x, "datasetid"), store, fmt, callopts)
  loc <- if (store$store == "disk") resp else "memory"
  temp_table <- read_table(resp, fmt)
  # change response type if csv
  if (fmt == 'csv'){
    dds_url <- sub('csv', 'dds', url)
    # strip off constraints
    amp_location <- regexpr("&", dds_url)
    if (amp_location[1] > 0) {
      dds_url <- substr(dds_url, 1, amp_location[1] - 1)
    }
    dds <- try(suppressWarnings(utils::read.table(dds_url)), silent = TRUE)
    # if (class(dds) == 'try-error') {
    if (methods::is(dds, 'try-error')) {
      print('failed to get variable datatype information')
      print('will leave units unchanged')
    } else{
      temp_table <- set_units(temp_table, dds)
    }
  }
  
  # go through columns get units to add as an attribute
  # if parquet file also set missing value to NA
  temp_table_names <- colnames(temp_table)
  icount = 0
  for (myName in temp_table_names){
    icount <- icount + 1
    units_loc <- which(x$alldata[[myName]]$attribute_name == 'units')
    if (length(units_loc) > 0) {
      temp_units <- x$alldata[[myName]]$value[[units_loc]]
      if (icount == 1){
        temp_table_units <- temp_units
      } else {
        temp_table_units <- c(temp_table_units, temp_units)
      }
    } else {
      if (icount == 1){
        temp_table_units <- NA
      } else {
        temp_table_units <- c(temp_table_units, NA)
      }
    }
    fillLoc <- which(x$alldata[[myName]]$attribute_name == '_FillValue')
  }
   
    
  structure(
    #read_table(resp),
    temp_table,
    class = c("tabledap", "data.frame"),
    datasetid = attr(x, "datasetid"),
    path = loc,
    url = url,
    units = temp_table_units
  )
}

#' @export
print.tabledap <- function(x, ...) {
  finfo <- file_info(attr(x, "path"))
  cat(sprintf("<ERDDAP tabledap> %s", attr(x, "datasetid")), sep = "\n")
  path <- attr(x, "path")
  path2 <- if (file.exists(path)) path else "<beware: file deleted>"
  cat(sprintf("   Path: [%s]", path2), sep = "\n")
  if (attr(x, "path") != "memory") {
    cat(sprintf("   Last updated: [%s]", finfo$mtime), sep = "\n")
    cat(sprintf("   File size:    [%s mb]", finfo$size), sep = "\n")
  }
  print(tibble::as_tibble(x))
}

erd_tab_GET <- function(url, dset, store, fmt, callopts) {
  cli <- crul::HttpClient$new(url = url, opts = callopts)
  if (store$store == "disk") {
    # store on disk
    if (fmt == 'csv') {
      key <- gen_key(url, NULL, "csv")
    } else {
      key <- gen_key(url, NULL, "parquet")
    }
    if ( file.exists(file.path(store$path, key)) ) {
      file.path(store$path, key)
    } else {
      dir.create(store$path, showWarnings = FALSE, recursive = TRUE)
      if (!store$overwrite) {
        stop('overwrite was `FALSE`, see ?disk')
      }
      # res <- cli$get(disk = file.path(store$path, key))
      response <- tryCatch(
        {
          res <- cli$get(disk = file.path(store$path, key))  # Attempt to fetch
        },
        error = function(e) {
          message("Curl request failed to get table: ", e$message)
          quit(save = "no", status = 1)  # Gracefully exit R session
        }
      )
      err_handle(res, store, key)
      res$content
    }
  } else {
    # res <- cli$get()
    response <- tryCatch(
      {
        res <- cli$get()  # Attempt to fetch
      },
      error = function(e) {
        message("Curl request failed to get table: ", e$message)
        quit(save = "no", status = 1)  # Gracefully exit R session
      }
    )
    err_handle(res, store, key)
    res
  }
}

makevar <- function(x, y){
  if (!is.null(x)) {
    x <- paste0(x, collapse = ",")
    sprintf(y, x)
  } else {
    NULL
  }
}

set_units <- function(temp_t, dds) {
    for (i in seq(3, (nrow(dds) - 2))) {
        var_type <- dds[i, 1]
        var_name <- dds[i, 2]
        var_name <- substr(var_name, 1, nchar(var_name)-1)
        if (var_name == 'time') {
            # make certain lubridate can convert time
            temp_time <- suppressWarnings(lubridate::as_datetime(temp_t$time))
           if (!any(is.na(temp_time))) {
               temp_t$time <- temp_time
           }
       } else {
           if (grepl('Float', var_type)) {
               temp_t[[var_name]] <- as.numeric(temp_t[[var_name]])
           } else if(grepl('Int', var_type)) {
              temp_t[[var_name]] <- as.integer(temp_t[[var_name]])
           }
       }
    } 
  return(temp_t)
}

#' List cached files
#'
#' @export
#' @family cache
#' @examples \dontrun{
#' # list files in cache
#' cache_list()
#'
#' # List info for files
#' ## download some data first
#' tabledap('erdCinpKfmBT')
#' griddap('erdVHNchlamday',
#'  time = c('2015-04-01','2015-04-10'),
#'  latitude = c(18, 21),
#'  longitude = c(-120, -119)
#' )
#' 
#' (x <- cache_list())
#' cache_details(x$nc[1])
#' cache_details(x$csv[1])
#' cache_details()
#'
#' # delete files by name in cache
#' # cache_delete(x$nc[1])
#' # cache_delete(x$nc[2:3])
#' }
cache_list <- function() {
  nc_files <- list.files(rrcache$cache_path_get(), pattern = ".nc")
  csv_files <- list.files(rrcache$cache_path_get(), pattern = ".csv")
  structure(list(nc = nc_files, csv = csv_files), class = "rerddap_cache")
}

#' @export
print.rerddap_cache <- function(x, ...) {
  cat("<rerddap cached files>", sep = "\n")
  cat(" NetCDF files: ", strwrap(x$nc, indent = 5), sep = "\n")
  cat(" CSV files: ", strwrap(x$csv, indent = 5), sep = "\n")
}

#' cmocean colors
#' The cmocean color palette by Kristen Thyng as implemented in the R package "oce"
#'
#' str(colors)
#' List of 13
#'  $ viridis
#'  $ cdom
#'  $ chlorophyll
#'  $ density
#'  $ freesurface
#'  $ oxygen
#'  $ par
#'  $ phase
#'  $ salinity
#'  $ temperature
#'  $ turbidity
#'  $ velocity
#'  $ vorticity
"colors"

#' Options for saving ERDDAP datasets.
#'
#' @export
#' @param path Path to store files in. A directory, not a file.
#' Default: the root cache path, see \code{\link{cache_setup}}
#' @param overwrite (logical) Overwrite an existing file of the same name?
#' Default: \code{TRUE}
disk <- function(path = NULL, overwrite = TRUE) {
  if (is.null(path)) {
    # path is NULL - use cache path already setup
    path <- rrcache$cache_path_get()
  }
  list(store = "disk", path = path, overwrite = overwrite)
}

#' @export
#' @rdname disk
memory <- function() list(store = "memory")

#' Convert a CF Standard Name to/from a GCMD Science Keyword
#'
#' @export
#' @param udunits character; A UDUNITS character string
#' https://www.unidata.ucar.edu/software/udunits/
#' @param ucum character; A UCUM character string
#' https://ucum.org/ucum.html
#' @param url Base URL of the ERDDAP server. See [eurl()] for 
#' more information
#' @param ... Curl options passed on to [crul::verb-GET]
#' @examples  \dontrun{
#' convert_units(udunits = "degree_C meter-1")
#' convert_units(ucum = "Cel.m-1")
#' }

convert_units <- function(udunits = NULL, ucum = NULL, url = eurl(), ...) {
  check1notboth(udunits, ucum)
  args <- rc(list(UDUNITS = udunits, UCUM = ucum))
  cli <- crul::HttpClient$new(url = file.path(pu(url), 'convert/units.txt'), 
    opts = list(...))
  # res <- cli$get(query = args)
  response <- tryCatch(
    {
      res <- cli$get(query = args)  # Attempt to fetch
    },
    error = function(e) {
      message("Curl request failed to convert units: ", e$message)
      quit(save = "no", status = 1)  # Gracefully exit R session
    }
  )
  res$raise_for_status()
  res$parse("UTF-8")
}

#' Estimate the size of a rerddap::griddap() request
#'
#' Uses coordinate metadata from an ERDDAP info() object to estimate how
#' many grid cells will be returned and the total uncompressed byte count for
#' each requested data variable.  No network request is made.
#'
#' Coordinate dimension names are taken directly from the info object using
#' the same logic as rerddap:::dimvars() — the set-difference between all
#' keys in info$alldata and the data variable names plus "NC_GLOBAL".  This
#' means any coordinate system works: geographic lat/lon, projected x/y,
#' sigma-layer depth, ROMS xi_rho/eta_rho, etc.
#'
#' Dimension constraints are passed via ... using the exact coordinate
#' variable names reported by the dataset, the same way griddap() accepts
#' them.  Each constraint is a numeric (or character for time) vector of
#' length 2: c(min, max).
#'
#' Spacing is resolved in this order for each dimension:
#'   1. User-supplied override in the spacing argument.
#'   2. For time: (t_max - t_min) / (nValues - 1) derived from the
#'      dimension's nValues row and actual_range attribute.  This correctly
#'      handles running composites where time steps are daily even though the
#'      composite window is e.g. 8 days.
#'   3. NC_GLOBAL attributes: geospatial_lat_resolution,
#'      geospatial_lon_resolution, time_coverage_resolution (ISO 8601).
#'   4. Coordinate variable attributes: point_spacing, resolution, spacing.
#'   5. For time: averageSpacing string from the nValues row as a last resort.
#'   6. If the constraint min == max the dimension contributes 1 point.
#'   7. Otherwise: NA — a warning is issued and 1 point is assumed.
#'
#' @param info    An object returned by rerddap::info().
#' @param ...     Named dimension constraints, one per coordinate variable to
#'                constrain.  Names must exactly match the coordinate variable
#'                names in the dataset (as returned by dimvars(info)).
#'                Each value is c(min, max); for time use ISO 8601 strings.
#' @param fields  Character vector of data variable names, or "all" (default).
#' @param stride  Integer scalar or named list of per-dimension stride values,
#'                using the same convention as griddap().  Default 1.
#' @param spacing Optional named list to override auto-detected spacing for
#'                one or more dimensions.  For time, supply seconds as
#'                time_sec (e.g. spacing = list(time_sec = 86400)).
#'                For other dimensions use the coordinate variable name
#'                (e.g. spacing = list(latitude = 0.01, xi_rho = 1)).
#' @param verbose Logical; print a formatted summary (default TRUE).
#'
#' @return Invisibly, a named list containing per-dimension point counts,
#'         spacing values, per-variable byte estimates, and total bytes.
#'
#' @examples
#' \dontrun{
#'   library(rerddap)
#'
#'   myURL <- "https://coastwatch.pfeg.noaa.gov/erddap/"
#'   response <- try(httr::HEAD(myURL, httr::timeout(10)), silent = TRUE)
#'   if (inherits(response, "try-error")) {
#'      stop("The ERDDAP\u2122 server is not responding")
#'   }
#'   info <- rerddap::info("erdMH1chla8day", url = myURL)
#'   estimate_griddap_size(info,
#'     latitude  = c(30, 50),
#'     longitude = c(-140, -110),
#'     time      = c("2020-01-01", "2020-12-31"))
#' }

estimate_griddap_size <- function(info,
                                  ...,
                                  fields  = "all",
                                  stride  = 1L,
                                  spacing = list(),
                                  verbose = TRUE) {
  
  # ── helpers ──────────────────────────────────────────────────────────────
  
  `%||%` <- function(a, b) if (!is.null(a)) a else b
  
  .attr <- function(var_name, attr_names) {
    if (!var_name %in% names(info$alldata)) return(NA_character_)
    df  <- info$alldata[[var_name]]
    for (a in attr_names) {
      val <- df$value[df$attribute_name == a]
      if (length(val) && !is.na(val[1L]) && nzchar(val[1L]))
        return(as.character(val[1L]))
    }
    NA_character_
  }
  
  .bytes <- function(dtype) {
    switch(tolower(as.character(dtype)),
           "float"   = , "float32" = , "real"    = 4L,
           "double"  = , "float64" = 8L,
           "int"     = , "int32"   = , "integer" = 4L,
           "short"   = , "int16"   = 2L,
           "byte"    = , "int8"    = , "ubyte"   = 1L,
           "long"    = , "int64"   = 8L,
           "ushort"  = , "uint16"  = 2L,
           "uint"    = , "uint32"  = 4L,
           "string"  = , "char"    = 1L,
           8L
    )
  }
  
  .fmt <- function(b) {
    if      (b < 1024^1) sprintf("%.1f B",  b)
    else if (b < 1024^2) sprintf("%.1f KB", b / 1024)
    else if (b < 1024^3) sprintf("%.1f MB", b / 1024^2)
    else                 sprintf("%.2f GB", b / 1024^3)
  }
  
  .iso_to_sec <- function(x) {
    if (is.null(x) || is.na(x) || !nzchar(x)) return(NA_real_)
    x <- trimws(as.character(x))
    if (!grepl("^P", x, ignore.case = TRUE)) return(NA_real_)
    body   <- sub("^P", "", x, ignore.case = TRUE)
    parts  <- strsplit(body, "T", fixed = TRUE)[[1L]]
    date_s <- parts[1L]
    time_s <- if (length(parts) > 1L) parts[2L] else ""
    rx <- function(s, unit) {
      m <- regmatches(s, regexpr(paste0("[0-9.]+(?=", unit, ")"), s, perl = TRUE))
      if (length(m)) as.numeric(m) else 0
    }
    out <- rx(date_s, "[Yy]") * 365.25 * 86400 +
      rx(date_s, "[Mm]") *  30.44 * 86400 +
      rx(date_s, "[Ww]") *      7 * 86400 +
      rx(date_s, "[Dd]") *          86400 +
      rx(time_s, "[Hh]") *           3600 +
      rx(time_s, "[Mm]") *             60 +
      rx(time_s, "[Ss]")
    if (out == 0) NA_real_ else out
  }
  
  ## Derive time spacing from (t_max - t_min) / (nValues - 1).
  ## This is the most reliable method: it reflects the actual number of time
  ## steps in the dataset regardless of whether the dataset is a running
  ## composite (daily steps) or a true N-day snapshot (N-day steps).
  .time_spacing_from_nvalues <- function(vname) {
    df <- info$alldata[[vname]]
    
    # nValues is in the row_type == "dimension" row
    dim_row <- df$value[df$row_type == "dimension"]
    if (!length(dim_row)) return(NA_real_)
    m <- regmatches(dim_row, regexpr("(?<=nValues=)[0-9]+", dim_row, perl = TRUE))
    if (!length(m)) return(NA_real_)
    nval <- as.integer(m)
    if (is.na(nval) || nval < 2L) return(NA_real_)
    
    # actual_range is "t_min, t_max" in seconds since epoch
    range_str <- df$value[df$attribute_name == "actual_range"]
    if (!length(range_str)) return(NA_real_)
    rng <- suppressWarnings(as.numeric(strsplit(range_str, ",\\s*")[[1L]]))
    if (length(rng) < 2L || any(is.na(rng))) return(NA_real_)
    
    span <- abs(rng[2L] - rng[1L])
    if (span == 0) return(NA_real_)
    span / (nval - 1L)
  }
  
  ## Parse ERDDAP averageSpacing string as a last resort,
  ## e.g. "nValues=890, evenlySpaced=false, averageSpacing=7 days 23h 40m 34s"
  .avg_spacing_to_sec <- function(vname) {
    df      <- info$alldata[[vname]]
    avg_str <- df$value[df$row_type == "dimension"]
    if (!length(avg_str)) return(NA_real_)
    m <- regmatches(avg_str, regexpr("averageSpacing=.+", avg_str))
    if (!length(m)) return(NA_real_)
    s    <- sub("averageSpacing=", "", m)
    pick <- function(pattern) {
      x <- regmatches(s, regexpr(pattern, s, perl = TRUE))
      if (length(x)) as.numeric(x) else 0
    }
    total <- pick("[0-9]+(?= days?)") * 86400 +
      pick("[0-9]+(?=h)")      *  3600 +
      pick("[0-9]+(?=m)")      *    60 +
      pick("[0-9]+(?=s)")
    if (total > 0) total else NA_real_
  }
  
  .stride_for <- function(dim_name) {
    s <- if (is.list(stride)) {
      stride[[dim_name]] %||% stride[["default"]] %||% 1L
    } else {
      as.integer(stride)
    }
    max(1L, as.integer(s))
  }
  
  # ── get dimension variable names ──────────────────────────────────────────
  dim_vars <- dimvars(info)
  
  if (length(dim_vars) == 0L)
    stop("No dimension variables found in info object. ",
         "Is this a griddap (not tabledap) dataset?")
  
  # ── collect and validate user dimension constraints ───────────────────────
  dimargs <- list(...)
  
  unknown <- setdiff(names(dimargs), dim_vars)
  if (length(unknown))
    warning("Dimension(s) not in dataset, ignoring: ",
            paste(unknown, collapse = ", "),
            "\n  Available: ", paste(dim_vars, collapse = ", "))
  
  dimargs <- dimargs[intersect(names(dimargs), dim_vars)]
  
  # ── detect whether a dimension is time-like ───────────────────────────────
  .is_time_dim <- function(vname) {
    if (tolower(vname) == "time") return(TRUE)
    units_val <- .attr(vname, "units")
    !is.na(units_val) && grepl("since", units_val, ignore.case = TRUE)
  }
  
  # ── resolve spacing for each dimension ───────────────────────────────────
  dim_spacing <- stats::setNames(lapply(dim_vars, function(vname) {
    
    ## 1. User override by coordinate name
    if (!is.null(spacing[[vname]]))
      return(list(value = abs(as.numeric(spacing[[vname]])), source = "user"))
    
    ## 2. Time dimension
    if (.is_time_dim(vname)) {
      if (!is.null(spacing[["time_sec"]]))
        return(list(value = as.numeric(spacing[["time_sec"]]),
                    source = "user (time_sec)"))
      
      # Primary: derive from (t_max - t_min) / (nValues - 1) — correctly
      # handles running composites with daily steps
      sec <- .time_spacing_from_nvalues(vname)
      if (!is.na(sec))
        return(list(value = sec, source = "nValues / actual_range"))
      
      # Secondary: NC_GLOBAL ISO 8601 duration
      raw <- .attr("NC_GLOBAL", c("time_coverage_resolution", "time_resolution"))
      sec <- .iso_to_sec(raw)
      if (!is.na(sec))
        return(list(value = sec, source = paste0("NC_GLOBAL (", raw, ")")))
      
      # Last resort: averageSpacing string in the nValues row
      sec <- .avg_spacing_to_sec(vname)
      if (!is.na(sec))
        return(list(value = sec, source = "dimension averageSpacing"))
      
      return(list(value = NA_real_, source = "unknown"))
    }
    
    ## 3. Latitude-like dimension
    if (grepl("^lat", tolower(vname))) {
      v <- suppressWarnings(as.numeric(.attr("NC_GLOBAL",
                                             c("geospatial_lat_resolution", "geospatial_lat_step"))))
      if (!is.na(v) && v > 0)
        return(list(value = abs(v), source = "NC_GLOBAL::geospatial_lat_resolution"))
    }
    
    ## 4. Longitude-like dimension
    if (grepl("^lon", tolower(vname))) {
      v <- suppressWarnings(as.numeric(.attr("NC_GLOBAL",
                                             c("geospatial_lon_resolution", "geospatial_lon_step"))))
      if (!is.na(v) && v > 0)
        return(list(value = abs(v), source = "NC_GLOBAL::geospatial_lon_resolution"))
    }
    
    ## 5. Generic: check coordinate variable's own attributes
    v <- suppressWarnings(as.numeric(.attr(vname,
                                           c("point_spacing", "resolution", "spacing",
                                             "geospatial_vertical_resolution"))))
    if (!is.na(v) && v > 0)
      return(list(value = abs(v), source = paste0(vname, "::point_spacing")))
    
    list(value = NA_real_, source = "unknown")
  }), dim_vars)
  
  # ── compute n_points for each constrained dimension ───────────────────────
  dim_npts <- stats::setNames(lapply(dim_vars, function(vname) {
    constraint <- dimargs[[vname]]
    if (is.null(constraint)) return(NA_integer_)
    
    if (length(constraint) != 2L)
      stop("Constraint for '", vname, "' must be c(min, max)")
    
    vals <- if (.is_time_dim(vname)) {
      as.numeric(as.POSIXct(constraint, tz = "UTC"))
    } else {
      as.numeric(constraint)
    }
    span <- abs(vals[2L] - vals[1L])
    if (span == 0) return(1L)
    
    sp <- dim_spacing[[vname]]$value
    if (is.na(sp) || sp <= 0) {
      warning("Spacing unknown for '", vname,
              "'; assuming 1 point. Use spacing = list(", vname,
              " = <value>) to override.")
      return(1L)
    }
    floor(span / sp) + 1L
  }), dim_vars)
  
  dim_npts_strided <- stats::setNames(lapply(dim_vars, function(vname) {
    n <- dim_npts[[vname]]
    if (is.na(n)) return(NA_integer_)
    ceiling(n / .stride_for(vname))
  }), dim_vars)
  
  constrained   <- dim_vars[!vapply(dim_npts, is.na, logical(1L))]
  unconstrained <- dim_vars[ vapply(dim_npts, is.na, logical(1L))]
  
  n_cells <- prod(unlist(dim_npts_strided[constrained]))
  
  # ── data variables ────────────────────────────────────────────────────────
  all_vars <- info$variables
  req_vars <- if (identical(fields, "all")) {
    all_vars
  } else {
    all_vars[all_vars$variable_name %in% fields, , drop = FALSE]
  }
  
  var_bytes <- stats::setNames(
    vapply(seq_len(nrow(req_vars)),
           function(i) as.numeric(n_cells) * .bytes(req_vars$data_type[i]),
           numeric(1L)),
    req_vars$variable_name)
  
  total_bytes <- sum(var_bytes)
  
  # ── verbose output ────────────────────────────────────────────────────────
  if (verbose) {
    dsid <- attr(info, "datasetid") %||% "(unknown)"
    sep  <- paste(rep("-", 65), collapse = "")
    cat("=== griddap request size estimate ===\n")
    cat("  Dataset  :", dsid, "\n")
    cat("  All dims :", paste(dim_vars, collapse = ", "), "\n\n")
    
    if (length(unconstrained))
      cat("  NOTE: unconstrained dim(s) not included in estimate:",
          paste(unconstrained, collapse = ", "), "\n\n")
    
    cat(sprintf("  %-20s  %10s  %10s  %s\n",
                "Dimension", "n_pts", "strided", "spacing / source"))
    cat(sep, "\n")
    for (vname in dim_vars) {
      n   <- dim_npts[[vname]]
      ns  <- dim_npts_strided[[vname]]
      sp  <- dim_spacing[[vname]]
      sp_str <- if (is.na(sp$value)) {
        paste0("unknown  [", sp$source, "]")
      } else if (.is_time_dim(vname)) {
        paste0(sp$value, " s  [", sp$source, "]")
      } else {
        paste0(sp$value, "  [", sp$source, "]")
      }
      cat(sprintf("  %-20s  %10s  %10s  %s\n",
                  vname,
                  if (is.na(n))  "(full)" else format(n,  big.mark = ","),
                  if (is.na(ns)) "(full)" else format(ns, big.mark = ","),
                  sp_str))
    }
    cat(sep, "\n")
    cat(sprintf("  Total cells (constrained) : %s\n\n",
                format(n_cells, big.mark = ",")))
    
    cat(sprintf("  %-30s  %-10s  %s\n", "Variable", "dtype", "size"))
    for (nm in names(var_bytes)) {
      dtype <- req_vars$data_type[req_vars$variable_name == nm]
      cat(sprintf("  %-30s  %-10s  %s\n", nm, dtype, .fmt(var_bytes[nm])))
    }
    cat(sep, "\n")
    cat(sprintf("  Estimated size (NetCDF3, no compression) : %s\n",
                .fmt(total_bytes)))
  }
  
  invisible(list(
    dim_vars         = dim_vars,
    constrained      = constrained,
    unconstrained    = unconstrained,
    dim_npts         = dim_npts,
    dim_npts_strided = dim_npts_strided,
    dim_spacing      = dim_spacing,
    n_cells          = n_cells,
    var_bytes        = var_bytes,
    total_bytes      = total_bytes,
    total_fmt        = .fmt(total_bytes)
  ))
}

#' ERDDAP™ server URLS and other info
#'
#' @export
#' @param ... curl options passed on to [crul::verb-GET]
#' @return data.frame with 3 columns:
#' 
#' - name (character): ERDDAP™ name
#' - url (character): ERDDAP™ url
#' - public (logical): whether it's public or not
#' 
#' @examples \dontrun{
#' servers()
#' }
servers <- function(...) {
  surl <- "https://irishmarineinstitute.github.io/awesome-erddap/erddaps.json"
  # tt <- crul::HttpClient$new(url = surl, opts = list(...))$get()
  response <- tryCatch(
    {
      tt <- crul::HttpClient$new(url = surl, opts = list(...))$get()  # Attempt to fetch
    },
    error = function(e) {
      message("Curl request failed to get server list: ", e$message)
      quit(save = "no", status = 1)  # Gracefully exit R session
    }
  )
  tt$raise_for_status()
  tibble::as_tibble(jsonlite::fromJSON(tt$parse("UTF-8")))
}

#' @title rerddap
#' @description General purpose R client for ERDDAP™ servers
#' @section ERDDAP™ info:
#' NOAA's ERDDAP™ service holds many datasets of interest. It's built on top of
#' OPenDAP. You can search for datasets via
#' [ed_search()], list datasets via [ed_datasets()],
#' get information on a single dataset via [info()], then get
#' data you want for either tabledap type via [tabledap()], or
#' for griddap type via [griddap()]
#'
#' @section tabledap/griddap:
#' tabledap and griddap have different interfaces to query for data, so
#' [tabledap()] and [griddap()] are separated out as
#' separate functions even though some of the internals are the same. In
#' particular, with tabledap you can query on/subset all variables, whereas
#' with gridddap, you can only query on/subset the dimension varibles (e.g.,
#' latitude, longitude, altitude).
#'
#' @section Data size:
#' With griddap data via [griddap()] you can get a lot of
#' data quickly. Try small searches of a dataset to start to get a sense for
#' the data, then you can increase the amount of data you get. See
#' [griddap()] for more details.
#' 
#' @section Caching:
#' \pkg{rerddap} by default caches the requests you make, so that if you happen to 
#' make the same request again, the data is restored from the cache, rather than 
#' having to go out and retrieve it remotely.  For most applications, this is good, 
#' as it can speed things up when doing a lot of request in a script, and works 
#' because in most cases an ERDDAP™ request is "idempotent".  This means that the 
#' the request will always return the same thing no matter what requests came 
#' before - it doesn't depend on state. However this is not true if the 
#' script uses either "last" in [griddap()] or "now" in [tabledap()] as these 
#' will return different values as time elapses and data are added to the 
#' datasets.  While it is desirable to have ERDDAP™ purely idempotent,  the 
#' "last" and "now" constructs are very helpful for people using ERDDAP™ 
#' in dashboards, webpages, regular input to models and the like, and the 
#' benefits far outweigh the problems.  However, if you are using either "last" 
#' or "now" in an \pkg{rerddap} based script, you want to be very careful to 
#' clear the \pkg{rerddap} cache, otherwise the request will be viewed as the 
#' same,  and the data from the last request, rather than the latest data, 
#' will be returned.
#'
#' @importFrom utils head read.csv read.delim URLencode
#' @importFrom crul HttpClient
#' @importFrom jsonlite fromJSON
#' @importFrom data.table rbindlist
#' @importFrom ncdf4 nc_open nc_close ncvar_get
#' @importFrom xml2 xml_text xml_find_all read_html xml_find_first
#' @importFrom dplyr as_data_frame bind_rows
#' @importFrom digest digest
#' @importFrom hoardr hoard
#' @docType package
#' @name rerddap
#' @keywords internal
"_PACKAGE"
## usethis namespace: start
## usethis namespace: end

#' institutions
#'
#' @docType data
#' @keywords datasets
#' @format A character vector
#' @name institutions
NULL

#' ioos_categories
#'
#' @docType data
#' @keywords datasets
#' @format A character vector
#' @name ioos_categories
NULL

#' keywords
#'
#' @docType data
#' @keywords datasets
#' @format A character vector
#' @name keywords
NULL

#' longnames
#'
#' @docType data
#' @keywords datasets
#' @format A character vector
#' @name longnames
NULL

#' standardnames
#'
#' @docType data
#' @keywords datasets
#' @format A character vector
#' @name standardnames
NULL

#' variablenames
#'
#' @docType data
#' @keywords datasets
#' @format A character vector
#' @name variablenames
NULL

#' Delete cached files
#'
#' @export
#' @param x File names
#' @param force (logical) Should files be force deleted? Default: \code{FALSE}
#' @family cache
#' @examples \dontrun{
#' # delete files by name in cache
#' # cache_delete('9911750294a039b8b517c8bf288978ea.csv')
#' # cache_delete(c('9911750294a039b8b517c8bf288978ea.csv',
#' #                'b26825b6737da13d6a52c28c8dfe690f.csv'))
#'
#' # You can delete from the output of griddap or tabledap fxns
#' ## tabledap
#' (table_res <- tabledap('erdCinpKfmBT'))
#' cache_delete(table_res)
#'
#' ## griddap
#' (out <- info('erdQMekm14day'))
#' (grid_res <- griddap(out,
#'  time = c('2015-12-28','2016-01-01'),
#'  latitude = c(24, 23),
#'  longitude = c(88, 90)
#' ))
#' cache_delete(grid_res)
#' }
cache_delete <- function(x, force = FALSE) {
  UseMethod("cache_delete")
}

#' @export
cache_delete.tabledap <- function(x, force = FALSE) {
  cdel(basename(attr(x, "path")), force)
}

#' @export
cache_delete.griddap_nc <- function(x, force = FALSE) {
  cdel(basename(attr(x, "path")), force)
}

#' @export
cache_delete.griddap_csv <- function(x, force = FALSE) {
  cdel(basename(attr(x, "path")), force)
}

#' @export
cache_delete.list <- function(x, force = FALSE) {
  cdel(unlist(x), force)
}

#' @export
cache_delete.character <- function(x, force = FALSE) {
  cdel(x, force)
}

# cache_delete helper
cdel <- function(files, force = FALSE) {
  files <- file.path(rrcache$cache_path_get(), files)
  if (!all(file.exists(files))) {
    stop("These files don't exist or can't be found: \n",
         strwrap(files[!file.exists(files)], indent = 5), call. = FALSE)
  }
  unlink(files, force = force)
}

#' @export
#' @rdname cache_delete
cache_delete_all <- function(force = FALSE) {
  files <- list.files(rrcache$cache_path_get(), full.names = TRUE)
  unlink(files, force = force)
}

#' Convert a FIPS County Code to/from a County Name
#'
#' @export
#' @param county character; A county name.
#' @param code numeric; A FIPS code.
#' @param url A URL for an ERDDAP™ server. Default:
#' https://upwell.pfeg.noaa.gov/erddap/ - See [eurl()] for 
#' more information
#' @param ... Curl options passed on to [crul::verb-GET]
#' @examples  \dontrun{
#' fipscounty(code = "06053")
#' fipscounty(county = "CA, Monterey")
#' fipscounty(county = "OR, Multnomah")
#' }

fipscounty <- function(county = NULL, code = NULL, url = eurl(), ...){
  either_or_fips(county, code)
  args <- rc(list(county = county, code = code))
  cli <- crul::HttpClient$new(url = file.path(pu(url), 'convert/fipscounty.txt'), 
    opts = list(...))
  # res <- cli$get(query = args)
  response <- tryCatch(
    {
      res <- cli$get(query = args)  # Attempt to fetch
    },
    error = function(e) {
      message("Curl request failed to get FIPS country code: ", e$message)
      quit(save = "no", status = 1)  # Gracefully exit R session
    }
  )
  res$raise_for_status()
  res$parse("UTF-8")
}

either_or_fips <- function(county, code) {
  if (!xor(!is.null(county), !is.null(code))) {
    stop("Provide either county or code, not both", call. = FALSE)
  }
}

#' Get information on an ERDDAP(TM) dataset.
#'
#' @export
#'
#' @param datasetid Dataset id
#' @param url A URL for an ERDDAP(TM) server. Default:
#' https://upwell.pfeg.noaa.gov/erddap/ - See [eurl()] for 
#' more information
#' @param ... Further args passed on to [crul::verb-GET] (must be a
#' named parameter)
#' @param x A datasetid or the output of `info`
#' @return Prints a summary of the data on return, but you can index to
#' various information.
#'
#' The data is a list of length two with:
#' 
#' - variables - Data.frame of variables and their types
#' - alldata - List of data variables and their full attributes
#' 
#' Where `alldata` element has many data.frame's, one for each variable,
#' with metadata for that variable. E.g., for griddap dataset
#' `noaa_pfeg_696e_ec99_6fa6`, `alldata` has:
#' 
#' - NC_GLOBAL
#' - time
#' - latitude
#' - longitude
#' - sss
#' 
#' @references https://upwell.pfeg.noaa.gov/erddap/index.html
#' @examples \dontrun{
#' # grid dap datasets
#' info('erdATastnhday')
#'
#' (out <- ed_search(query='temperature'))
#' info(out$info$dataset_id[5])
#' info(out$info$dataset_id[15])
#' info(out$info$dataset_id[25])
#' info(out$info$dataset_id[150])
#' info(out$info$dataset_id[400])
#' info(out$info$dataset_id[678])
#'
#' out <- info(datasetid='erdMBchla1day')
#' ## See brief overview of the variables and range of possible values, if given
#' out$variables
#' ## all information on longitude
#' out$alldata$longitude
#' ## all information on chlorophyll
#' out$alldata$chlorophyll
#'
#' # table dap datasets
#' (out <- ed_search(query='temperature', which = "table"))
#' info(out$info$dataset_id[1])
#' info(out$info$dataset_id[2])
#' info(out$info$dataset_id[3])
#' info(out$info$dataset_id[4])
#'
#' info('erdCinpKfmBT')
#' out <- info('erdCinpKfmBT')
#' ## See brief overview of the variables and range of possible values, if given
#' out$variables
#' ## all information on longitude
#' out$alldata$longitude
#' ## all information on Haliotis_corrugata_Mean_Density
#' out$alldata$Haliotis_corrugata_Mean_Density
#'
#' # use a different ERDDAP(TM) server
#' ## Marine Institute (Ireland)
#' info("IMI_CONN_2D", url = "http://erddap.marine.ie/erddap/")
#' }

info <- function(datasetid, url = eurl(), ...){
  url <- sub("/$", "", url)
  json <- erddap_GET(sprintf(file.path(url, 'info/%s/index.json'), datasetid),
                      NULL, ...)
  colnames <- vapply(tolower(json$table$columnNames),
                     function(z) gsub("\\s", "_", z), "", USE.NAMES = FALSE)
  dfs <- lapply(json$table$rows, function(x){
    tmp <- data.frame(x, stringsAsFactors = FALSE)
    names(tmp) <- colnames
    tmp
  })
  lists <- lapply(json$table$rows, stats::setNames, nm=colnames)
  names(lists) <- vapply(lists, function(b) b$variable_name, "",
                         USE.NAMES = FALSE)
  outout <- list()
  for (i in seq_along(lists)) {
    outout[[names(lists[i])]] <-
      unname(lists[ names(lists) %in% names(lists)[i] ])
  }

  df <- data.frame(rbindlist(dfs))
  vars <- df[ df$row_type == 'variable', names(df) %in%
                c('variable_name','data_type')]
  actual <- vapply(split(df, df$variable_name), function(z){
    tmp <- z[ z$attribute_name %in% 'actual_range' , "value"]
    if(length(tmp)==0) "" else tmp
  }, "")
  actualdf <- data.frame(variable_name=names(actual),
                         actual_range=unname(actual))
  vars <- merge(vars, actualdf, by="variable_name")
  oo <- lapply(outout, function(x) data.frame(rbindlist(x)))
  structure(list(variables=vars, alldata=oo, base_url=url),
            class="info",
            datasetid=datasetid,
            type=table_or_grid(datasetid, url))
}

#' @export
print.info <- function(x, ...) {
  global <- x$alldata$NC_GLOBAL
  tt <- global[ global$attribute_name %in%
                  c('time_coverage_end','time_coverage_start'), "value", ]
  dims <- x$alldata[dimvars(x)]
  vars <- x$alldata[x$variables$variable_name]
  cat(sprintf("<ERDDAP(TM) info> %s", attr(x, "datasetid")), "\n")
  cat(paste0(" Base URL: ", x$base_url), "\n")
  cat(paste0(" Dataset Type: ", attr(x, "type")), "\n")
  if(attr(x, "type") == "griddap") cat(" Dimensions (range): ", "\n")
  for(i in seq_along(dims)){
    if(names(dims[i]) == "time"){
      cat(sprintf("     time: (%s, %s)", tt[2], tt[1]), "\n")
    } else {
      cat(sprintf("     %s: (%s)", names(dims[i]), foo(dims[[i]],
                                                       "actual_range")), "\n")
    }
  }
  cat(" Variables: ", "\n")
  for(i in seq_along(vars)){
    cat(sprintf("     %s:", names(vars[i])), "\n")
    ar <- foo(vars[[i]], "actual_range")
    if(!length(ar) == 0) cat("         Range:",
                             foo(vars[[i]], "actual_range"), "\n")
    un <- foo(vars[[i]], "units")
    if(!length(un) == 0) cat("         Units:", foo(vars[[i]], "units"), "\n")
  }
}

foo <- function(x, y){
  x[ x$attribute_name == y, "value"]
}

#' @export
#' @rdname info
as.info <- function(x, url) {
  UseMethod("as.info")
}

#' @export
as.info.info <- function(x, url) {
  x
}

#' @export
as.info.character <- function(x, url) {
  info(x, url)
}

table_or_grid <- function(datasetid, url) {
  url <- sub("/$", "", url)
  table_url <- file.path(url, 'tabledap/index.json')
  tab <- toghelper(table_url)
  if (datasetid %in% tab) "tabledap" else "griddap"
}

toghelper <- function(url) {
  out <- erddap_GET(url, list(page = 1, itemsPerPage = 10000L))
  nms <- out$table$columnNames
  lists <- lapply(out$table$rows, stats::setNames, nm = nms)
  vapply(lists, "[[", "", "Dataset ID")
}

# get metadata and data -----------------------
ncdf4_get <- function(file){
  nc <- ncdf4::nc_open(file)
  on.exit(ncdf4::nc_close(nc))
  tmp <- unclass(nc)

  dims <- names(nc$dim)

  lat_name = ifelse('lat' %in% dims, 'lat', 'latitude')
  lon_name = ifelse('lon' %in% dims, 'lon', 'longitude')
  out <- list()
  for (i in seq_along(dims)) {
    out[[dims[i]]] <- ncdf4::ncvar_get(nc, nc$dim[[dims[i]]])
  }
  if ('time' %in% names(out)){
    out$time <- sapply(out$time, convert_time)
    
  }
  vars <- names(nc$var)
  outvars <- list()
  for (i in seq_along(vars)) {
    outvars[[ vars[i] ]] <- as.vector(ncdf4::ncvar_get(nc, vars[i]))
  }
  df <- do.call("cbind.data.frame", outvars)
  rows <- length(outvars[[1]])

  out <- out[length(out):1]
  meta <- expand.grid(out, stringsAsFactors = FALSE)
  # make data.frame
  alldf <- if (NROW(meta) > 0) cbind(meta, df) else df

  # output
  list(summary = tmp, data = alldf)
}

# get just metadata -----------------------
ncdf_summary <- function(file) {
  nc <- ncdf4::nc_open(file, readunlim = FALSE)
  tmp <- unclass(nc)
  on.exit(ncdf4::nc_close(nc))
  list(summary = tmp, data = data.frame(NULL))
}

#' Default ERDDAP server URL
#' 
#' @details default url is https://upwell.pfeg.noaa.gov/erddap/
#' 
#' You can set a default using an environment variable so you 
#' don't have to pass anything to the URL parameter in your 
#' function calls. 
#' 
#' In your .Renviron file or similar set a URL for the environment 
#' variable `RERDDAP_DEFAULT_URL`, like
#' `RERDDAP_DEFAULT_URL=https://upwell.pfeg.noaa.gov/erddap/`
#' 
#' It's important that you include a trailing slash in your URL
#'
#' @export
#' @examples
#' eurl()
#' Sys.setenv(RERDDAP_DEFAULT_URL = "https://google.com")
#' Sys.getenv("RERDDAP_DEFAULT_URL")
#' eurl()
#' Sys.unsetenv("RERDDAP_DEFAULT_URL")
#' eurl()
eurl <- function() {
  def_url <- "https://upwell.pfeg.noaa.gov/erddap/"
  url <- Sys.getenv("RERDDAP_DEFAULT_URL", "")
  if (identical(url, "")) return(def_url)
  url <- check_url(url)
  return(url)
}

check_url <- function(x) {
  # check if has scheme or not
  if (!grepl("https?://", x)) stop("Not a proper url")
  # add trailing slash if not present
  if (!grepl("\\/$", x)) paste0(x, "/") else x
}

#' Advanced search for ERDDAP™ tabledep or griddap datasets
#'
#' @export
#'
#' @param query (character) Search terms
#' @param page (integer) Page number. Default: 1
#' @param page_size (integer) Results per page: Default: 1000
#' @param protocol (character) One of any (default), tabledep or griddap
#' @param cdm_data_type (character) One of grid, other, point, profile,
#' timeseries, timeseriesprofile, trajectory, trajectoryprofile
#' @param institution (character) An institution. See the dataset
#' `institutions`
#' @param ioos_category (character) An ioos category See the dataset
#' `ioos_categories`
#' @param keywords (character) A keywords. See the dataset `keywords`
#' @param long_name (character) A long name. See the dataset `longnames`
#' @param standard_name (character) A standar dname. See the dataset
#' `standardnames`
#' @param variableName (character) A variable name. See the dataset
#' `variablenames`
#' @param minLat,maxLat (numeric) Minimum and maximum latitude, between -90
#' and 90
#' @param minLon,maxLon (numeric) Minimum and maximum longitude. Some datasets
#' have longitude values within -180 to 180, others use 0 to 360. If you
#' specify min and max Longitude within -180 to 180 (or 0 to 360), ERDDAP™ will
#' only find datasets that match the values you specify. Consider doing one
#' search: longitude -180 to 360, or two searches: longitude -180 to 180,
#' and 0 to 360.
#' @param minTime,maxTime (numeric/character) Minimum and maximum time. Time
#' string with the format "yyyy-MM-ddTHH:mm:ssZ, (e.g., 2009-01-21T23:00:00Z).
#' If you specify something, you must include at least yyyy-MM-dd; you can
#' omit Z, :ss, :mm, :HH, and T. Always use UTC (GMT/Zulu) time. Or specify
#' the number of seconds since 1970-01-01T00:00:00Z.
#' @param url A URL for an ERDDAP™ server. Default:
#' https://upwell.pfeg.noaa.gov/erddap/ - See [eurl()] for 
#' more information
#' @param ... Curl options passed on to [crul::verb-GET] (must be
#' named parameters)
#' @references https://upwell.pfeg.noaa.gov/erddap/index.html
#' @examples \dontrun{
#' ed_search_adv(query = 'temperature')
#' ed_search_adv(query = 'temperature', protocol = "griddap")
#' ed_search_adv(query = 'temperature', protocol = "tabledap")
#' ed_search_adv(maxLat = 63, minLon = -107, maxLon = -87, minLat = 50,
#'   protocol = "griddap")
#' ed_search_adv(maxLat = 63, minLon = -107, maxLon = -87, minLat = 50,
#'   protocol = "tabledap")
#' ed_search_adv(minTime = "2010-01-01T00:00:00Z",
#'   maxTime="2010-02-01T00:00:00Z")
#' (out <- ed_search_adv(maxLat = 63, minLon = -107, maxLon = -87, minLat = 50,
#'              minTime = "2010-01-01T00:00:00Z",
#'              maxTime="2010-02-01T00:00:00Z"))
#' out$alldata[[1]]
#' ed_search_adv(variableName = 'upwelling')
#' ed_search_adv(query = 'upwelling', protocol = "tabledap")
#'
#' # use a different URL
#' ed_search_adv(query = 'temperature', url = servers()$url[6])
#' }

ed_search_adv <- function(query = NULL, page = 1, page_size = 1000,
  protocol = NULL, cdm_data_type = NULL, institution = NULL,
  ioos_category = NULL, keywords = NULL, long_name = NULL, standard_name = NULL,
  variableName = NULL, maxLat = NULL, minLon = NULL, maxLon = NULL,
  minLat = NULL, minTime = NULL, maxTime = NULL, url = eurl(), ...) {

  check_args(query, page, page_size, protocol, cdm_data_type, institution,
             ioos_category, keywords, long_name, standard_name, variableName,
             maxLat, minLon, maxLon, minLat, minTime, maxTime)
  args <- rc(list(
    searchFor = query, page = page, itemsPerPage = page_size,
    protocol = protocol, cdm_data_type = cdm_data_type,
    institution = institution, ioos_category = ioos_category,
    keywords = keywords, long_name = long_name,
    standard_name = standard_name,
    variableName = variableName, maxLat = maxLat, minLon = minLon,
    maxLon = maxLon, minLat = minLat, minTime = minTime, maxTime = maxTime))
  json <- erddap_GET(paste0(url, 'search/advanced.json'), args, ...)
  colnames <- vapply(
    tolower(json$table$columnNames), function(z) gsub("\\s", "_", z), "",
    USE.NAMES = FALSE
  )
  dfs <- lapply(json$table$rows, function(x){
    names(x) <- colnames
    x <- x[c('title', 'dataset_id')]
    dplyr::as_tibble(x)
  })
  df <- dplyr::bind_rows(dfs)
  lists <- lapply(json$table$rows, stats::setNames, nm = colnames)
  res <- list(info = df, alldata = lists)
  structure(res, class = "ed_search_adv")
}

#' @export
print.ed_search_adv <- function(x, ...){
  cat(sprintf("%s results, showing first 20", nrow(x$info)), "\n")
  print(head(x$info, n = 20))
}

check_args <- function(query, page, page_size, protocol, cdm_data_type,
  institution, ioos_category, keywords, long_name, standard_name,
  variableName, maxLat, minLon, maxLon, minLat, minTime, maxTime) {

  check_arg(query, "character")
  check_arg(page, "numeric")
  check_arg(page_size, "numeric")
  check_arg(protocol, "character")
  check_arg(cdm_data_type, "character")
  check_arg(institution, "character")
  check_arg(ioos_category, "character")
  check_arg(keywords, "character")
  check_arg(long_name, "character")
  check_arg(standard_name, "character")
  check_arg(variableName, "character")
  check_arg(maxLat, "numeric")
  check_arg(minLon, "numeric")
  check_arg(maxLon, "numeric")
  check_arg(minLat, "numeric")
  check_arg(minTime, c("numeric", "character"))
  check_arg(maxTime, c("numeric", "character"))
}

check_arg <- function(x, class) {
  var <- deparse(substitute(x))
  if (!is.null(x)) {
    if (!class(x) %in% class) {
      if (length(class) > 1) class <- paste0(class, collapse = ", ")
      stop(var, " not of class ", class, call. = FALSE)
    }
  }
}

#' Get ERDDAP™ version
#'
#' @export
#' @param url A URL for an ERDDAP™ server. Default:
#' https://upwell.pfeg.noaa.gov/erddap/ - See [eurl()] for 
#' more information
#' @param ... Curl options passed on to [crul::verb-GET]
#' @examples \dontrun{
#' version()
#' ss <- servers()
#' version(ss$url[2])
#' version(ss$url[3])
#' }
version <- function(url = eurl(), ...){
  cli <- crul::HttpClient$new(url = file.path(pu(url), 'version'), 
    opts = list(...))
  # res <- cli$get()
  response <- tryCatch(
    {
      res <- cli$get()  # Attempt to fetch
    },
    error = function(e) {
      message("Curl request failed to get version: ", e$message)
      quit(save = "no", status = 1)  # Gracefully exit R session
    }
  )
  res$raise_for_status()
  sub("\n", "", res$parse("UTF-8"))
}

#' Browse a dataset webpage.
#'
#' @export
#'
#' @param x  datasetid or an object associated with a datasetid such
#' [info()], [griddap()] or [tabledap()]
#' @param url A URL for an ERDDAP™ server. Default:
#' https://upwell.pfeg.noaa.gov/erddap/ - See [eurl()] for 
#' more information
#' @param ... Further args passed on to `utils::browseURL`
#' (must be a named parameter)
#' @return if in interactive mode, opens a URL in your default browser; 
#' if not, then prints the URL in the console
#' @author Ben Tupper \email{btupper@@bigelow.org}
#' @examples \dontrun{
#' if (interactive()) {
#' # browse by dataset_id
#' browse('erdATastnhday')
#'
#' # browse info class
#' my_info <- info('erdATastnhday')
#' browse(my_info)
#'
#' # browse tabledap class
#' my_tabledap <- tabledap('erdCalCOFIlrvsiz', fields=c('latitude','longitude','larvae_size',
#'    'itis_tsn'), 'time>=2011-10-25', 'time<=2011-10-31')
#' browse(my_tabledap)
#' }}
browse <- function(x, url = eurl(), ...){
  UseMethod("browse", x)
}


#' @export
browse.character <- function(x, url = eurl(), ...){
  if (missing(x)) stop("datasetid is required")
  uri <- sprintf(paste0(url, 'info/%s/index.html'), x)
  
  if (interactive()) {
    utils::browseURL(uri)
  } else {
    message("URL: ", uri)
    return(uri)
  }
}

#' @export
browse.info <- function(x, url = eurl(), ...){
  datasetid <- attr(x, "datasetid")
  browse(datasetid, ...)
}

#' @export
browse.tabledap <- function(x, url = eurl(), ...){
  datasetid <- attr(x, "datasetid")
  browse(datasetid, ...)
}

#' @export
browse.griddap_nc <- function(x, url = eurl(), ...){
  datasetid <- attr(x, "datasetid")
  browse(datasetid, ...)
}

#' @export
browse.griddap_csv <- function(x, url = eurl(), ...){
  datasetid <- attr(x, "datasetid")
  browse(datasetid, ...)
}

#' @export
browse.default <- function(x, url = eurl(), ...){
  att <- attributes(x)
  if (!is.null(att)) {
    if ('datasetid' %in% names(att)) browse(att, ...)
  } else {
    stop(sprintf("browse method not implemented for %s.", class(x)[1]),
         call. = FALSE)
  }
}

#' Setup cache path
#'
#' @export
#' @param full_path (character) the full path to use for storing cached
#' files.
#' @param temp_dir (logical) if \code{TRUE} use a randomly assigned
#' \code{tempdir} (and \code{full_path} is ignored), if \code{FALSE}, you
#' can use \code{full_path}.
#' @family cache
#' @return the full cache path, a directory (character)
#' @details On opening,  by default a temporary directory is created for caching
#' files.  To have files cached elsewhere,  give the full path of where to
#' cache files.  Adding \code{temp_dir = TRUE} will again use a temporary
#' dirctory for cacheing.
#' @examples \dontrun{
#' # default path
#' cache_setup()
#'
#' # you can define your own path
#' cache_setup(path = "foobar")
#'
#' # set a tempdir - better for programming with to avoid prompt
#' cache_setup(temp_dir = TRUE)
#'
#' # cache info
#' cache_info()
#' }
cache_setup <- function(full_path = NULL, temp_dir = FALSE) {
  if (!is.null(full_path)) rrcache$cache_path_set(full_path = full_path)
  if (temp_dir) rrcache$cache_path_set("rerddap", type = "tempdir")
  rrcache$cache_path_get()
}

#' @export
#' @rdname cache_setup
cache_info <- function() {
  list(
    path = rrcache$cache_path_get(),
    no_files = length(rrcache$list())
  )
}

#' Get details of cached files
#'
#' @export
#' @param x File names
#' @family cache
#' @details Can be used to list details for all files, both .nc and .csv
#' types, or details for just individual files of class \code{tabledap},
#' \code{griddap_nc}, and \code{griddap_csv}
#' @examples \dontrun{
#' # List details for all cached files
#' cache_details()
#' }
cache_details <- function(x) {
  UseMethod("cache_details")
}

#' @export
cache_details.default <- function(x) {
  cdetails(NULL)
}

#' @export
cache_details.tabledap <- function(x) {
  cdetails(basename(attr(x, "path")))
}

#' @export
cache_details.griddap_nc <- function(x) {
  cdetails(basename(attr(x, "path")))
}

#' @export
cache_details.griddap_csv <- function(x) {
  cdetails(basename(attr(x, "path")))
}

#' @export
cache_details.list <- function(x) {
  cdetails(unlist(x))
}

#' @export
cache_details.character <- function(x) {
  cdetails(x)
}

# Helper fxn
cdetails <- function(files = NULL) {
  if (is.null(files)) {
    files <- list.files(rrcache$cache_path_get(), full.names = TRUE)
    structure(lapply(files, file_info_), class = "rerddap_cache_info")
  } else {
    files <- file.path(rrcache$cache_path_get(), files)
    structure(lapply(files, file_info_), class = "rerddap_cache_info")
  }
}

file_info_ <- function(x) {
  tmp <- strsplit(x, '\\.')[[1]]
  ext <- tmp[length(tmp)]
  fs <- file.info(x)$size
  switch(ext,
     nc = {
       list(type = "netcdf",
            size = if (!is.na(fs)) getsize(fs) else NA,
            info = if (!is.na(fs)) ncdf_summary(x)$summary else NA
       )
     },
     csv = {
       list(type = "csv",
            size = if (!is.na(fs)) getsize(fs) else NA,
            info = if (!is.na(fs))
              names(read.csv(x, nrows = 1, stringsAsFactors = FALSE)) else NA
       )
     }
  )
}

getsize <- function(x) {
  round(x/10 ^ 6, 3)
}

#' @export
print.rerddap_cache_info <- function(x, ...) {
  cat("<rerddap cached files>", sep = "\n")
  for (i in seq_along(x)) {
    cat(paste0("Type: ", x[[i]]$type), sep = "\n")
    cat(paste0("Size: ", x[[i]]$size, " mb"), sep = "\n")
    if (x[[i]]$type == "netcdf") {
      cat("info: ", sep = "\n")
      if (!any(is.na(x[[i]]$info))) {
        print(x[[i]]$info)
      }
    } else {
      cat("info: (csv columns)", sep = "\n")
      if (!any(is.na(x[[i]]$info))) {
        print(x[[i]]$info)
      }
    }
    cat("\n")
  }
}

#' Convert a UDUNITS compatible time to ISO time
#'
#' @export
#' @param n numeric; A unix time number.
#' @param isoTime character; A string time representation.
#' @param units character; Units to return. Default:
#' "seconds since 1970-01-01T00:00:00Z"
#' @param url Base URL of the ERDDAP™ server. See [eurl()] for 
#' more information
#' @param method (character) One of local or web. Local simply uses
#' [as.POSIXct()], while web method uses the ERDDAP™ time conversion service
#' `/erddap/convert/time.txt`
#' @param ... Curl options passed on to [crul::verb-GET]
#' @details When `method = "web"` time zone is GMT/UTC
#' @examples  \dontrun{
#' # local conversions
#' convert_time(n = 473472000)
#' convert_time(isoTime = "1985-01-02T00:00:00Z")
#'
#' # using an ERDDAP™ web service
#' convert_time(n = 473472000, method = "web")
#' convert_time(isoTime = "1985-01-02T00:00:00Z", method = "web")
#' }

convert_time <- function(n = NULL, isoTime = NULL,
  units = "seconds since 1970-01-01T00:00:00Z", url = eurl(),
  method = "local", ...) {

  if (!is.null(n)) stopifnot(is.numeric(n))
  if (!is.null(isoTime)) stopifnot(is.character(isoTime))
  check1notboth(n, isoTime)
  if (method == "local") {
    format(as.POSIXct(rc(list(n, isoTime))[[1]],
                      origin = "1970-01-01T00:00:00Z", tz = "UTC"),
           format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
  } else {
    args <- rc(list(n = n, isoTime = isoTime, units = units))
    cli <- crul::HttpClient$new(url = file.path(pu(url), 'convert/time.txt'), 
      opts = list(...))
    # res <- cli$get(query = args)
    response <- tryCatch(
      {
        res <- cli$get(query = args)  # Attempt to fetch
      },
      error = function(e) {
        message("Curl request failed to convert time: ", e$message)
        quit(save = "no", status = 1)  # Gracefully exit R session
      }
    )
    res$raise_for_status()
    res$parse("UTF-8")
  }
}

depsub <- function(x) {
  deparse(substitute(x, env = parent.env()))
}

check1notboth <- function(x, y) {
  if (is.null(x) && is.null(y)) {
    stop(sprintf("One of %s or %s must be non-NULL",
                 deparse(substitute(x)), deparse(substitute(y))), call. = FALSE)
  }
  if (!is.null(x) && !is.null(y)) {
    stop(sprintf("Supply only one of %s or %s",
                 deparse(substitute(x)), deparse(substitute(y))), call. = FALSE)
  }
}

## ----initialize, echo = FALSE-------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE, 
  comment = "#>", 
  cache.path = "cache/",
  fig.path = "man/figures/"
)
library(rerddap)
data("colors")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# tabledap <- function(x, ..., fields=NULL, distinct=FALSE, orderby=NULL,
#   orderbymax=NULL, orderbymin=NULL, orderbyminmax=NULL, units=NULL, fmt = 'csv',
#   url = eurl(), store = disk(), callopts=list())

## ----eval = FALSE-------------------------------------------------------------
# install.packages("rerddap")

## ----eval = FALSE-------------------------------------------------------------
# remotes::install_github("ropensci/rerddap")

## -----------------------------------------------------------------------------
library("rerddap")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# library("dplyr")
# library("ggplot2")
# library("mapdata")
# library("ncdf4")
# library("plot3D")

## ----eval = FALSE-------------------------------------------------------------
# ?rerddap

## ----eval = FALSE-------------------------------------------------------------
# whichSST <- ed_search(query = "SST")

## ----eval = FALSE-------------------------------------------------------------
# whichSST <- ed_search(query = "SST MODIS")

## -----------------------------------------------------------------------------
servers()

## ----eval = FALSE, echo = TRUE------------------------------------------------
# browse('jplMURSST41')
# browse('siocalcofiHydroCast')

## ----eval = FALSE, echo = TRUE------------------------------------------------
# info('jplMURSST41')
# info('siocalcofiHydroCast')

## -----------------------------------------------------------------------------
100*100*4*8*365

## ----eval = FALSE, echo = TRUE------------------------------------------------
# myURL <- "https://coastwatch.pfeg.noaa.gov/erddap/"
# info <- rerddap::info("erdMH1chla8day", url = myURL)
# estimate_griddap_size(info,
#   latitude  = c(30, 50),
#   longitude = c(-140, -110),
#   time      = c("2020-01-01", "2020-12-31")
#   )

## ----eval = FALSE, echo = TRUE------------------------------------------------
# === griddap request size estimate ===
#   Dataset  : erdMH1chla8day
#   All dims : time, latitude, longitude
# 
#   Dimension                  n_pts     strided  spacing / source
# -----------------------------------------------------------------
#   time                          46          46  690032.432432432 s  [nValues / actual_range]
#   latitude                     481         481  0.0416666658948831  [NC_GLOBAL::geospatial_lat_resolution]
#   longitude                    720         720  0.0416666743836092  [NC_GLOBAL::geospatial_lon_resolution]
# -----------------------------------------------------------------
#   Total cells (constrained) : 15,930,720
# 
#   Variable                        dtype       size
#   chlorophyll                     float       60.8 MB
# -----------------------------------------------------------------
#   Estimated size (NetCDF3, no compression) : 60.8 MB

## ----eval = FALSE, echo = TRUE------------------------------------------------
# latitude = c(22., 51.)
# longitude = c(-140., -105)
# time = c("2017-01-01", "2017-01-02")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# sstInfo <- info('jplMURSST41')
# murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c("2017-01-01", "2017-01-02"), fields = 'analysed_sst')
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c("2017-01-01", "2017-01-02"), stride = c(1, 1, 5), fields = 'analysed_sst')
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# 'time>=2010-01-01'
# 'time<=2010-12-31'
# 'scientific_name="Sardinops sagax"'
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# CPSinfo <- info('FRDCPSTrawlLHHaulCatch')
# sardines <- tabledap(CPSinfo, fields = c('latitude',  'longitude', 'time', 'scientific_name', 'subsample_count'), 'time>=2010-01-01', 'time<=2010-12-31', 'scientific_name="Sardinops sagax"' )
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# CPSinfo <- info('FRDCPSTrawlLHHaulCatch')
# oncorhynchus <- tabledap(CPSinfo,
#                          fields = c('latitude',  'longitude', 'time', 'scientific_name',
#                          'subsample_count'),
#                          'scientific_name=~"Oncorhynchus.*"'
#                          )
# head(oncorhynchus)
# 
# <ERDDAP™ tabledap> FRDCPSTrawlLHHaulCatch
#    Path: [/var/folders/46/jyz1mm5x5bvbf59b5g8f7f580000gn/T//RtmpLpXHMI/R/rerddap/0ce4939b104016aafebee121efeeba00.csv]
#    Last updated: [2024-10-29 11:14:10.975094]
#    File size:    [0.02 mb]
# # A tibble: 6 × 5
#   latitude longitude time                scientific_name      subsample_count
#      <dbl>     <dbl> <dttm>              <chr>                          <int>
# 1     43.0     -125. 2003-07-09 06:08:00 Oncorhynchus kisutch               1
# 2     43.0     -125. 2003-07-09 07:08:00 Oncorhynchus kisutch               1
# 3     43       -125. 2003-07-09 08:07:00 Oncorhynchus                       1
# 4     46.0     -124. 2003-07-12 08:52:00 Oncorhynchus                      10
# 5     46.0     -124. 2003-07-12 10:24:00 Oncorhynchus                      10
# 6     46.0     -124. 2003-07-12 12:09:00 Oncorhynchus                       2
# 
# 

## ----MUR, eval = FALSE, echo = TRUE-------------------------------------------
# require("ggplot2")
# require("mapdata")
# require("rerddap")
# sstInfo <- info('jplMURSST41')
# # get latest daily sst
# murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c('last','last'), fields = 'analysed_sst')
# mycolor <- colors$temperature
# w <- map_data("worldHires", ylim = c(22., 51.), xlim = c(-140, -105))
# ggplot(data = murSST$data, aes(x = longitude, y = latitude, fill = analysed_sst)) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     geom_raster(interpolate = FALSE) +
#     scale_fill_gradientn(colours = mycolor, na.value = NA) +
#     theme_bw() + ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = c(-140, -105),  ylim = c(22., 51.)) + ggtitle("Latest MUR SST")

## ----VIIRS, echo = TRUE, eval = FALSE-----------------------------------------
# require("ggplot2")
# require("mapdata")
# require("rerddap")
# sstInfo <- info('erdVHsstaWS3day', url = "https://coastwatch.pfeg.noaa.gov/erddap/")
# # get latest 3-day composite sst
# viirsSST <- griddap(sstInfo, latitude = c(41., 31.), longitude = c(-128., -115), time = c('last','last'), fields = 'sst')
# # remap latitiudes and longitudes to even grid
# myLats <- unique(viirsSST$data$latitude)
# myLons <- unique(viirsSST$data$longitude)
# myLats <- seq(range(myLats)[1], range(myLats)[2], length.out = length(myLats))
# myLons <- seq(range(myLons)[1], range(myLons)[2], length.out = length(myLons))
# # melt these out to full grid
# mapFrame <- expand.grid(x = myLons, y = myLats)
# mapFrame$y <- rev(mapFrame$y)
# # form a frame with the new values and the data
# tempFrame <- data.frame(sst = viirsSST$data$sst, lat = mapFrame$y, lon = mapFrame$x)
# mycolor <- colors$temperature
# w <- map_data("worldHires", ylim = c(30., 42.), xlim = c(-128, -114))
# ggplot(data = tempFrame, aes(x = lon, y = lat, fill = sst)) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     geom_raster(interpolate = FALSE) +
#     scale_fill_gradientn(colours = mycolor, na.value = NA) +
#     theme_bw() + ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = c(-128, -114),  ylim = c(30., 42.)) + ggtitle("Latest VIIRS 3-day SST")
# 

## ----echo = TRUE, eval = FALSE------------------------------------------------
# require("ggplot2")
# require("rerddap")
# viirsSST1 <- griddap(sstInfo, latitude = c(36., 36.),
#                      longitude = c(-126., -126.),
#                      time = c('2015-01-01','2015-12-31'), fields = 'sst')
# tempTime <- as.Date(viirsSST1$data$time, origin = '1970-01-01', tz = "GMT")
# tempFrame <- data.frame(time = tempTime, sst = viirsSST1$data$sst)

## ----echo = TRUE, eval = FALSE------------------------------------------------
# ggplot(tempFrame, aes(time, sst)) +
#   geom_line() +
#   theme_bw() +
#   ylab("sst") +
#   ggtitle("VIIRS SST at (36N, 126W)")

## ----VHNCHla, eval = FALSE, echo = TRUE---------------------------------------
# require("ggplot2")
# require("mapdata")
# require("rerddap")
# chlaInfo <- info('erdVHNchla3day')
# viirsCHLA <- griddap(chlaInfo, latitude = c(41., 31.),
#                      longitude = c(-128., -115), time = c('last','last'),
#                      fields = 'chla')

## ----eval = FALSE, echo = TRUE------------------------------------------------
# mycolor <- colors$chlorophyll
# w <- map_data("worldHires", ylim = c(30., 42.), xlim = c(-128, -114))
# ggplot(data = viirsCHLA$data, aes(x = longitude, y = latitude, fill = log(chla))) +
#   geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   geom_raster(interpolate = FALSE) +
#   scale_fill_gradientn(colours = mycolor, na.value = NA) +
#   theme_bw() + ylab("latitude") + xlab("longitude") +
#   coord_fixed(1.3, xlim = c(-128, -114),  ylim = c(30., 42.)) +
#   ggtitle("Latest VIIRS 3-day Chla")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# require("rerddap")
# dataInfo <- rerddap::info('hawaii_d90f_20ee_c4cb')
# xpos <- c(135.25, 240.25)
# ypos <- c(20.25, 60.25)
# zpos <- c(70.02, 70.02)
# tpos <- c('2010-12-15', '2010-12-15')
# soda70 <- griddap(dataInfo,  longitude = xpos, latitude = ypos,
#                   time = tpos, depth = zpos, fields = 'temp' )
# 

## ----soda70Plot, eval = FALSE, echo = TRUE------------------------------------
# require("ggplot2")
# require("mapdata")
# xlim <- c(135, 240)
# ylim <- c(20, 60)
# my.col <- colors$temperature
# ## Must do a kludge to remove countries that wrap and mess up the plot
# w1 <- map("world2Hires", xlim = c(135, 240), ylim = c(20, 60), fill = TRUE, plot = FALSE)
# remove <- c("UK:Great Britain", "France", "Spain", "Algeria", "Mali", "Burkina Faso", "Ghana", "Togo")
# w <- map_data("world2Hires", regions = w1$names[!(w1$names %in% remove)], ylim = ylim, xlim = xlim)
# myplot <- ggplot() +
#     geom_raster(data = soda70$data, aes(x = longitude, y = latitude, fill = temp), interpolate = FALSE) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(-3,30), name = "temperature") +
#     ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#     ggtitle(paste("70m temperature ", soda70$data$time[1]))
# myplot

## ----NAtlSSS, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# urlBase <- "https://erddap.marine.ie/erddap/"
# parameter <- "sea_surface_salinity"
# sssTimes <- c("last", "last")
# sssLats <- c(48.00625, 57.50625)
# sssLons <- c(-17.99375, -1.00625)
# dataInfo <- rerddap::info("IMI_NEATL", url = urlBase)
# NAtlSSS <- griddap(dataInfo, longitude = sssLons, latitude = sssLats, time = sssTimes, fields = parameter, url = urlBase)

## ----NAtlSSSplot, eval = FALSE, echo = TRUE-----------------------------------
# require("ggplot2")
# require("mapdata")
# xlim <- c(-17.99375, -1.00625)
# ylim <- c(48.00625, 57.50625)
# my.col <- colors$salinity
# w <- map_data("worldHires", ylim = ylim, xlim = xlim)
# myplot <- ggplot() +
#     geom_raster(data = NAtlSSS$data, aes(x = longitude, y = latitude, fill = sea_surface_salinity), interpolate = FALSE) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(34, 36), name = "salinity") +
#     ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#     ggtitle(paste("salinity", NAtlSSS$data$time[1]))
# myplot

## ----IFREMER, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# urlBase <- "https://www.ifremer.fr/erddap/"
# parameter <- "Salinity"
# ifrTimes <- c("2010-05-16", "2010-05-16")
# ifrLats <- c(30., 50.)
# ifrLons <- c(-140., -110.)
# ifrDepth <- c(75., 75.)
# dataInfo <- rerddap::info("SDC_GLO_CLIM_TS_V2_2", url = urlBase)
# ifrPSAL <- griddap(dataInfo, longitude = ifrLons, latitude = ifrLats, time = ifrTimes, depth = ifrDepth,  fields = parameter, url = urlBase)

## ----ifrPSALplot, eval = FALSE, echo = TRUE-----------------------------------
# ## ggplot2 has trouble with unequal y's
#  require("ggplot2")
#  require("mapdata")
#   xlim <- c(-140, -110)
#   ylim <- c(30, 51)
#   my.col <- colors$salinity
#   w <- map_data("worldHires", ylim = ylim, xlim = xlim)
#   myplot <- ggplot() +
#     geom_raster(data = ifrPSAL$data, aes(x = longitude, y = latitude, fill = Salinity), interpolate = FALSE) +
#     geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(32, 35), name = "salinity") +
#     ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) + ggtitle(paste("salinity at 75 meters",ifrPSAL$data$time[1] ))
#  myplot

## ----eval = FALSE, echo = TRUE------------------------------------------------
# require("rerddap")
# hydroInfo <- info('siocalcofiHydroCast')

## ----calCOFI, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# calcofi.df <- tabledap(hydroInfo, fields = c('cst_cnt',  'date', 'year', 'month', 'julian_date', 'julian_day', 'rpt_line', 'rpt_sta', 'cruz_num', 'intchl', 'intc14', 'time'), 'time>=1984-01-01T00:00:00Z', 'time<=2014-04-17T05:35:00Z')

## ----calCOFInum, eval = FALSE, echo = TRUE------------------------------------
# if (packageVersion('rerddap') < '1.1.0'){
#     calcofi.df$cruz_num <- as.numeric(calcofi.df$cruz_num)
#     calcofi.df$intc14 <- as.numeric(calcofi.df$intc14)
#     calcofi.df$time <- as.Date(calcofi.df$time, origin = '1970-01-01', tz = "GMT")
# }

## ----calCOFIPlot, eval = FALSE,  echo = TRUE----------------------------------
# require("dplyr")
# 
# # calculate cruise means
# by_cruznum <- group_by(calcofi.df, cruz_num)
# tempData <- select(by_cruznum, year, month, cruz_num, intchl, intc14)
# CruiseMeans <- summarize(by_cruznum, cruisechl = mean(intchl, na.rm = TRUE), cruisepp = mean(intc14, na.rm = TRUE), year = median(year, na.rm = TRUE), month = median(month, na.rm = TRUE))
# tempTimes <- paste0(CruiseMeans$year,'-',CruiseMeans$month,'-1')
# cruisetimes <- as.Date(tempTimes, origin = '1970-01-01', tz = "GMT")
# CruiseMeans$cruisetimes <- cruisetimes
# # calculate monthly "climatologies"
# byMonth <- group_by(CruiseMeans, month)
# climate <- summarize(byMonth, ppClimate = mean(cruisepp, na.rm = TRUE), chlaClimate = mean(cruisechl, na.rm = TRUE))
# # calculate anomalies
# CruiseMeans$chlanom <- CruiseMeans$cruisechl - climate$chlaClimate[CruiseMeans$month]
# CruiseMeans$ppanom <- CruiseMeans$cruisepp - climate$ppClimate[CruiseMeans$month]
# # calculate mean yearly anomaly
# byYear <- select(CruiseMeans, year)
# tempData <- select(CruiseMeans, year, chlanom, ppanom )
# byYear <- group_by(tempData, year)
# yearlyAnom <- summarize(byYear, ppYrAnom = mean(ppanom, na.rm = TRUE), chlYrAnom = mean(chlanom, na.rm = TRUE))
# yearlyAnom$year <- ISOdate(yearlyAnom$year, 01, 01, hour = 0)
# ggplot(yearlyAnom, aes(year, chlYrAnom)) + geom_line() +
#   theme_bw() + ggtitle('yearly chla anom')
# ggplot(yearlyAnom, aes(year, ppYrAnom)) + geom_line() +
#   theme_bw() + ggtitle('yearly pp anom')

## ----calCOFI-parquet, eval = FALSE, echo = TRUE-------------------------------
# require("rerddap")
# calcofi.df <- tabledap(hydroInfo, fields = c('cst_cnt',  'date', 'year', 'month', 'julian_date', 'julian_day', 'rpt_line', 'rpt_sta', 'cruz_num', 'intchl', 'intc14', 'time'), 'time>=1984-01-01T00:00:00Z', 'time<=2014-04-17T05:35:00Z', fmt = 'parquet')

## ----CPS Query, eval = FALSE, echo = TRUE-------------------------------------
# require("rerddap")
# CPSquery <- ed_search(query = 'CPS Trawl')
# CPSquery$alldata[[1]]$summary
# CPSquery$alldata[[1]]$tabledap
# CPSquery$alldata[[1]]$dataset_id

## ----CPSINfo, eval = FALSE, echo = TRUE---------------------------------------
# require("rerddap")
# CPSinfo <- info('FRDCPSTrawlLHHaulCatch')

## ----eval = FALSE, echo = TRUE------------------------------------------------
# require("dplyr")
# require("rerddap")
# sardines <- tabledap(CPSinfo, fields = c('latitude',  'longitude', 'time', 'scientific_name', 'subsample_count'), 'time>=2010-01-01', 'time<=2012-01-01', 'scientific_name="Sardinops sagax"' )
# if (packageVersion('rerddap') < '1.1.0'){
#     sardines$time <- as.Date(sardines$time, origin = '1970-01-01', tz = "GMT")
#     sardines$latitude <- as.numeric(sardines$latitude)
#     sardines$longitude <- as.numeric(sardines$longitude)
# }
# sardines$time <- as.Date(sardines$time, origin = '1970-01-01', tz = "GMT")
# sardines$latitude <- as.numeric(sardines$latitude)
# sardines$longitude <- as.numeric(sardines$longitude)
# sardine2010 <- filter(sardines, time < as.Date('2010-12-01'))

## ----eval = FALSE, echo = TRUE------------------------------------------------
# # get the dataset info
# sstInfo <- info('erdMWsstdmday')
# # get 201004 monthly sst
# sst201004 <- griddap('erdMWsstdmday', latitude = c(22., 51.), longitude = c(220., 255), time = c('2010-04-16','2010-04-16'), fields = 'sst')
# # get 201104 monthly sst
# sst201104 <- griddap('erdMWsstdmday', latitude = c(22., 51.), longitude = c(220., 255), time = c('2011-04-16','2011-04-16'), fields = 'sst')

## ----CPSPlot, eval = FALSE, echo = TRUE---------------------------------------
# # get polygons of coast for this area
# w <- map_data("worldHires", ylim = c(22., 51.), xlim = c(220 - 360, 250 - 360))
# # plot 201004 sst on the map
# sardine2010 <- filter(sardines, time < as.Date('2010-12-01', origin = '1970-01-01', tz = "GMT"))
# sardine2011 <- filter(sardines, time > as.Date('2010-12-01', origin = '1970-01-01', tz = "GMT"))
# mycolor <- colors$temperature
# p1 <- ggplot() +
#   geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   geom_raster(data = sst201004$data, aes(x = (longitude - 360), y = latitude, fill = sst), interpolate = FALSE) +
#   scale_fill_gradientn(colours = mycolor, na.value = NA, limits = c(5,30)) +
#   theme_bw() + ylab("latitude") + xlab("longitude") +
#   coord_fixed(1.3, xlim = c(220 - 360, 250 - 360),  ylim = c(22., 51.))
# 
# # plot 201104 sst on the map
# p2 <- ggplot() +
#   geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   geom_raster(data = sst201104$data, aes(x = (longitude - 360), y = latitude, fill = sst), interpolate = FALSE) +
#   geom_point(data = sardine2011, aes(x = longitude, y = latitude, colour = subsample_count)) +
#   scale_fill_gradientn(colours = mycolor, na.value = NA, limits = c(5,30)) +
#   theme_bw() + ylab("latitude") + xlab("longitude") +
#   coord_fixed(1.3, xlim = c(220 - 360, 250 - 360),  ylim = c(22., 51.))
# p1 + geom_point(data = sardine2010, aes(x = longitude, y = latitude, colour = subsample_count)) + scale_colour_gradient(space = "Lab", na.value = NA, limits = c(0,80))
# 
# p2 +   geom_point(data = sardine2011, aes(x = longitude, y = latitude, colour = subsample_count)) + scale_colour_gradient(space = "Lab", na.value = NA, limits = c(0,80))

## ----eval = FALSE,  echo = TRUE-----------------------------------------------
# sardinops <- tabledap(CPSinfo, fields = c('longitude', 'latitude', 'time'),  'scientific_name="Sardinops sagax"')
# if (packageVersion('rerddap') < '1.1.0') {
#     sardinops$time <- as.Date(sardinops$time, origin = '1970-01-01', tz = "GMT")
#     sardinops$year <- as.factor(format(sardinops$time, '%Y'))
#     sardinops$latitude <- as.numeric(sardinops$latitude)
#     sardinops$longitude <- as.numeric(sardinops$longitude)
# }
# sardinops$year <- as.factor(format(sardinops$time, '%Y'))
# 

## ----sardinesPlot, eval = FALSE, echo = TRUE----------------------------------
# xlim <- c(-135, -110)
# ylim <- c(30, 51)
# coast <- map_data("worldHires", ylim = ylim, xlim = xlim)
# ggplot() +
#     geom_point(data = sardinops, aes(x = longitude, y = latitude, colour = year)) +
#     geom_polygon(data = coast, aes(x = long, y = lat, group = group), fill = "grey80") +
#     theme_bw() + ylab("latitude") + xlab("longitude") +
#     coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#     ggtitle("Location of sardines by year in EPM Trawls")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# # get location and station ID of NDBC buoys in a region
# BuoysInfo <- info('cwwcNDBCMet')
# locationBuoys <- tabledap(BuoysInfo, distinct = TRUE, fields = c("station", "longitude", "latitude"), "longitude>=-124", "longitude<=-121", "latitude>=37", "latitude<=47")
# if (packageVersion('rerddap') < '1.1.0') {
#    locationBuoys$latitude <- as.numeric(locationBuoys$latitude)
#    locationBuoys$longitude <- as.numeric(locationBuoys$longitude)
# }

## ----NDBC, eval = FALSE,  echo = TRUE-----------------------------------------
# xlim <- c(-130, -110)
# ylim <- c(35, 50)
# coast <- map_data("worldHires", ylim = ylim, xlim = xlim)
# ggplot() +
#    geom_point(data = locationBuoys, aes(x = longitude , y = latitude, colour = factor(station) )) +
#    geom_polygon(data = coast, aes(x = long, y = lat, group = group), fill = "grey80") +
#    theme_bw() + ylab("latitude") + xlab("longitude") +
#    coord_fixed(1.3, xlim = xlim, ylim = ylim) +
#    ggtitle("Location of buoys in given region")
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# buoyData <- tabledap(BuoysInfo, fields = c("time", "wspd"), 'station="46012"', 'time>=2012-01-01', 'time<=2013-01-01')
# if (packageVersion('rerddap') < '1.1.0') {
#     buoyData$wspd <- as.numeric(buoyData$wspd)
#     buoyData$time <- as.Date(buoyData$time, origin = '1970-01-01', tz = "GMT")
# }

## ----NDBCTS, eval = FALSE, echo = TRUE----------------------------------------
# ggplot(buoyData, aes(time, wspd)) +
#   geom_line() +
#   theme_bw() +
#   ylab("wind speed") +
#   ggtitle("Wind Speed in 2012 from buoy 46012 ")

## ----eval = FALSE, echo = TRUE------------------------------------------------
# urlBase <- "https://data.ioos.us/gliders/erddap/"
# gliderInfo <- info("sp064-20161214T1913",  url = urlBase)
# glider <- tabledap(gliderInfo, fields = c("longitude", "latitude", "depth", "salinity"), 'time>=2016-12-14', 'time<=2016-12-23', url = urlBase)
# if (packageVersion('rerddap') < '1.1.0') {
#     glider$longitude <- as.numeric(glider$longitude)
#     glider$latitude <- as.numeric(glider$latitude)
#     glider$depth <- as.numeric(glider$depth)
# }

## ----glider, eval = FALSE, echo = TRUE----------------------------------------
# require("plot3D")
# scatter3D(x = glider$longitude , y = glider$latitude , z = -glider$depth, colvar = glider$salinity,              col = colors$salinity, phi = 40, theta = 25, bty = "g", type = "p",
#            ticktype = "detailed", pch = 10, clim = c(33.2,34.31), clab = 'Salinity',
#            xlab = "longitude", ylab = "latitude", zlab = "depth",
#            cex = c(0.5, 1, 1.5))

## ----echo = TRUE, eval = FALSE------------------------------------------------
# atnURL <- 'https://oceanview.pfeg.noaa.gov/erddap/'
# atnInfo <- info('gtoppAT', url = atnURL)
# atnData <- tabledap(atnInfo, fields = c("time", "longitude", "latitude"), 'toppID="1807001"', url = atnURL)
# atnData$latitude <- as.numeric(atnData$latitude)
# atnData$longitude <- as.numeric(atnData$longitude)
# ncdcSST = array(NA_real_, dim = length(atnData$time))
# ncdcSSTInfo = info('ncdcOisst2Agg')
# time_bound <- c(as.character(atnData$time[1]), as.character(atnData$time[66]))
# for (i in 1:length(atnData$time)) {
#   extract <- griddap(ncdcSSTInfo,
#                      fields = 'sst',
#                      latitude = c(atnData$latitude[i], atnData$latitude[i]),
#                      longitude = c(atnData$longitude[i], atnData$longitude[i]),
#                      time = time_bound
#                      )
# ncdcSST[i] <- extract$data$sst
# }

## ----eval = FALSE,  echo = TRUE-----------------------------------------------
# ylim <- c(32.5, 34)
# xlim <- c(-119, -116.5)
# mycolor <- colors$temperature
# w <- map_data("worldHires", ylim = ylim, xlim = xlim)
# alldata <- data.frame(sst = ncdcSST, longitude = atnData$longitude - 360, latitude = atnData$latitude)
# z <- ggplot(alldata, aes(x = longitude, y = latitude)) +
#    geom_point(aes(colour = sst), size = .5)
# z + geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
#   theme_bw() +
#   scale_colour_gradientn(colours = mycolor, limits = c(16.9, 17.3), "SST") +
#   coord_fixed(1.3, xlim = xlim, ylim = ylim) + ggtitle("SST Along Track")
# 

## ----eval = FALSE, echo = TRUE------------------------------------------------
# urlBase <- 'https://coastwatch.pfeg.noaa.gov/erddap/'
# nphInfo <- info('erdNph', url = urlBase)
# nphData <- tabledap(nphInfo, fields = c("year", "maxSLP" ), 'month=2', 'year>=1987', url = urlBase)
# nphData$maxSLP <- as.numeric(nphData$maxSLP)
# urlBase <- 'https://oceanview.pfeg.noaa.gov/erddap/'
# cohoInfo <- info('cciea_SM_CA_CO_ABND', url = urlBase)
# cohoData <- tabledap(cohoInfo, fields = c("abundance_anomaly", "time"),  url = urlBase)
# if (packageVersion('rerddap') < '1.1.0') {
#     cohoData$abundance_anomaly <- as.numeric(cohoData$abundance_anomaly)
# }
# alldata <- data.frame(coho = cohoData$abundance_anomaly[1:27], maxSLP = nphData$maxSLP, year = nphData$year)

## ----eval = FALSE, echo = TRUE------------------------------------------------
# ggplot(alldata) + geom_line(aes(x = year, y = coho), colour = 'blue') + theme_bw() + ggtitle("coho abundance anomaly")
# ggplot(alldata) + geom_line(aes(x = year, y = maxSLP), colour = 'red') + theme_bw() + ggtitle("MaxSLP")

## ----echo = TRUE, eval = FALSE------------------------------------------------
# 
# bbox_set <- function(plotobj, landmask = TRUE, xlim = NULL, ylim = NULL, interactive = FALSE)
# 

## ----echo = TRUE, eval = FALSE------------------------------------------------
# library(plotly)
# 
# myURL <- "https://coastwatch.pfeg.noaa.gov/erddap/"
# 
# sstInfo <- rerddap::info('erdVHsstaWS3day', url = myURL)
# # get latest 3-day composite sst
# viirsSST <- rerddap::griddap(sstInfo,
#                              latitude = c(41., 31.),
#                              longitude = c(-128., -115),
#                              time = c('last','last'),
#                              fields = 'sst')
# 
# w <- map("worldHires", xlim = c(-140., -114), ylim = c(30., 42.), fill = TRUE, plot = FALSE)
# my_plot <- plotdap::add_griddap(plotdap::plotdap(mapData = w), viirsSST, ~sst, fill = "thermal" )
# 
# my_plot1 <- plotdap::bbox_set(my_plot, xlim = c(-128., -115), ylim = c(31, 41.), interactive = TRUE)
# 
# # must have '"tooltip = "text"' to work properly.
# plotly::ggplotly(my_plot1, tooltip = "text")
# 

## ----echo = TRUE, eval = FALSE------------------------------------------------
# r <- plotdap:::get_raster(viirsSST, ~sst)
# r <- plotdap:::latlon_adjust(r)

## ----eval = FALSE, echo = TRUE------------------------------------------------
# ?cache_delete
# ?cache_delete_all
# ?cache_details
# ?cache_list

## -----------------------------------------------------------------------------
require("ncdf4")
exampleFile <- system.file("extdata", "MWsstd1day.nc", package = "rerddap")
sstFile <- nc_open(exampleFile)

## -----------------------------------------------------------------------------
names(sstFile$dim)

## -----------------------------------------------------------------------------
sstFile$dim$longitude$vals

## -----------------------------------------------------------------------------
names(sstFile$var)

## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(1, 1, 1, 1), count = c(1, 1, -1, -1))

## -----------------------------------------------------------------------------
latMin <- which(sstFile$dim$latitude$vals == 30.0)
latMax <- which(sstFile$dim$latitude$vals == 30.5)
lonMin <- which(sstFile$dim$longitude$vals == 210.0)
lonMax <- which(sstFile$dim$longitude$vals == 210.5)

## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMin, latMin, 1, 1), count = c( (lonMax - lonMin + 1), (latMax - latMin + 1), 1, 1 ))

## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMax, latMin, 1, 1), count = c(1, 1, 1, -1 ))

## -----------------------------------------------------------------------------
latMin <- max(which(sstFile$dim$latitude$vals <= 30.1))
latMax <- min(which(sstFile$dim$latitude$vals >= 30.3))
lonMin <- max(which(sstFile$dim$longitude$vals <= 210.1))
lonMax <- min(which(sstFile$dim$longitude$vals >= 210.3))


## -----------------------------------------------------------------------------
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMin, latMin, 1, 1), count = c( (lonMax - lonMin + 1), (latMax - latMin + 1), 1, 1 ))

#' Search for ERDDAP™ tabledep or griddap datasets
#'
#' @export
#'
#' @param query (character) Search terms
#' @param page (integer) Page number
#' @param page_size (integer) Results per page
#' @param which (character) One of tabledep or griddap.
#' @param url A URL for an ERDDAP™ server. Default:
#' https://upwell.pfeg.noaa.gov/erddap/ - See [eurl()] for 
#' more information
#' @param ... Curl options passed on to [crul::verb-GET] (must be
#' named parameters)
#' @references https://upwell.pfeg.noaa.gov/erddap/index.html
#' @examples \dontrun{
#' (out <- ed_search(query='temperature'))
#' out$alldata[[1]]
#' (out <- ed_search(query='size'))
#' out$info
#'
#' # List datasets
#' ed_datasets('table')
#' ed_datasets('grid')
#'
#' # use a different ERDDAP™ server
#' ## Marine Institute (Ireland)
#' ed_search("temperature", url = "http://erddap.marine.ie/erddap/")
#' }

ed_search <- function(query, page=NULL, page_size=NULL, which='griddap',
                      url = eurl(), ...){

  check_arg(query, "character")
  check_arg(page, "numeric")
  check_arg(page_size, "numeric")
  which <- match.arg(which, c("tabledap","griddap"), FALSE)
  args <- rc(list(searchFor = query, page = page, itemsPerPage = page_size))
  json <- erddap_GET(paste0(url, 'search/index.json'), args, ...)
  colnames <- vapply(
    tolower(json$table$columnNames), function(z) gsub("\\s", "_", z),
    "", USE.NAMES = FALSE)
  dfs <- lapply(json$table$rows, function(x){
    names(x) <- colnames
    x <- x[c('title','dataset_id')]
    data.frame(x, stringsAsFactors = FALSE)
  })
  df <- data.frame(rbindlist(dfs))
  lists <- lapply(json$table$rows, stats::setNames, nm = colnames)
  df$gd <- vapply(
    lists, function(x) {
      if (x$griddap == "") "tabledap" else "griddap"
    }, character(1))
  df <- df[ df$gd == which, -3 ]
  row.names(df) <- NULL
  res <- list(info = df, alldata = lists)
  structure(res, class = "ed_search")
}

#' @export
print.ed_search <- function(x, ...){
  print(tibble::as_tibble(x$info))
}

#' @export
#' @rdname ed_search
ed_datasets <- function(which = 'tabledap', url = eurl()){
  which <- match.arg(which, c("tabledap","griddap"), FALSE)
  url <- sprintf('%s%s/index.json', url, which)
  out <- erddap_GET(url, list(page = 1, itemsPerPage = 10000L))
  nms <- out$table$columnNames
  lists <- lapply(out$table$rows, stats::setNames, nm = nms)
  tibble::as_tibble(rbindlist(lapply(lists, data.frame,
                                         stringsAsFactors = FALSE)))
}

---
title: "Using rerddap to Access Data from ERDDAP™ Servers"
author: "Roy Mendelssohn and Scott Chamberlain"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Using rerddap to Access Data from ERDDAP™ Servers}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r initialize, echo = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE, 
  comment = "#>", 
  cache.path = "cache/",
  fig.path = "man/figures/"
)
library(rerddap)
data("colors")
```


## Introduction

`rerddap` is a general purpose <span style="color:red">R</span> client for working with <span style="color:red">ERDDAP™</span> servers.  <span style="color:red">ERDDAP™</span> is a web service developed by Bob Simons of NOAA.  At the time of this writing, there are over sixty <span style="color:red">ERDDAP™</span> servers  (though not all are public facing) providing access to literally petabytes of data and model output relevant to oceanography, meteorology, fisheries and marine mammals, among other areas. <span style="color:red">ERDDAP™</span> is a simple to use, RESTful web service, that allows data to be subsetted and returned in a variety of formats.

This vignette goes over some of the nuts and bolts of using the `rerddap` package, and shows the power of the combination of the `rerddap` package with <span style="color:red">ERDDAP™</span> servers.  Some of the examples are taken from the `xtractomatic` package (available from CRAN - https://cran.r-project.org/package=xtractomatic), and some from the `rerddapXtracto` package available on Github (https://github.com/rmendels/rerddapXtracto),  but reworked to use `rerddap` directly.   Other examples are new to this vignette, and include both gridded and non-gridded datasets from several <span style="color:red">ERDDAP™s</span>.

Some new features in `rerddap` (as of version 1.2.0) are support for downloading `tabledap()` requests as parquet files,  which can greatly reduce the size of the download (this will only work for <span style="color:red">ERDDAP™</span> server greater or equal to version 2.25),  inclusion of units in the dataframe created by `tabledap()`, and all variables in a dataframe created by `tabledap()` now have the datatype defined in the <span style="color:red">ERDDAP™</span> server,  that is for example latitude and longitude will always be numbers,  not character strings as in the past.

The option to download files as parquet files means the `tabledap()` function definition is slightly changed,  it is now:

```{r, eval = FALSE, echo = TRUE }
tabledap <- function(x, ..., fields=NULL, distinct=FALSE, orderby=NULL,
  orderbymax=NULL, orderbymin=NULL, orderbyminmax=NULL, units=NULL, fmt = 'csv',
  url = eurl(), store = disk(), callopts=list())
```

Since the "fmt" argument is optional and defaults to "csv", which was used in previous versions of `tabledap()`,  this change will not affect existing scripts.  In order to downlaod as a parquet file,  just set "fmt = 'parquet' in the `tabledap()` options.  See below for an example.



## Installation

The first step is to install the `rerddap` package, the stable version is available from CRAN:


```{r, eval = FALSE}
install.packages("rerddap")
```

or the development version can be installed from GitHub:

```{r, eval = FALSE}
remotes::install_github("ropensci/rerddap")
```

and to load the library:

```{r}
library("rerddap")
```

Besides `rerddap` the following libraries are used in this vignette:



```{r, eval = FALSE, echo = TRUE}
library("akima")
library("dplyr")
library("ggplot2")
library("mapdata")
library("ncdf4")
library("plot3D")
```

Code chunks are always given with the required libraries so that the chunks are more standalone in nature. Many of the plots use an early version of the `cmocean` colormaps designed by Kristen Thyng (see https://matplotlib.org/cmocean/ and https://github.com/matplotlib/cmocean) translatd from the original Python implementation. However, there is now a `cmocean` package for <span style="color:red">R</span>.  As past scripts may use the built-in color palette it is maintained here,  but use of the `cmocean` package is advisable as it is kept up-to-date.  In the present version of the `cmocean` package the names have changed, what is "temperature" here is now "thermal", "chlorophyll" is now "algae", and "salinity" is now "haline".

## The main `rerddap` functions

The complete list of rerddap functions can be seen by looking at he `rerddap` package help:

```{r, eval = FALSE}
?rerddap
```

and selecting the index of the package.  The main functions used here are:

* the list of servers `rerddap` knows about - `server()`
* search an <span style="color:red">ERDDAP™</span> server for terms - `ed_search(query, page = NULL, page_size = NULL, which = "griddap", url = eurl(), ...)`
* search a list of  <span style="color:red">ERDDAP™</span> servers for terms - `global_search(query, server_list, which_service)`
* get a list of datasets on an <span style="color:red">ERDDAP™</span> server - `ed_datasets(which = "tabledap", url = eurl())`
* obtain information about a dataset - `info(datasetid, url = eurl(), ...)`
* extract data from a griddap dataset - `griddap(x, ..., fields = "all", stride = 1, fmt = "nc", url = eurl(), store = disk(), read = TRUE, callopts = list())`
* extract data from a tabledap dataset - `tabledap(x, ..., fields = NULL, distinct = FALSE, orderby = NULL, orderbymax = NULL, orderbymin = NULL, orderbyminmax = NULL, units = NULL, url = eurl(), store = disk(), callopts = list())`

Be careful when using the functions `ed_search()`, `ed_datasets()` and `global_search()`.  The default <span style="color:red">ERDDAP™</span> has over 9,000 datasets,  most of which are grids, so that a list of all the gridded datasets can be quite long.  A seemly reasonable search:

```{r, eval = FALSE}
whichSST <- ed_search(query = "SST")
```

returns about 1000 responses.  The more focused query:

```{r, eval = FALSE}
whichSST <- ed_search(query = "SST MODIS")
```

still returns 172 responses.  If the simple search doesn't narrow things enough,  look at the advanced search function `ed_search_adv()`.

## Finding the Data You Want

The first way to find a dataset is to browse the builtin web page for a particular <span style="color:red">ERDDAP™</span> server.
A list of some of the public available <span style="color:red">ERDDAP™</span> servers can be obtained from the `rerddap` command:

```{r}
servers()
```

The list of <span style="color:red">ERDDAP™</span> servers is based on the list maintained at the *Awesome ERDDAP™* site compiled by the  Irish Marine Institute.

The second way to find and obtain the desired data is to use functions in `rerddap`.  The basic steps are:

1. Find the dataset on an <span style="color:red">ERDDAP™</span> server (`rerddap::servers()`, `rerddap::ed_search()`, `rerddap::ed_datasets()` ).
2. Get the needed information about the dataset (`rerddap::info()` )
3. Think about what you are going to do.
4. Make the request for the data  (`rerddap::griddap()` or `rerddap::tabledap()` ).

We discuss each of these steps in more detail, and then look at some realistic uses of the package.


### Think about what you are going to do.

This may seem to be a strange step in the process, but it is important because many of the datasets are high-resolution, and data requests can get very large, very quickly.  As an example, based on a real use case.  The MUR SST (	Multi-scale Ultra-high Resolution (MUR) SST Analysis fv04.1, see https://coastwatch.pfeg.noaa.gov/erddap/griddap/jplMURSST41.html )  is a daily, high-quality, high-resolution sea surface temperature product.  The user wanted the MUR data for a 2x2-degree box, daily for a year.  That seems innocuous enough.  Except that MURsst is at a one-hundreth of degree resolution.  If we assume just a binary representation of the data, assuming 8-bytes per value, and do the math:

```{r}
100*100*4*8*365
```


Yes, 116,800,000 bytes or roughly 115MB for that request.  Morever the user wanted the data as a .csv file, which usually makes the resulting file 8-10 times larger,  so now we are over a 1GB for the request. Even more so, there are four parameters in that dataset, and in `rerddap::griddap()` if "fields" is not specified,  all fields are downloaded, therefore the resulting files will be four times as large as given above.

So the gist of this is to think about your request before you make it.  Do a little mental math to get a rough estimate of the size of the download.  There are times receiving the data as a .csv file is convenient,  but make certain the request will not be too large.  For larger requests, obtain the data as netCDF files.  By default, `rerddap::griddap()` "melts"" the data into a dataframe,  so a .csv only provides a small convenience.  But for really large downloads,  you should select the option in `rerddap::griddap()` to not read in the data, and use instead the `netcdf4` package to read in the data, as this allows for only reading in parts of the data at a time.  [Below](#ncdf4) we provide a brief tutorial on reading in data using the `ncdf4` package.

## Some ERDDAP™ Basics

One of the main advantages of a service such as <span style="color:red">ERDDAP™</span> is that you only need to download the subset of the data you desire,  rather than the entire dataset,  which is both convenient and essential for large datasets. The underlying data model in <span style="color:red">ERDDAP™</span> is quite simple - everything is either a (multi-dimensional) grid  (think <span style="color:red">R</span> array) or a table (think a simple spreadsheet or data table). Grids are subsetted using the function `griddap()` and tables are subset using the function `tabledap()`.

If you know the datasetID of the data you are after, and are unsure if it is a grid or a table,  there are several ways to find out. We will look at two datasets, 'jplMURSST41' and 'siocalcofiHydroCasts'.  The first method is to use the `rerddap` function `browse()`

```{r, eval = FALSE, echo = TRUE}
browse('jplMURSST41')
browse('siocalcofiHydroCast')
```

which brings up information on the datasets in the browser, in the first case the "data" link is under "griddap", the second is under "tabledap".

The other method is to use the `rerddap` function `info`:

```{r, eval = FALSE, echo = TRUE}
info('jplMURSST41')
info('siocalcofiHydroCast')
```

Notice the information on 'jplMURSST41' lists the dimensions (that is a grid) while that of 'siocalcofiHydroCasts' does not (that is a table).

### Subsetting griddap()

Like an <span style="color:red">R</span> array, <span style="color:red">ERDDAP™</span> grids are subsetted by setting limits on the dimension variables, the difference being that a subset is defined in coordinate space  (latitude values, longitude values, time values) rather than array space as is done with <span style="color:red">R</span> arrays.  Thus for 'jplMURSST41' if the desired area of the data extract is latitude limits of (22N, 51N), longitude limits of (140W, 105W), and with time limits of (2017-01-01, 2017-01-02) the following would be passed to the function `griddap()`:

```{r, eval = FALSE, echo = TRUE}
latitude = c(22., 51.)
longitude = c(-140., -105)
time = c("2017-01-01", "2017-01-02")
```

A full `griddap()` request to retrieve "analysed_sst" with these constraints would be:

```{r, eval = FALSE, echo = TRUE}
sstInfo <- info('jplMURSST41')
murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c("2017-01-01", "2017-01-02"), fields = 'analysed_sst')

```

#### Strides

Strides allow to retrieve data within a coordinate bound only at every "n" values,  where "n" is an integer - think of the "by" part of the <span style="color:red">R</span> function `seq()`.  This is useful say for a monthly dataset where only the December values are desired, or you want to subsample a very high resolution dataset.  The default is a stride of 1 for all dimensions.  If you want to change the stride value in any dimension, then the value must be given for all dimensions.  So in the previous example, if only every fifth longitude is desired, the call would be:

```{r, eval = FALSE, echo = TRUE}
murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c("2017-01-01", "2017-01-02"), stride = c(1,1,5), fields = 'analysed_sst')

```

Strides are done in array space, not in coordinate space - so it is not skipping say a number of degrees of longitude, but is skipping a number of values of the array index - if longitude is thought of as an array, then every fifth value is used.  There are many cases where having strides work in coordinate space would be handy,  but it can cause a lot of problems.  Consider the case where neither the starting longitude, nor the ending longitude in the request lie on the actual data grid,  and the stride is in coordinate units in such a way that no value requested actually lies on an actual grid value. This would be equivalent to the more complicated problem of data regridding and data interpolation.

When <span style="color:red">ERDDAP™</span> receives a request where the bounds are not on the actual dataset grid,  <span style="color:red">ERDDAP™</span> finds the closest values on the grid to the requested bounds, and returns those points and all grid points in between.  If a stride is added of a value greater than one but resricted to array space, this guarantees that every value returned lies on the dataset grid.




### Subsetting tabledap()

Tables in <span style="color:red">ERDDAP™</span> are subset by using "constraint expressions" on any variable in the table.  The valid operators are =, != (not equals), =~ (a regular expression test), <, <=, >, and >=.  The constraint is constructed as the parameter on the left, value on the right, and the operator in the middle, all within a set of quotes. For example, if in the SWFSC/FRD trawl catch data (datasetID 'FRDCPSTrawlLHHaulCatch'), only sardines for 2010  were desired,  the following constraints would be set in the `tabledap()` call:

```{r, eval = FALSE, echo = TRUE}
'time>=2010-01-01'
'time<=2010-12-31'
'scientific_name="Sardinops sagax"'

```

Note that in `tabledap()` character strings usually must be passed as "double-quoted", as seen in the example with the scientific name. A full `tabledap()` request to retrieve 'latitude',  'longitude', 'time', 'scientific_name', and 'subsample_count' with these constraints would be:

```{r, eval = FALSE, echo = TRUE }
CPSinfo <- info('FRDCPSTrawlLHHaulCatch')
sardines <- tabledap(CPSinfo, fields = c('latitude',  'longitude', 'time', 'scientific_name', 'subsample_count'), 'time>=2010-01-01', 'time<=2010-12-31', 'scientific_name="Sardinops sagax"' )

```

#### tabledap() and regular expressions

As mentioned above, the 'tabledap()' function allows using a regular expression to subset the variable of interest.  While regular expressions can be used for variables that are not strings (see https://coastwatch.pfeg.noaa.gov/erddap/tabledap/documentation.html#query), they are most useful for subsetting strings, as numerical variables have a natural ordering that are better suited for the more standard constraint expressions.

As an example,  in the CPS trawl survey dataset above,  to find all the entries where the species names contains "Oncorhynchus"  (species of Pacific salmon and Pacific Trout):

```{r, eval = FALSE, echo = TRUE }
CPSinfo <- info('FRDCPSTrawlLHHaulCatch')
oncorhynchus <- tabledap(CPSinfo, 
                         fields = c('latitude',  'longitude', 'time', 'scientific_name', 
                         'subsample_count'), 
                         'scientific_name=~"Oncorhynchus.*"'
                         )
head(oncorhynchus)

<ERDDAP™ tabledap> FRDCPSTrawlLHHaulCatch
   Path: [/var/folders/46/jyz1mm5x5bvbf59b5g8f7f580000gn/T//RtmpLpXHMI/R/rerddap/0ce4939b104016aafebee121efeeba00.csv]
   Last updated: [2024-10-29 11:14:10.975094]
   File size:    [0.02 mb]
# A tibble: 6 × 5
  latitude longitude time                scientific_name      subsample_count
     <dbl>     <dbl> <dttm>              <chr>                          <int>
1     43.0     -125. 2003-07-09 06:08:00 Oncorhynchus kisutch               1
2     43.0     -125. 2003-07-09 07:08:00 Oncorhynchus kisutch               1
3     43       -125. 2003-07-09 08:07:00 Oncorhynchus                       1
4     46.0     -124. 2003-07-12 08:52:00 Oncorhynchus                      10
5     46.0     -124. 2003-07-12 10:24:00 Oncorhynchus                      10
6     46.0     -124. 2003-07-12 12:09:00 Oncorhynchus                       2


```



##  Searching

The functions in `rerddap` work with any <span style="color:red">ERDDAP™</span> server as long as the base URL is provided.  A list of advertised <span style="color:red">ERDDAP™</span> servers is provided by the Irish Marine Institute's *Awesome ERDDAP*  web page, and the `rerddap` function `servers()` will download this list.

Given a base <span style="color:red">ERDDAP™</span> URL, the function `ed_search()` will search the server for the given search terms.  The default <span style="color:red">ERDDAP™</span> server is https://upwell.pfeg.noaa.gov/erddap/.  Alternately, the function `global_search()` will search a list of <span style="color:red">ERDDAP™</span> servers for the given search terms.  The function `ed_datasets()` lists the datasets available from the given <span style="color:red">ERDDAP™</span> server.


## griddap



### MUR SST

MUR (Multi-scale Ultra-high Resolution) is an analyzed SST product at 0.01-degree resolution going back to 2002, providing one of the longest satellite based time series at such high resolution (see https://podaac.jpl.nasa.gov/dataset/MUR-JPL-L4-GLOB-v4.1). The latest data available for a region off the west coast can be extracted and plotted by:

```{r MUR, eval = FALSE, echo = TRUE}
require("ggplot2")
require("mapdata")
require("rerddap")
sstInfo <- info('jplMURSST41')
# get latest daily sst
murSST <- griddap(sstInfo, latitude = c(22., 51.), longitude = c(-140., -105), time = c('last','last'), fields = 'analysed_sst')
mycolor <- colors$temperature
w <- map_data("worldHires", ylim = c(22., 51.), xlim = c(-140, -105))
ggplot(data = murSST$data, aes(x = longitude, y = latitude, fill = analysed_sst)) +
    geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
    geom_raster(interpolate = FALSE) +
    scale_fill_gradientn(colours = mycolor, na.value = NA) +
    theme_bw() + ylab("latitude") + xlab("longitude") +
    coord_fixed(1.3, xlim = c(-140, -105),  ylim = c(22., 51.)) + ggtitle("Latest MUR SST")
```

![](MUR_sst.png)

### VIIRS SST and Chlorophyll

VIIRS (Visible Infrared Imaging Radiometer Suite)  is a scanning radiometer, that collects visible and infrared imagery and radiometric measurements of the land, atmosphere, cryosphere, and oceans. VIIRS data is used to measure cloud and aerosol properties, ocean color, sea and land surface temperature, ice motion and temperature, fires, and Earth's albedo.   Both NASA and NOAA provide VIIRS-based high resolution SST and chlorophyll products.

ERD provides a 3-day composite SST product at 750 meter resolution developed from a real-time NOAA product. The most recent values can be obtained by setting "time" to be "last".  (Note that <span style="color:red">R</span> sees the latitude-longitude grid as slightly uneven (even though it is in fact even), and that produces artificial lines in `ggplot2::geom_raster()`.  In order to remove those lines, the latitude-longitude grid is remapped to an evenly-space grid.)

```{r VIIRS, echo = TRUE, eval = FALSE}
require("ggplot2")
require("mapdata")
require("rerddap")
sstInfo <- info('erdVHsstaWS3day', url = "https://coastwatch.pfeg.noaa.gov/erddap/")
# get latest 3-day composite sst
viirsSST <- griddap(sstInfo, latitude = c(41., 31.), longitude = c(-128., -115), time = c('last','last'), fields = 'sst')
# remap latitiudes and longitudes to even grid
myLats <- unique(viirsSST$data$latitude)
myLons <- unique(viirsSST$data$longitude)
myLats <- seq(range(myLats)[1], range(myLats)[2], length.out = length(myLats))
myLons <- seq(range(myLons)[1], range(myLons)[2], length.out = length(myLons))
# melt these out to full grid
mapFrame <- expand.grid(x = myLons, y = myLats)
mapFrame$y <- rev(mapFrame$y)
# form a frame with the new values and the data
tempFrame <- data.frame(sst = viirsSST$data$sst, lat = mapFrame$y, lon = mapFrame$x)
mycolor <- colors$temperature
w <- map_data("worldHires", ylim = c(30., 42.), xlim = c(-128, -114))
ggplot(data = tempFrame, aes(x = lon, y = lat, fill = sst)) +
    geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
    geom_raster(interpolate = FALSE) +
    scale_fill_gradientn(colours = mycolor, na.value = NA) +
    theme_bw() + ylab("latitude") + xlab("longitude") +
    coord_fixed(1.3, xlim = c(-128, -114),  ylim = c(30., 42.)) + ggtitle("Latest VIIRS 3-day SST")

```

![](VIIRS_sst.png)


A time series from the same dataset at a given location,  here (36., -126.):

```{r, echo = TRUE, eval = FALSE}
require("ggplot2")
require("rerddap")
viirsSST1 <- griddap(sstInfo, latitude = c(36., 36.), 
                     longitude = c(-126., -126.), 
                     time = c('2015-01-01','2015-12-31'), fields = 'sst')
tempTime <- as.Date(viirsSST1$data$time, origin = '1970-01-01', tz = "GMT")
tempFrame <- data.frame(time = tempTime, sst = viirsSST1$data$sst)
```

```{r, echo = TRUE, eval = FALSE}
ggplot(tempFrame, aes(time, sst)) + 
  geom_line() + 
  theme_bw() + 
  ylab("sst") +
  ggtitle("VIIRS SST at (36N, 126W)")
```
![](VIIRS_sst_series.png)


A similar 3-day composite for chloropyll for the same region from a scientific quality product developed by NOAA:

```{r VHNCHla, eval = FALSE, echo = TRUE}
require("ggplot2")
require("mapdata")
require("rerddap")
chlaInfo <- info('erdVHNchla3day')
viirsCHLA <- griddap(chlaInfo, latitude = c(41., 31.), 
                     longitude = c(-128., -115), time = c('last','last'), 
                     fields = 'chla')
```

```{r, eval = FALSE, echo = TRUE}
mycolor <- colors$chlorophyll
w <- map_data("worldHires", ylim = c(30., 42.), xlim = c(-128, -114))
ggplot(data = viirsCHLA$data, aes(x = longitude, y = latitude, fill = log(chla))) +
  geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
  geom_raster(interpolate = FALSE) +
  scale_fill_gradientn(colours = mycolor, na.value = NA) +
  theme_bw() + ylab("latitude") + xlab("longitude") +
  coord_fixed(1.3, xlim = c(-128, -114),  ylim = c(30., 42.)) + 
  ggtitle("Latest VIIRS 3-day Chla")
```

![](VIIRS_chla.png)


### Temperature at 70m in the north Pacific from the SODA model output

This is an example of an extract from a 4-D dataset (results from the "Simple Ocean Data Assimilation (SODA)" model - see https://soda.umd.edu), and illustrate the case where the z-coordinate does not have the default name "altitude".  Water temperature at 70m depth is extracted for the North Pacific Ocean.


```{r eval = FALSE, echo = TRUE}
require("rerddap")
dataInfo <- rerddap::info('hawaii_d90f_20ee_c4cb')
xpos <- c(135.25, 240.25)
ypos <- c(20.25, 60.25)
zpos <- c(70.02, 70.02)
tpos <- c('2010-12-15', '2010-12-15')
soda70 <- griddap(dataInfo,  longitude = xpos, latitude = ypos, 
                  time = tpos, depth = zpos, fields = 'temp' )

```

Since the data cross the dateline, it is necessary to use the new "world2Hires" continental outlines in the package `mapdata` which is Pacific Ocean centered.  Unfortunatley there is a small problem where the outlines from certain countries wrap and mistakenly appear in plots, and those countries must be removed,  see code below.


```{r soda70Plot, eval = FALSE, echo = TRUE}
require("ggplot2")
require("mapdata")
xlim <- c(135, 240)
ylim <- c(20, 60)
my.col <- colors$temperature
## Must do a kludge to remove countries that wrap and mess up the plot
w1 <- map("world2Hires", xlim = c(135, 240), ylim = c(20, 60), fill = TRUE, plot = FALSE)
remove <- c("UK:Great Britain", "France", "Spain", "Algeria", "Mali", "Burkina Faso", "Ghana", "Togo")
w <- map_data("world2Hires", regions = w1$names[!(w1$names %in% remove)], ylim = ylim, xlim = xlim)
myplot <- ggplot() +
    geom_raster(data = soda70$data, aes(x = longitude, y = latitude, fill = temp), interpolate = FALSE) +
    geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
    theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(-3,30), name = "temperature") +
    ylab("latitude") + xlab("longitude") +
    coord_fixed(1.3, xlim = xlim, ylim = ylim) +
    ggtitle(paste("70m temperature ", soda70$data$time[1]))
myplot
```

![](soda70.png)

### Irish Marine Institute {#hourly}

The Irish Marine Institute has an <span style="color:red">ERDDAP™</span> server at https://erddap.marine.ie/erddap.  Among other datasets, they have hourly output from a model of the North Altantic ocean, with a variety of ocean related parameters, see the dataset *IMI_NEATL*.  To obtain the latest sea surface salinity for the domain of the model:

```{r NAtlSSS, eval = FALSE, echo = TRUE}
require("rerddap")
urlBase <- "https://erddap.marine.ie/erddap/"
parameter <- "sea_surface_salinity"
sssTimes <- c("last", "last")
sssLats <- c(48.00625, 57.50625)
sssLons <- c(-17.99375, -1.00625)
dataInfo <- rerddap::info("IMI_NEATL", url = urlBase)
NAtlSSS <- griddap(dataInfo, longitude = sssLons, latitude = sssLats, time = sssTimes, fields = parameter, url = urlBase)
```


and the extracted data plotted:

```{r NAtlSSSplot, eval = FALSE, echo = TRUE}
require("ggplot2")
require("mapdata")
xlim <- c(-17.99375, -1.00625)
ylim <- c(48.00625, 57.50625)
my.col <- colors$salinity
w <- map_data("worldHires", ylim = ylim, xlim = xlim)
myplot <- ggplot() +
    geom_raster(data = NAtlSSS$data, aes(x = longitude, y = latitude, fill = sea_surface_salinity), interpolate = FALSE) +
    geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
    theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(34, 36), name = "salinity") +
    ylab("latitude") + xlab("longitude") +
    coord_fixed(1.3, xlim = xlim, ylim = ylim) +
    ggtitle(paste("salinity", NAtlSSS$data$time[1]))
myplot
```

![](NAtlSSS.png)

### IFREMER

The French agency IFREMER also has an <span style="color:red">ERDDAP™</span> server. Here salinity climatologies at 75 meters  off the west coast of the United States, calculated from the World Ocean Database 2013 using the program DIVA, is extracted and plotted.

```{r IFREMER, eval = FALSE, echo = TRUE}
require("rerddap")
urlBase <- "https://www.ifremer.fr/erddap/"
parameter <- "Salinity"
ifrTimes <- c("2010-05-16", "2010-05-16")
ifrLats <- c(30., 50.)
ifrLons <- c(-140., -110.)
ifrDepth <- c(75., 75.)
dataInfo <- rerddap::info("SDC_GLO_CLIM_TS_V2_2", url = urlBase)
ifrPSAL <- griddap(dataInfo, longitude = ifrLons, latitude = ifrLats, time = ifrTimes, depth = ifrDepth,  fields = parameter, url = urlBase)
```

The `ggplot2` function `geom_raster()` is not designed for unevenly spaced coordinates, as are the latitudes from this model.  The function `interp()` from the package `akima` is used to interpolate the data which are then plotted.


```{r ifrPSALplot, eval = FALSE, echo = TRUE}
## ggplot2 has trouble with unequal y's
 require("akima")
 require("dplyr")
 require("ggplot2")
 require("mapdata")
  xlim <- c(-140, -110)
  ylim <- c(30, 51)
  my.col <- colors$salinity
  myplot <- ggplot() +
    geom_raster(data = ifrPSAL$data, aes(x = longitude, y = latitude, fill = Salinity), interpolate = FALSE) +
    geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
    theme_bw() + scale_fill_gradientn(colours = my.col, na.value = NA, limits = c(32, 35), name = "salinity") +
    ylab("latitude") + xlab("longitude") +
    coord_fixed(1.3, xlim = xlim, ylim = ylim) + ggtitle(paste("salinity at 75 meters",ifrPSAL$data$time[1] ))
 myplot
```

![](ifrPSA.png)

## tabledap


### CalCOFI data

CalCOFI (California Cooperative Oceanic Fisheries Investigations - https://calcofi.org) is a multi-agency partnership formed in 1949 to investigate the collapse of the sardine population off California. The organization's members are from NOAA Fisheries Service, Scripps Institution of Oceanography, and California Department of Fish and Wildlife. The scope of this research has evolved into the study of marine ecosystems off California and the management of its fisheries resources.  The nearly complete CalCOFI data, both physical and biological, are available through <span style="color:red">ERDDAP™</span>.

The following example is a modification of a script developed by Dr. Andrew Leising of the Southwest Fisheries Science Center.  The original script has been used to automate the generation of several yearly reports about the California Current Ecosystem.   The script gets chlorophyll data and a measure of primary productivity from the hydrocasts,and then calculates a seasoanlly adjusted chlorophyll anomaly as well as a seasonally adjusted primary productivity anomaly.  The first step is to get the information about the particular dataset:

```{r, eval = FALSE, echo = TRUE}
require("rerddap")
hydroInfo <- info('siocalcofiHydroCast')
```

and then get the desired data from 1984 through 2014:

```{r calCOFI, eval = FALSE, echo = TRUE}
require("rerddap")
calcofi.df <- tabledap(hydroInfo, fields = c('cst_cnt',  'date', 'year', 'month', 'julian_date', 'julian_day', 'rpt_line', 'rpt_sta', 'cruz_num', 'intchl', 'intc14', 'time'), 'time>=1984-01-01T00:00:00Z', 'time<=2014-04-17T05:35:00Z')
```


In versions of 'rerddap' before v1.1.0,  "intchl" and "intC14" are returned as character strings, and are easier to work with as numbers, so in earlier versions of you need to:

```{r calCOFInum, eval = FALSE, echo = TRUE}
if (packageVersion('rerddap') < '1.1.0'){
    calcofi.df$cruz_num <- as.numeric(calcofi.df$cruz_num)
    calcofi.df$intc14 <- as.numeric(calcofi.df$intc14)
    calcofi.df$time <- as.Date(calcofi.df$time, origin = '1970-01-01', tz = "GMT")
}
```


At this point the requested data are in the <span style="color:red">R</span> workspace - the rest of the code performs the calculations to derive the seasonally adjusted values and plot them.

```{r calCOFIPlot, eval = FALSE,  echo = TRUE}
require("dplyr")

# calculate cruise means
by_cruznum <- group_by(calcofi.df, cruz_num)
tempData <- select(by_cruznum, year, month, cruz_num, intchl, intc14)
CruiseMeans <- summarize(by_cruznum, cruisechl = mean(intchl, na.rm = TRUE), cruisepp = mean(intc14, na.rm = TRUE), year = median(year, na.rm = TRUE), month = median(month, na.rm = TRUE))
tempTimes <- paste0(CruiseMeans$year,'-',CruiseMeans$month,'-1')
cruisetimes <- as.Date(tempTimes, origin = '1970-01-01', tz = "GMT")
CruiseMeans$cruisetimes <- cruisetimes
# calculate monthly "climatologies"
byMonth <- group_by(CruiseMeans, month)
climate <- summarize(byMonth, ppClimate = mean(cruisepp, na.rm = TRUE), chlaClimate = mean(cruisechl, na.rm = TRUE))
# calculate anomalies
CruiseMeans$chlanom <- CruiseMeans$cruisechl - climate$chlaClimate[CruiseMeans$month]
CruiseMeans$ppanom <- CruiseMeans$cruisepp - climate$ppClimate[CruiseMeans$month]
# calculate mean yearly anomaly
byYear <- select(CruiseMeans, year)
tempData <- select(CruiseMeans, year, chlanom, ppanom )
byYear <- group_by(tempData, year)
yearlyAnom <- summarize(byYear, ppYrAnom = mean(ppanom, na.rm = TRUE), chlYrAnom = mean(chlanom, na.rm = TRUE))
yearlyAnom$year <- ISOdate(yearlyAnom$year, 01, 01, hour = 0)
ggplot(yearlyAnom, aes(year, chlYrAnom)) + geom_line() +
  theme_bw() + ggtitle('yearly chla anom')
ggplot(yearlyAnom, aes(year, ppYrAnom)) + geom_line() +
  theme_bw() + ggtitle('yearly pp anom')
```

![](calCOFIPlot-1.png)
![](calCOFIPlot-2.png)

### Parquet files

A new feature in `tabledap()` is to choose to download the data as a parquet file rather than as a csv file - the parquet file can be up to 90% smaller in our tests  (note this will only work if the <span style="color:red">ERDDAP™</span> server is at least version 2.25,  otherwise the program will throw an error).  As any example, for the CalCOFI example above:


```{r calCOFI-parquet, eval = FALSE, echo = TRUE}
require("rerddap")
calcofi.df <- tabledap(hydroInfo, fields = c('cst_cnt',  'date', 'year', 'month', 'julian_date', 'julian_day', 'rpt_line', 'rpt_sta', 'cruz_num', 'intchl', 'intc14', 'time'), 'time>=1984-01-01T00:00:00Z', 'time<=2014-04-17T05:35:00Z', fmt = 'parquet')
```




### CPS Trawl Surveys


The CPS (Coastal Pelagic Species) Trawl Life History Length Frequency Data contains the length distribution of a subset of individuals from a species (mainly non-target) caught during SWFSC-FRD fishery independent trawl surveys of coastal pelagic species. Measured lengths for indicated length type (fork, standard, total, or mantle) were grouped in 10 mm bins (identified by the midpoint of the length class) and counts are recorded by sex.

The number and location of sardines (Sardinops sagax) in the tows in March 2010 and 2011 are extracted, and compared with monthly SST from satellites.  First, query the <span style="color:red">ERDDAP™</span> server to see if CPS Trawl data are available through the <span style="color:red">ERDDAP™</span> server, and if so, obtain the datasetID for the dataset.

```{r CPS Query, eval = FALSE, echo = TRUE}
require("rerddap")
CPSquery <- ed_search(query = 'CPS Trawl')
CPSquery$alldata[[1]]$summary
CPSquery$alldata[[1]]$tabledap
CPSquery$alldata[[1]]$dataset_id
```

Then get the information for the CPS dataset:

```{r CPSINfo, eval = FALSE, echo = TRUE}
require("rerddap")
CPSinfo <- info('FRDCPSTrawlLHHaulCatch')
```

extract the desired CPS data:

```{r, eval = FALSE, echo = TRUE}
require("dplyr")
require("rerddap")
sardines <- tabledap(CPSinfo, fields = c('latitude',  'longitude', 'time', 'scientific_name', 'subsample_count'), 'time>=2010-01-01', 'time<=2012-01-01', 'scientific_name="Sardinops sagax"' )
if (packageVersion('rerddap') < '1.1.0'){
    sardines$time <- as.Date(sardines$time, origin = '1970-01-01', tz = "GMT")
    sardines$latitude <- as.numeric(sardines$latitude)
    sardines$longitude <- as.numeric(sardines$longitude)
}
sardines$time <- as.Date(sardines$time, origin = '1970-01-01', tz = "GMT")
sardines$latitude <- as.numeric(sardines$latitude)
sardines$longitude <- as.numeric(sardines$longitude)
sardine2010 <- filter(sardines, time < as.Date('2010-12-01'))
```

and plot the data versus monthly SST values:

```{r, eval = FALSE, echo = TRUE}
# get the dataset info
sstInfo <- info('erdMWsstdmday')
# get 201004 monthly sst
sst201004 <- griddap('erdMWsstdmday', latitude = c(22., 51.), longitude = c(220., 255), time = c('2010-04-16','2010-04-16'), fields = 'sst')
# get 201104 monthly sst
sst201104 <- griddap('erdMWsstdmday', latitude = c(22., 51.), longitude = c(220., 255), time = c('2011-04-16','2011-04-16'), fields = 'sst')
```

```{r CPSPlot, eval = FALSE, echo = TRUE}
# get polygons of coast for this area
w <- map_data("worldHires", ylim = c(22., 51.), xlim = c(220 - 360, 250 - 360))
# plot 201004 sst on the map
sardine2010 <- filter(sardines, time < as.Date('2010-12-01', origin = '1970-01-01', tz = "GMT"))
sardine2011 <- filter(sardines, time > as.Date('2010-12-01', origin = '1970-01-01', tz = "GMT"))
mycolor <- colors$temperature
p1 <- ggplot() +
  geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
  geom_raster(data = sst201004$data, aes(x = (longitude - 360), y = latitude, fill = sst), interpolate = FALSE) +
  scale_fill_gradientn(colours = mycolor, na.value = NA, limits = c(5,30)) +
  theme_bw() + ylab("latitude") + xlab("longitude") +
  coord_fixed(1.3, xlim = c(220 - 360, 250 - 360),  ylim = c(22., 51.))

# plot 201104 sst on the map
p2 <- ggplot() +
  geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
  geom_raster(data = sst201104$data, aes(x = (longitude - 360), y = latitude, fill = sst), interpolate = FALSE) +
  geom_point(data = sardine2011, aes(x = longitude, y = latitude, colour = subsample_count)) +
  scale_fill_gradientn(colours = mycolor, na.value = NA, limits = c(5,30)) +
  theme_bw() + ylab("latitude") + xlab("longitude") +
  coord_fixed(1.3, xlim = c(220 - 360, 250 - 360),  ylim = c(22., 51.))
p1 + geom_point(data = sardine2010, aes(x = longitude, y = latitude, colour = subsample_count)) + scale_colour_gradient(space = "Lab", na.value = NA, limits = c(0,80))

p2 +   geom_point(data = sardine2011, aes(x = longitude, y = latitude, colour = subsample_count)) + scale_colour_gradient(space = "Lab", na.value = NA, limits = c(0,80))
```

![](CPSPlot-1.png)
![](CPSPlot-2.png)

Also of interest is the distribution of sardines through the years:

```{r, eval = FALSE,  echo = TRUE}
sardinops <- tabledap(CPSinfo, fields = c('longitude', 'latitude', 'time'),  'scientific_name="Sardinops sagax"')
if (packageVersion('rerddap') < '1.1.0') {
    sardinops$time <- as.Date(sardinops$time, origin = '1970-01-01', tz = "GMT")
    sardinops$year <- as.factor(format(sardinops$time, '%Y'))
    sardinops$latitude <- as.numeric(sardinops$latitude)
    sardinops$longitude <- as.numeric(sardinops$longitude)
}
sardinops$year <- as.factor(format(sardinops$time, '%Y'))

```

```{r sardinesPlot, eval = FALSE, echo = TRUE}
xlim <- c(-135, -110)
ylim <- c(30, 51)
coast <- map_data("worldHires", ylim = ylim, xlim = xlim)
ggplot() +
    geom_point(data = sardinops, aes(x = longitude, y = latitude, colour = year)) +
    geom_polygon(data = coast, aes(x = long, y = lat, group = group), fill = "grey80") +
    theme_bw() + ylab("latitude") + xlab("longitude") +
    coord_fixed(1.3, xlim = xlim, ylim = ylim) +
    ggtitle("Location of sardines by year in EPM Trawls")
```

![](sardines.png)


### NDBC Buoys

NOAA's National Data Buoy Center (NDBC) collects world-wide data from buoys in the ocean. <span style="color:red">ERDDAP™</span> can be searched for the location of all buoys in a bounding box with latitudes(37N, 47N) and longitudes (124W, 121W) and the results plotted:

```{r, eval = FALSE, echo = TRUE}
# get location and station ID of NDBC buoys in a region
BuoysInfo <- info('cwwcNDBCMet')
locationBuoys <- tabledap(BuoysInfo, distinct = TRUE, fields = c("station", "longitude", "latitude"), "longitude>=-124", "longitude<=-121", "latitude>=37", "latitude<=47")
if (packageVersion('rerddap') < '1.1.0') {
   locationBuoys$latitude <- as.numeric(locationBuoys$latitude)
   locationBuoys$longitude <- as.numeric(locationBuoys$longitude)
}
```

```{r NDBC, eval = FALSE,  echo = TRUE}
xlim <- c(-130, -110)
ylim <- c(35, 50)
coast <- map_data("worldHires", ylim = ylim, xlim = xlim)
ggplot() +
   geom_point(data = locationBuoys, aes(x = longitude , y = latitude, colour = factor(station) )) +
   geom_polygon(data = coast, aes(x = long, y = lat, group = group), fill = "grey80") +
   theme_bw() + ylab("latitude") + xlab("longitude") +
   coord_fixed(1.3, xlim = xlim, ylim = ylim) +
   ggtitle("Location of buoys in given region")

```

![](NDBC.png)

Looking at wind speed for 2012 for station "46012"

```{r, eval = FALSE, echo = TRUE}
buoyData <- tabledap(BuoysInfo, fields = c("time", "wspd"), 'station="46012"', 'time>=2012-01-01', 'time<=2013-01-01')
if (packageVersion('rerddap') < '1.1.0') {
    buoyData$wspd <- as.numeric(buoyData$wspd)
    buoyData$time <- as.Date(buoyData$time, origin = '1970-01-01', tz = "GMT")
}
```

```{r NDBCTS, eval = FALSE, echo = TRUE}
ggplot(buoyData, aes(time, wspd)) + 
  geom_line() + 
  theme_bw() + 
  ylab("wind speed") +
  ggtitle("Wind Speed in 2012 from buoy 46012 ")
```
![](NDBCTS.png)

###  IOOS Glider Data

The mission of the IOOS Glider DAC is to provide glider operators with a simple process for submitting glider data sets to a centralized location, enabling the data to be visualized, analyzed, widely distributed via existing web services and the Global Telecommunications System (GTS) and archived at the National Centers for Environmental Information (NCEI).
The IOOS Glider Dac is accessible through `rerddap` (https://gliders.ioos.us/erddap/index.html).  Extracting and plotting salinity from part of the path of one glider deployed by the Scripps Institution of Oceanography:

```{r, eval = FALSE, echo = TRUE}
urlBase <- "https://data.ioos.us/gliders/erddap/"
gliderInfo <- info("sp064-20161214T1913",  url = urlBase)
glider <- tabledap(gliderInfo, fields = c("longitude", "latitude", "depth", "salinity"), 'time>=2016-12-14', 'time<=2016-12-23', url = urlBase)
if (packageVersion('rerddap') < '1.1.0') {
    glider$longitude <- as.numeric(glider$longitude)
    glider$latitude <- as.numeric(glider$latitude)
    glider$depth <- as.numeric(glider$depth)
}
```

```{r glider, eval = FALSE, echo = TRUE}
require("plot3D")
scatter3D(x = glider$longitude , y = glider$latitude , z = -glider$depth, colvar = glider$salinity,              col = colors$salinity, phi = 40, theta = 25, bty = "g", type = "p",
           ticktype = "detailed", pch = 10, clim = c(33.2,34.31), clab = 'Salinity',
           xlab = "longitude", ylab = "latitude", zlab = "depth",
           cex = c(0.5, 1, 1.5))
```

![](glider.png)

### Animal Telemetry Network (ATN)

The Integrated Ocean Observing System Animal Telemetry Network (IOOS ATN) is designed to serve as an access point to search, discover and access animal telemetry data, and associated oceanographic datasets, from a wide variety of species and platforms.  The track of one of the tagged animals is extracted, and plotted with SST value from satellite along the track.


```{r, echo = TRUE, eval = FALSE}
atnURL <- 'https://oceanview.pfeg.noaa.gov/erddap/'
atnInfo <- info('gtoppAT', url = atnURL)
atnData <- tabledap(atnInfo, fields = c("time", "longitude", "latitude"), 'toppID="1807001"', url = atnURL)
atnData$latitude <- as.numeric(atnData$latitude)
atnData$longitude <- as.numeric(atnData$longitude)
ncdcSST = array(NA_real_, dim = length(atnData$time))
ncdcSSTInfo = info('ncdcOisst2Agg')
time_bound <- c(as.character(atnData$time[1]), as.character(atnData$time[66]))
for (i in 1:length(atnData$time)) {
  extract <- griddap(ncdcSSTInfo, 
                     fields = 'sst', 
                     latitude = c(atnData$latitude[i], atnData$latitude[i]), 
                     longitude = c(atnData$longitude[i], atnData$longitude[i]), 
                     time = time_bound
                     )
ncdcSST[i] <- extract$data$sst
}
```

```{r, eval = FALSE,  echo = TRUE}
ylim <- c(32.5, 34)
xlim <- c(-119, -116.5)
mycolor <- colors$temperature
w <- map_data("worldHires", ylim = ylim, xlim = xlim)
alldata <- data.frame(sst = ncdcSST, longitude = atnData$longitude - 360, latitude = atnData$latitude)
z <- ggplot(alldata, aes(x = longitude, y = latitude)) +
   geom_point(aes(colour = sst), size = .5)
z + geom_polygon(data = w, aes(x = long, y = lat, group = group), fill = "grey80") +
  theme_bw() +
  scale_colour_gradientn(colours = mycolor, limits = c(16.9, 17.3), "SST") +
  coord_fixed(1.3, xlim = xlim, ylim = ylim) + ggtitle("SST Along Track")

```

![](atn.png)

### California Current System Integrated Ecosystem Assessment (CCSIEA)

The primary goals of the CCIEA are to better understand the web of interactions that drive patterns and trends of components within the California Current ecosystem, and forecast how changing environmental conditions and management actions affect the status of these components. The conceptual model of the social-ecological system of the California Current illustrates how humans and their social systems are inextricably linked to these marine, coastal, and upland environments (see \code("www.integratedecosystemassessment.noaa.gov/regions/california-current/projects").

The over 300 indices developed for the CCSIEA are available through `rerddap`.  Here an index of coho abundance in California is compared with the February value of an index of the strength and location of the North Pacific High, developed in "The North Pacific High and wintertime pre-conditioning of California current productivity" by Schroeder et al.  (Geophys. Res. Lett., 40, 541–546).

```{r, eval = FALSE, echo = TRUE}
urlBase <- 'https://coastwatch.pfeg.noaa.gov/erddap/'
nphInfo <- info('erdNph', url = urlBase)
nphData <- tabledap(nphInfo, fields = c("year", "maxSLP" ), 'month=2', 'year>=1987', url = urlBase)
nphData$maxSLP <- as.numeric(nphData$maxSLP)
urlBase <- 'https://oceanview.pfeg.noaa.gov/erddap/'
cohoInfo <- info('cciea_SM_CA_CO_ABND', url = urlBase)
cohoData <- tabledap(cohoInfo, fields = c("abundance_anomaly", "time"),  url = urlBase)
if (packageVersion('rerddap') < '1.1.0') {
    cohoData$abundance_anomaly <- as.numeric(cohoData$abundance_anomaly)
}
alldata <- data.frame(coho = cohoData$abundance_anomaly[1:27], maxSLP = nphData$maxSLP, year = nphData$year)
```

```{r, eval = FALSE, echo = TRUE}
ggplot(alldata) + geom_line(aes(x = year, y = coho), colour = 'blue') + theme_bw() + ggtitle("coho abundance anomaly")
ggplot(alldata) + geom_line(aes(x = year, y = maxSLP), colour = 'red') + theme_bw() + ggtitle("MaxSLP")
```

![](coho_anomaly.png)
![](maxslp.png)

## Interactive plots

The `plotdap` plotting package has a new version of the function "box_set()" which can both reset the limits of the plot as well as return a
`ggplot2` object that works with the `plotly` function "ggplotly()" to produce an intertactive plot.

"box_set()" has the form:

```{r, echo = TRUE, eval = FALSE}

bbox_set <- function(plotobj, landmask = TRUE, xlim = NULL, ylim = NULL, interactive = FALSE)

```

If the values for "xlim" and "ylim" are not given,  then the function will determine the limits and projection of the `plotdap` object,  and set this in the `ggplot2` part of the the `plotdap` object.  If "interactive" is set to "TRUE",  then a `ggplot2` object is returned that works properly with the function `plotly::ggplotly()`.

```{r, echo = TRUE, eval = FALSE}
library(plotly)

myURL <- "https://coastwatch.pfeg.noaa.gov/erddap/"

sstInfo <- rerddap::info('erdVHsstaWS3day', url = myURL)
# get latest 3-day composite sst
viirsSST <- rerddap::griddap(sstInfo, 
                             latitude = c(41., 31.), 
                             longitude = c(-128., -115), 
                             time = c('last','last'), 
                             fields = 'sst')

w <- map("worldHires", xlim = c(-140., -114), ylim = c(30., 42.), fill = TRUE, plot = FALSE)
my_plot <- plotdap::add_griddap(plotdap::plotdap(mapData = w), viirsSST, ~sst, fill = "thermal" )

my_plot1 <- plotdap::bbox_set(my_plot, xlim = c(-128., -115), ylim = c(31, 41.), interactive = TRUE)

# must have '"tooltip = "text"' to work properly.
plotly::ggplotly(my_plot1, tooltip = "text")

```

The resulting plot is an interactive plot that as you mouse over the data will display the latitude,  longitude and value at the point.  As the plot takes a long time to create and is quote large,  both of which cause problems with CRAN, it is not displayed here.


## Cacheing, "last", "now", idempotency, and a gotcha

`rerddap` by default caches the requests you make, so that if you happen to make the same request again, the data is restored from the cache, rather than having to go out and retrieve it remotely.  For most applications, this a boon (such as when "knitting" and "reknitting" this document), as it can speed things up when doing a lot of request in a script, and works because in most cases an <span style="color:red">ERDDAP™</span> request is "idempotent".  This means that the the request will always return the same thing no matter what requests came before - it doesn't depend on state. However this is not true if the script uses either "last" in `griddap()` or "now" in `tabledap()` as these will return different values as time elapses and data are added to the datasets.  While it is desirable to have <span style="color:red">ERDDAP™</span> purely idempotent,  the "last" and "now" constructs are very helpful for people using <span style="color:red">ERDDAP™</span> in dashboards, webpages, regular input to models and the like, and the benefits far outweigh the problems.  However, if you are using either "last" or "now" in an `rerddap` based script, you want to be very careful to clear the `rerddap` cache, otherwise the request will be viewed as the same,  and the data from the last request, rather than the latest data, will be returned.  Note that several examples in this vignette use "last", and therefore the graphics may look different depending on when you "knitted" the vignette.

For help in dealing with the cache, see:

```{r,  eval = FALSE, echo = TRUE}
?cache_delete
?cache_delete_all
?cache_details
?cache_list
```



## Reading data from a netCDF file. {#ncdf4}

Here we give a brief summary of how to read in part of the data from a netCDF file.  The basic steps are:

* Open the netCDF file
* Map coordinate values to array indices
* Extract the data

A sample netCDF file, "MWsstd1day.nc" is included.  This is a small file and is a toy example,  but the basic principles remain the same for a larger file.

Open the netCDF file:

```{r}
require("ncdf4")
exampleFile <- system.file("extdata", "MWsstd1day.nc", package = "rerddap")
sstFile <- nc_open(exampleFile)
```

While it is possible to obtain the coordinate values by extracting them from the file,  `ncdf4` by default does so automatically in `nc_open`. The names of the coordinate dimensions can be found from:

```{r}
names(sstFile$dim)
```

(note that the coordinate names here are given in 'C' order,  while for an extract the coordinates will be in the opposite, "Fortran" order) and the values of any coordinate, say longitude, can be found from:

```{r}
sstFile$dim$longitude$vals
```

The names of the variables in the netCDF file can be found from:

```{r}
names(sstFile$var)
```


An extract is done by giving the pointer to the netCDF file ("sstFile" in this instance), the name of the variable ot be extracted ("sst" in this instance), the  starting index value (in array coordinates) for each dimension, and the the count (the number of other index values) to include.  If all values of a particular dimension are wanted, then "-1" can be used for the count.  So for example to extract all of the values for the first day:

```{r}
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(1, 1, 1, 1), count = c(1, 1, -1, -1))
```

Suppose we only want the latitudes from (30.0, 30.5) and longitudes (210.0, 210.5) for the first day.  We need to find the array indices that match those coordinate values:

```{r}
latMin <- which(sstFile$dim$latitude$vals == 30.0)
latMax <- which(sstFile$dim$latitude$vals == 30.5)
lonMin <- which(sstFile$dim$longitude$vals == 210.0)
lonMax <- which(sstFile$dim$longitude$vals == 210.5)
```

and then extract the data:

```{r}
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMin, latMin, 1, 1), count = c( (lonMax - lonMin + 1), (latMax - latMin + 1), 1, 1 ))
```


If we wanted a time series at (30N, 210E) it would be:

```{r}
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMax, latMin, 1, 1), count = c(1, 1, 1, -1 ))
```

For this example, it is easy to visually peruse the dimension values but for a large extract this might not be the possible.  Suppose we wanted all values in a latitude range or (30.1, 30.3) and since the range of values we are interested in might not be on a grid boundary we want the smallest range that would include these values  (that is if not on the grid then the smallest value may be less than 30.1)  and similarly for longitude, say (210.1, 210.3):

```{r}
latMin <- max(which(sstFile$dim$latitude$vals <= 30.1))
latMax <- min(which(sstFile$dim$latitude$vals >= 30.3))
lonMin <- max(which(sstFile$dim$longitude$vals <= 210.1))
lonMax <- min(which(sstFile$dim$longitude$vals >= 210.3))

```

and then perform the extract as before:

```{r}
require("ncdf4")
day1SST <- ncvar_get(sstFile, "sst", start = c(lonMin, latMin, 1, 1), count = c( (lonMax - lonMin + 1), (latMax - latMin + 1), 1, 1 ))
```

---
title: rerddap introduction
author: Scott Chamberlain
date: "2023-06-29"
output: rmarkdown::html_vignette
vignette: >
    %\VignetteIndexEntry{rerddap introduction}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
---




`rerddap` is a general purpose R client for working with ERDDAP™ servers. ERDDAP™ is a server built on top of OPenDAP, which serves some NOAA data. You can get gridded data ([griddap](https://upwell.pfeg.noaa.gov/erddap/griddap/documentation.html)), which lets you query from gridded datasets, or table data ([tabledap](https://upwell.pfeg.noaa.gov/erddap/tabledap/documentation.html)) which lets you query from tabular datasets. In terms of how we interface with them, there are similarties, but some differences too. We try to make a similar interface to both data types in `rerddap`.

## NetCDF

`rerddap` supports NetCDF format, and is the default when using the `griddap()` function. NetCDF is a binary file format, and will have a much smaller footprint on your disk than csv. The binary file format means it's harder to inspect, but the `ncdf4` package makes it easy to pull data out and write data back into a NetCDF file. Note the the file extension for NetCDF files is `.nc`. Whether you choose NetCDF or csv for small files won't make much of a difference, but will with large files.

## Caching

Data files downloaded are cached in a single hidden directory `~/.rerddap` on your machine. It's hidden so that you don't accidentally delete the data, but you can still easily delete the data if you like.

When you use `griddap()` or `tabledap()` functions, we construct a MD5 hash from the base URL, and any query parameters - this way each query is separately cached. Once we have the hash, we look in `~/.rerddap` for a matching hash. If there's a match we use that file on disk - if no match, we make a http request for the data to the ERDDAP™ server you specify.

## ERDDAP™ servers

You can get a data.frame of ERDDAP™ servers using the function `servers()`. The list of ERDDAP™ servers is drawn from the *Awesome ERDDAP™* page maintained by the Irish Marine Institute .  If you know of more ERDDAP™ servers, follow the instructions on that page to add the server.

## Install

Stable version from CRAN


```r
install.packages("rerddap")
```

Or, the development version from GitHub


```r
remotes::install_github("ropensci/rerddap")
```


```r
library("rerddap")
```

## Search

First, you likely want to search for data, specify either `griddadp` or `tabledap`


```r
ed_search(query = 'size', which = "table")
#> # A tibble: 36 × 2
#>    title                                                              dataset_id
#>    <chr>                                                              <chr>     
#>  1 CCE Prey Size and Hard Part Size Regressions                       mmtdPreyS…
#>  2 CCE Teleost Prey Size and Hard Part Size Measurements              mmtdTeleo…
#>  3 CalCOFI Larvae Sizes                                               erdCalCOF…
#>  4 CCE Non-Teleost Prey Size and Hard Part Size Measurements          mmtdNonTe…
#>  5 Channel Islands, Kelp Forest Monitoring, Size and Frequency, Natu… erdCinpKf…
#>  6 File Names from the AWS S3 noaa-goes16 Bucket                      awsS3Noaa…
#>  7 File Names from the AWS S3 noaa-goes17 Bucket                      awsS3Noaa…
#>  8 PacIOOS Beach Camera 001: Waikiki, Oahu, Hawaii                    BEACHCAM-…
#>  9 PacIOOS Beach Camera 003: Waimea Bay, Oahu, Hawaii                 BEACHCAM-…
#> 10 PacIOOS Beach Camera 004: Waimea Bay (Offshore), Oahu, Hawaii      BEACHCAM-…
#> # ℹ 26 more rows
```


```r
ed_search(query = 'size', which = "grid")
#> # A tibble: 103 × 2
#>    title                                                              dataset_id
#>    <chr>                                                              <chr>     
#>  1 Audio data from a local source.                                    testGridW…
#>  2 Main Hawaiian Islands Multibeam Bathymetry Synthesis: 50-m Bathym… hmrg_bath…
#>  3 Main Hawaiian Islands Multibeam Bathymetry Synthesis: 50-m Bathym… hmrg_bath…
#>  4 Coastal Upwelling Transport Index (CUTI), Daily                    erdCUTIda…
#>  5 SST smoothed frontal gradients                                     FRD_SSTgr…
#>  6 Coastal Upwelling Transport Index (CUTI), Monthly                  erdCUTImo…
#>  7 SST smoothed frontal gradients, Lon0360                            FRD_SSTgr…
#>  8 Biologically Effective Upwelling Transport Index (BEUTI), Daily    erdBEUTId…
#>  9 Biologically Effective Upwelling Transport Index (BEUTI), Monthly  erdBEUTIm…
#> 10 Daily averaged and put on grid 4x daily NCEP reanalysis (psi.2012) noaa_psl_…
#> # ℹ 93 more rows
```

There is now a convenience function to search over a list of ERDDAP™ servers,  designed to work with the function `servers()`


```r
server_list <- c(
  emodnet_physics = 'https://erddap.emodnet-physics.eu/erddap/',
  irish_marine_institute = 'https://erddap.marine.ie/erddap/'
)
global_search(query = 'size', server_list, 'griddap')
#>                                                                                                                                                              title
#> 1                                          EMODnet Physics - Total Suspended Matter - GridSeriesObservation - Concentration of total suspended matter - BALTIC SEA
#> 2                                   EMODnet Physics - Total Suspended Matter - GridSeriesObservation - Concentration of total suspended matter - MEDITERRANEAN SEA
#> 3                  EMODnet Physics - Total Suspended Matter - GridSeriesObservation - Concentration of total suspended matter - MEDITERRANEAN SEA - LOW RESOLUTION
#> 4                                                                                                           EMODnet Physics - TEMPERATURE YEARLY RECORDING DENSITY
#> 5                                                                  EMODPACE - Monthly sea level derived from CMEMS-DUACS (DT-2018) satellite altimetry (1993-2019)
#> 6                                           EMODnet Physics - Total Suspended Matter - GridSeriesObservation - Concentration of total suspended matter - NORTH SEA
#> 7                                                       EMODPACE - Absolute sea level trend (1993 – 2019) - derived from CMEMS-DUACS (DT-2018) satellite altimetry
#> 8 EMODPACE - Sea Level monthly mean, EurAsia. This product is based, uses and reprocess the CMEMS product id. SEALEVEL_GLO_PHY_CLIMATE_L4_REP_OBSERVATIONS_008_057
#> 9                                                                                                                                 COMPASS-NEATL Hindcast 2016-2020
#>                                  dataset_id
#> 1                             TSM_BALTICSEA
#> 2                                 TSM_MBSEA
#> 3                   TSM_MBSEA_LOWRESOLUTION
#> 4                           ERD_EP_TEMP_DNS
#> 5 EMODPACE_SLEV_MONTHLY_MEAN_DESEASONALIZED
#> 6                              TSM_NORTHSEA
#> 7                       EMODPACE_SLEV_TREND
#> 8                EMODPACE_SLEV_MONTHLY_MEAN
#> 9               compass_neatl_hindcast_grid
#>                                         url
#> 1 https://erddap.emodnet-physics.eu/erddap/
#> 2 https://erddap.emodnet-physics.eu/erddap/
#> 3 https://erddap.emodnet-physics.eu/erddap/
#> 4 https://erddap.emodnet-physics.eu/erddap/
#> 5 https://erddap.emodnet-physics.eu/erddap/
#> 6 https://erddap.emodnet-physics.eu/erddap/
#> 7 https://erddap.emodnet-physics.eu/erddap/
#> 8 https://erddap.emodnet-physics.eu/erddap/
#> 9          https://erddap.marine.ie/erddap/
```

## Information

Then you can get information on a single dataset


```r
info('erdCalCOFIlrvsiz')
#> <ERDDAP™ info> erdCalCOFIlrvsiz 
#>  Base URL: https://upwell.pfeg.noaa.gov/erddap 
#>  Dataset Type: tabledap 
#>  Variables:  
#>      calcofi_species_code: 
#>          Range: 19, 946 
#>      common_name: 
#>      cruise: 
#>      itis_tsn: 
#>      larvae_10m2: 
...
```

## griddap (gridded) data

First, get information on a dataset to see time range, lat/long range, and variables.


```r
(out <- info('erdMBchla1day'))
#> <ERDDAP info> erdMBchla1day 
#>  Base URL: https://upwell.pfeg.noaa.gov/erddap 
#>  Dataset Type: griddap 
#>  Dimensions (range):  
#>      time: (2006-01-01T12:00:00Z, 2023-06-27T12:00:00Z) 
#>      altitude: (0.0, 0.0) 
#>      latitude: (-45.0, 65.0) 
#>      longitude: (120.0, 320.0) 
#>  Variables:  
#>      chlorophyll: 
#>          Units: mg m-3
```

Then query for gridded data using the `griddap()` function


```r
(res <- griddap(out,
  time = c('2015-01-01','2015-01-03'),
  latitude = c(14, 15),
  longitude = c(125, 126)
))
#> <ERDDAP™ griddap> erdMBchla1day
#>    Path: [/var/folders/xw/mcmsdzzx4mgbttplylgs7ysh0000gp/T//RtmpoME6FV/R/rerddap/4d844aa48552049c3717ac94ced5f9b8.nc]
#>    Last updated: [2023-06-29 13:18:53.945082]
#>    File size:    [0.03 mb]
#>    Dimensions (dims/vars):   [4 X 1]
#>    Dim names: time, altitude, latitude, longitude
#>    Variable names: Chlorophyll Concentration in Sea Water
#>    data.frame (rows/columns):   [5043 X 5]
#> # A tibble: 5,043 × 5
#>    longitude  latitude  altitude time                 chlorophyll
#>    <dbl[1d]> <dbl[1d]> <dbl[1d]> <chr>                      <dbl>
#>  1      125         14         0 2015-01-01T12:00:00Z          NA
#>  2      125.        14         0 2015-01-01T12:00:00Z          NA
#>  3      125.        14         0 2015-01-01T12:00:00Z          NA
#>  4      125.        14         0 2015-01-01T12:00:00Z          NA
#>  5      125.        14         0 2015-01-01T12:00:00Z          NA
#>  6      125.        14         0 2015-01-01T12:00:00Z          NA
#>  7      125.        14         0 2015-01-01T12:00:00Z          NA
#>  8      125.        14         0 2015-01-01T12:00:00Z          NA
#>  9      125.        14         0 2015-01-01T12:00:00Z          NA
#> 10      125.        14         0 2015-01-01T12:00:00Z          NA
#> # ℹ 5,033 more rows
```

The output of `griddap()` is a list that you can explore further. Get the summary


```r
res$summary
#> $filename
#> [1] "/var/folders/xw/mcmsdzzx4mgbttplylgs7ysh0000gp/T//RtmpoME6FV/R/rerddap/4d844aa48552049c3717ac94ced5f9b8.nc"
#> 
#> $writable
#> [1] FALSE
#> 
#> $id
#> [1] 65536
#> 
#> $error
#> [1] FALSE
#> 
#> $safemode
#> [1] FALSE
#> 
...
```

Get the dimension variables


```r
names(res$summary$dim)
#> [1] "time"      "altitude"  "latitude"  "longitude"
```

Get the data.frame (beware: you may want to just look at the `head` of the data.frame if large)


```r
head(res$data)
#>   longitude latitude altitude                 time chlorophyll
#> 1   125.000       14        0 2015-01-01T12:00:00Z          NA
#> 2   125.025       14        0 2015-01-01T12:00:00Z          NA
#> 3   125.050       14        0 2015-01-01T12:00:00Z          NA
#> 4   125.075       14        0 2015-01-01T12:00:00Z          NA
#> 5   125.100       14        0 2015-01-01T12:00:00Z          NA
#> 6   125.125       14        0 2015-01-01T12:00:00Z          NA
```

## tabledap (tabular) data


```r
(out <- info('erdCalCOFIlrvsiz'))
#> <ERDDAP™ info> erdCalCOFIlrvsiz 
#>  Base URL: https://upwell.pfeg.noaa.gov/erddap 
#>  Dataset Type: tabledap 
#>  Variables:  
#>      calcofi_species_code: 
#>          Range: 19, 946 
#>      common_name: 
#>      cruise: 
#>      itis_tsn: 
#>      larvae_10m2: 
...
```


```r
(dat <- tabledap('erdCalCOFIlrvsiz', fields=c('latitude','longitude','larvae_size',
  'scientific_name'), 'time>=2011-01-01', 'time<=2011-12-31'))
#> <ERDDAP™ tabledap> erdCalCOFIlrvsiz
#>    Path: [/var/folders/xw/mcmsdzzx4mgbttplylgs7ysh0000gp/T//RtmpoME6FV/R/rerddap/db7389db5b5b0ed9c426d5c13bc43d18.csv]
#>    Last updated: [2023-06-29 13:18:57.579066]
#>    File size:    [0.05 mb]
#> # A tibble: 1,304 × 4
#>    latitude  longitude  larvae_size scientific_name     
#>    <chr>     <chr>      <chr>       <chr>               
#>  1 32.956665 -117.305   4.5         Engraulis mordax    
#>  2 32.91     -117.4     5.0         Merluccius productus
#>  3 32.511665 -118.21167 2.0         Merluccius productus
#>  4 32.511665 -118.21167 3.0         Merluccius productus
#>  5 32.511665 -118.21167 5.5         Merluccius productus
#>  6 32.511665 -118.21167 6.0         Merluccius productus
#>  7 32.511665 -118.21167 2.8         Merluccius productus
#>  8 32.511665 -118.21167 3.0         Sardinops sagax     
#>  9 32.511665 -118.21167 5.0         Sardinops sagax     
#> 10 32.511665 -118.21167 2.5         Engraulis mordax    
#> # ℹ 1,294 more rows
```

Since both `griddap()` and `tabledap()` give back data.frame's, it's easy to do downstream manipulation. For example, we can use `dplyr` to filter, summarize, group, and sort:


```r
library("dplyr")
dat$larvae_size <- as.numeric(dat$larvae_size)
dat %>%
  group_by(scientific_name) %>%
  summarise(mean_size = mean(larvae_size)) %>%
  arrange(desc(mean_size))
#> # A tibble: 7 × 2
#>   scientific_name       mean_size
#>   <chr>                     <dbl>
#> 1 Anoplopoma fimbria        23.3 
#> 2 Engraulis mordax           9.26
#> 3 Sardinops sagax            7.28
#> 4 Merluccius productus       5.48
#> 5 Tactostoma macropus        5   
#> 6 Scomber japonicus          3.4 
#> 7 Trachurus symmetricus      3.29
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache_list.R
\name{cache_list}
\alias{cache_list}
\title{List cached files}
\usage{
cache_list()
}
\description{
List cached files
}
\examples{
\dontrun{
# list files in cache
cache_list()

# List info for files
## download some data first
tabledap('erdCinpKfmBT')
griddap('erdVHNchlamday',
 time = c('2015-04-01','2015-04-10'),
 latitude = c(18, 21),
 longitude = c(-120, -119)
)

(x <- cache_list())
cache_details(x$nc[1])
cache_details(x$csv[1])
cache_details()

# delete files by name in cache
# cache_delete(x$nc[1])
# cache_delete(x$nc[2:3])
}
}
\seealso{
Other cache:
\code{\link[=cache_delete]{cache_delete()}},
\code{\link[=cache_details]{cache_details()}},
\code{\link[=cache_setup]{cache_setup()}}
}
\concept{cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache_delete.R
\name{cache_delete}
\alias{cache_delete}
\alias{cache_delete_all}
\title{Delete cached files}
\usage{
cache_delete(x, force = FALSE)

cache_delete_all(force = FALSE)
}
\arguments{
\item{x}{File names}

\item{force}{(logical) Should files be force deleted? Default: \code{FALSE}}
}
\description{
Delete cached files
}
\examples{
\dontrun{
# delete files by name in cache
# cache_delete('9911750294a039b8b517c8bf288978ea.csv')
# cache_delete(c('9911750294a039b8b517c8bf288978ea.csv',
#                'b26825b6737da13d6a52c28c8dfe690f.csv'))

# You can delete from the output of griddap or tabledap fxns
## tabledap
(table_res <- tabledap('erdCinpKfmBT'))
cache_delete(table_res)

## griddap
(out <- info('erdQMekm14day'))
(grid_res <- griddap(out,
 time = c('2015-12-28','2016-01-01'),
 latitude = c(24, 23),
 longitude = c(88, 90)
))
cache_delete(grid_res)
}
}
\seealso{
Other cache:
\code{\link[=cache_details]{cache_details()}},
\code{\link[=cache_list]{cache_list()}},
\code{\link[=cache_setup]{cache_setup()}}
}
\concept{cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table.R
\name{tabledap}
\alias{tabledap}
\title{Get ERDDAP™ tabledap data.}
\usage{
tabledap(
  x,
  ...,
  fields = NULL,
  distinct = FALSE,
  orderby = NULL,
  orderbymax = NULL,
  orderbymin = NULL,
  orderbyminmax = NULL,
  units = NULL,
  fmt = "csv",
  url = eurl(),
  store = disk(),
  callopts = list()
)
}
\arguments{
\item{x}{Anything coercable to an object of class info. So the output of
a call to \code{\link[=info]{info()}}, or a datasetid, which will internally be passed
through \code{\link[=info]{info()}}}

\item{...}{Any number of key-value pairs in quotes as query constraints.
See Details & examples}

\item{fields}{Columns to return, as a character vector}

\item{distinct}{If \code{TRUE} ERDDAP™ will sort all of the rows in the results
table (starting with the first requested variable, then using the second
requested variable if the first variable has a tie, ...), then remove all
non-unique rows of data. In many situations, ERDDAP™ can return distinct
values quickly and efficiently. But in some cases, ERDDAP™ must look through
all rows of the source dataset.}

\item{orderby}{If used, ERDDAP™ will sort all of the rows in the results
table (starting with the first variable, then using the second variable
if the first variable has a tie, ...). Normally, the rows of data in the
response table are in the order they arrived from the data source. orderBy
allows you to request that the results table be sorted in a specific way.
For example, use \code{orderby=c("stationID,time")} to get the results
sorted by stationID, then time. The orderby variables MUST be included in
the list of requested variables in the fields parameter.}

\item{orderbymax}{Give a vector of one or more fields, that must be included
in the fields parameter as well. Gives back data given constraints. ERDDAP™
will sort all of the rows in the results table (starting with the first
variable, then using the second variable if the first variable has a
tie, ...) and then just keeps the rows where the value of the last sort
variable is highest (for each combination of other values).}

\item{orderbymin}{Same as \code{orderbymax} parameter, except returns
minimum value.}

\item{orderbyminmax}{Same as \code{orderbymax} parameter, except returns
two rows for every combination of the n-1 variables: one row with the
minimum value, and one row with the maximum value.}

\item{units}{One of 'udunits' (units will be described via the UDUNITS
standard (e.g.,degrees_C)) or 'ucum' (units will be described via the
UCUM standard (e.g., Cel)).}

\item{fmt}{whether download should be as '.csv' (default) or '.parquet'}

\item{url}{A URL for an ERDDAP™ server.
Default: https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{store}{One of \code{disk} (default) or \code{memory}. You can pass
options to \code{disk}}

\item{callopts}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET} (must be
named parameters)}
}
\value{
An object of class \code{tabledap}. This class is a thin wrapper
around a data.frame, so the data you get back is a data.frame with metadata
attached as attributes (datasetid, path (path where the csv is stored on
your machine), url (url for the request))
}
\description{
Get ERDDAP™ tabledap data.
}
\details{
For key-value pair query constraints, the valid operators are =,
!= (not equals), =~ (a regular expression test), <, <=, >, and >= . For
regular expressions you need to add a regular expression. For others, nothing
more is needed. Construct the entry like \code{'time>=2001-07-07'} with the
parameter on the left, value on the right, and the operator in the middle,
all within a set of quotes. Since ERDDAP accepts values other than \code{=},
we can't simply do \code{time = '2001-07-07'} as we normally would.

Server-side functionality: Some tasks are done server side. You don't have
to worry about what that means. They are provided via parameters in this
function. See \code{distinct}, \code{orderby}, \code{orderbymax},
\code{orderbymin}, \code{orderbyminmax}, and \code{units}.

Data is cached based on all parameters you use to get a dataset, including
base url, query parameters. If you make the same exact call in the same or
a different R session, as long you don't clear the cache, the function only
reads data from disk, and does not have to request the data from the web
again.

If you run into an error like "HTTP Status 500 - There was a (temporary?)
problem. Wait a minute, then try again.". it's likely they are hitting
up against a size limit, and they should reduce the amount of data they
are requesting either via space, time, or variables. Pass in
\code{config = verbose()} to the request, and paste the URL into your
browser to see if the output is garbled to examine if there's a problem
with servers or this package
}
\examples{
\dontrun{
# Just passing the datasetid without fields gives all columns back
tabledap('erdCinpKfmBT')

# Pass time constraints
tabledap('erdCinpKfmBT', 'time>=2006-08-24')

# Pass in fields (i.e., columns to retrieve) & time constraints
tabledap('erdCinpKfmBT',
  fields = c('longitude', 'latitude', 'Aplysia_californica_Mean_Density'),
  'time>=2006-08-24'
)

# Get info on a datasetid, then get data given information learned
info('erdCalCOFIlrvsiz')$variables
tabledap('erdCalCOFIlrvsiz', fields=c('latitude','longitude','larvae_size',
   'itis_tsn'), 'time>=2011-10-25', 'time<=2011-10-31')

# An example workflow
## Search for data
(out <- ed_search(query='fish', which = 'table'))
## Using a datasetid, search for information on a datasetid
id <- out$alldata[[1]]$dataset_id
vars <- info(id)$variables
## Get data from the dataset
vars$variable_name[1:3]
tabledap(id, fields = vars$variable_name[1:3])

# Time constraint
## Limit by time with date only
(info <- info('erdCinpKfmBT'))
tabledap(info, fields = c(
  'latitude','longitude','Haliotis_fulgens_Mean_Density'),
  'time>=2001-07-14')

# Use distinct parameter - compare to distinct = FALSE
tabledap('sg114_3',
   fields=c('longitude','latitude','trajectory'),
   'time>=2008-12-05', distinct = TRUE)

# Use units parameter
## In this example, values are the same, but sometimes they can be different
## given the units value passed
tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
   'time>=2007-09-19', 'time<=2007-09-21', units='udunits')
tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
   'time>=2007-09-19', 'time<=2007-09-21', units='ucum')

# Use orderby parameter
tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
   'time>=2007-09-19', 'time<=2007-09-21', orderby='temperature')
# Use orderbymax parameter
tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
   'time>=2007-09-19', 'time<=2007-09-21', orderbymax='temperature')
# Use orderbymin parameter
tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
   'time>=2007-09-19', 'time<=2007-09-21', orderbymin='temperature')
# Use orderbyminmax parameter
tabledap('erdCinpKfmT', fields=c('longitude','latitude','time','temperature'),
   'time>=2007-09-19', 'time<=2007-09-21', orderbyminmax='temperature')
# Use orderbymin parameter with multiple values
tabledap('erdCinpKfmT',
   fields=c('longitude','latitude','time','depth','temperature'),
   'time>=2007-06-10', 'time<=2007-09-21',
   orderbymax=c('depth','temperature')
)

# Integrate with taxize
out <- tabledap('erdCalCOFIlrvcntHBtoHI',
   fields = c('latitude','longitude','scientific_name','itis_tsn'),
   'time>=2007-06-10', 'time<=2007-09-21'
)
tsns <- unique(out$itis_tsn[1:100])
library("taxize")
classif <- classification(tsns, db = "itis")
head(rbind(classif)); tail(rbind(classif))

# Write to memory (within R), or to disk
(out <- info('erdCinpKfmBT'))
## disk, by default (to prevent bogging down system w/ large datasets)
## the 2nd call is much faster as it's mostly just the time of reading
## in the table from disk
system.time( tabledap('erdCinpKfmBT', store = disk()) )
system.time( tabledap('erdCinpKfmBT', store = disk()) )
## memory
tabledap('erdCinpKfmBT', store = memory())

# use a different ERDDAP™ server
## NOAA IOOS NERACOOS
url <- "http://www.neracoos.org/erddap/"
tabledap("E01_optics_hist", url = url)
}
}
\references{
https://upwell.pfeg.noaa.gov/erddap/index.html
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/browse.R
\name{browse}
\alias{browse}
\title{Browse a dataset webpage.}
\usage{
browse(x, url = eurl(), ...)
}
\arguments{
\item{x}{datasetid or an object associated with a datasetid such
\code{\link[=info]{info()}}, \code{\link[=griddap]{griddap()}} or \code{\link[=tabledap]{tabledap()}}}

\item{url}{A URL for an ERDDAP™ server. Default:
https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Further args passed on to \code{utils::browseURL}
(must be a named parameter)}
}
\value{
if in interactive mode, opens a URL in your default browser;
if not, then prints the URL in the console
}
\description{
Browse a dataset webpage.
}
\examples{
\dontrun{
if (interactive()) {
# browse by dataset_id
browse('erdATastnhday')

# browse info class
my_info <- info('erdATastnhday')
browse(my_info)

# browse tabledap class
my_tabledap <- tabledap('erdCalCOFIlrvsiz', fields=c('latitude','longitude','larvae_size',
   'itis_tsn'), 'time>=2011-10-25', 'time<=2011-10-31')
browse(my_tabledap)
}}
}
\author{
Ben Tupper \email{btupper@bigelow.org}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colors.R
\docType{data}
\name{colors}
\alias{colors}
\title{cmocean colors
The cmocean color palette by Kristen Thyng as implemented in the R package "oce"}
\format{
An object of class \code{list} of length 13.
}
\usage{
colors
}
\description{
str(colors)
List of 13
$ viridis
$ cdom
$ chlorophyll
$ density
$ freesurface
$ oxygen
$ par
$ phase
$ salinity
$ temperature
$ turbidity
$ velocity
$ vorticity
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info.R
\name{info}
\alias{info}
\alias{as.info}
\title{Get information on an ERDDAP(TM) dataset.}
\usage{
info(datasetid, url = eurl(), ...)

as.info(x, url)
}
\arguments{
\item{datasetid}{Dataset id}

\item{url}{A URL for an ERDDAP(TM) server. Default:
https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Further args passed on to \link[crul:verb-GET]{crul::verb-GET} (must be a
named parameter)}

\item{x}{A datasetid or the output of \code{info}}
}
\value{
Prints a summary of the data on return, but you can index to
various information.

The data is a list of length two with:
\itemize{
\item variables - Data.frame of variables and their types
\item alldata - List of data variables and their full attributes
}

Where \code{alldata} element has many data.frame's, one for each variable,
with metadata for that variable. E.g., for griddap dataset
\code{noaa_pfeg_696e_ec99_6fa6}, \code{alldata} has:
\itemize{
\item NC_GLOBAL
\item time
\item latitude
\item longitude
\item sss
}
}
\description{
Get information on an ERDDAP(TM) dataset.
}
\examples{
\dontrun{
# grid dap datasets
info('erdATastnhday')

(out <- ed_search(query='temperature'))
info(out$info$dataset_id[5])
info(out$info$dataset_id[15])
info(out$info$dataset_id[25])
info(out$info$dataset_id[150])
info(out$info$dataset_id[400])
info(out$info$dataset_id[678])

out <- info(datasetid='erdMBchla1day')
## See brief overview of the variables and range of possible values, if given
out$variables
## all information on longitude
out$alldata$longitude
## all information on chlorophyll
out$alldata$chlorophyll

# table dap datasets
(out <- ed_search(query='temperature', which = "table"))
info(out$info$dataset_id[1])
info(out$info$dataset_id[2])
info(out$info$dataset_id[3])
info(out$info$dataset_id[4])

info('erdCinpKfmBT')
out <- info('erdCinpKfmBT')
## See brief overview of the variables and range of possible values, if given
out$variables
## all information on longitude
out$alldata$longitude
## all information on Haliotis_corrugata_Mean_Density
out$alldata$Haliotis_corrugata_Mean_Density

# use a different ERDDAP(TM) server
## Marine Institute (Ireland)
info("IMI_CONN_2D", url = "http://erddap.marine.ie/erddap/")
}
}
\references{
https://upwell.pfeg.noaa.gov/erddap/index.html
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version.R
\name{version}
\alias{version}
\title{Get ERDDAP™ version}
\usage{
version(url = eurl(), ...)
}
\arguments{
\item{url}{A URL for an ERDDAP™ server. Default:
https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
Get ERDDAP™ version
}
\examples{
\dontrun{
version()
ss <- servers()
version(ss$url[2])
version(ss$url[3])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerddap-package.r
\docType{data}
\name{standardnames}
\alias{standardnames}
\title{standardnames}
\format{
A character vector
}
\description{
standardnames
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/servers.R
\name{servers}
\alias{servers}
\title{ERDDAP™ server URLS and other info}
\usage{
servers(...)
}
\arguments{
\item{...}{curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\value{
data.frame with 3 columns:
\itemize{
\item name (character): ERDDAP™ name
\item url (character): ERDDAP™ url
\item public (logical): whether it's public or not
}
}
\description{
ERDDAP™ server URLS and other info
}
\examples{
\dontrun{
servers()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_time.R
\name{convert_time}
\alias{convert_time}
\title{Convert a UDUNITS compatible time to ISO time}
\usage{
convert_time(
  n = NULL,
  isoTime = NULL,
  units = "seconds since 1970-01-01T00:00:00Z",
  url = eurl(),
  method = "local",
  ...
)
}
\arguments{
\item{n}{numeric; A unix time number.}

\item{isoTime}{character; A string time representation.}

\item{units}{character; Units to return. Default:
"seconds since 1970-01-01T00:00:00Z"}

\item{url}{Base URL of the ERDDAP™ server. See \code{\link[=eurl]{eurl()}} for
more information}

\item{method}{(character) One of local or web. Local simply uses
\code{\link[=as.POSIXct]{as.POSIXct()}}, while web method uses the ERDDAP™ time conversion service
\verb{/erddap/convert/time.txt}}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
Convert a UDUNITS compatible time to ISO time
}
\details{
When \code{method = "web"} time zone is GMT/UTC
}
\examples{
 \dontrun{
# local conversions
convert_time(n = 473472000)
convert_time(isoTime = "1985-01-02T00:00:00Z")

# using an ERDDAP™ web service
convert_time(n = 473472000, method = "web")
convert_time(isoTime = "1985-01-02T00:00:00Z", method = "web")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerddap-package.r
\docType{data}
\name{ioos_categories}
\alias{ioos_categories}
\title{ioos_categories}
\format{
A character vector
}
\description{
ioos_categories
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_adv.R
\name{ed_search_adv}
\alias{ed_search_adv}
\title{Advanced search for ERDDAP™ tabledep or griddap datasets}
\usage{
ed_search_adv(
  query = NULL,
  page = 1,
  page_size = 1000,
  protocol = NULL,
  cdm_data_type = NULL,
  institution = NULL,
  ioos_category = NULL,
  keywords = NULL,
  long_name = NULL,
  standard_name = NULL,
  variableName = NULL,
  maxLat = NULL,
  minLon = NULL,
  maxLon = NULL,
  minLat = NULL,
  minTime = NULL,
  maxTime = NULL,
  url = eurl(),
  ...
)
}
\arguments{
\item{query}{(character) Search terms}

\item{page}{(integer) Page number. Default: 1}

\item{page_size}{(integer) Results per page: Default: 1000}

\item{protocol}{(character) One of any (default), tabledep or griddap}

\item{cdm_data_type}{(character) One of grid, other, point, profile,
timeseries, timeseriesprofile, trajectory, trajectoryprofile}

\item{institution}{(character) An institution. See the dataset
\code{institutions}}

\item{ioos_category}{(character) An ioos category See the dataset
\code{ioos_categories}}

\item{keywords}{(character) A keywords. See the dataset \code{keywords}}

\item{long_name}{(character) A long name. See the dataset \code{longnames}}

\item{standard_name}{(character) A standar dname. See the dataset
\code{standardnames}}

\item{variableName}{(character) A variable name. See the dataset
\code{variablenames}}

\item{minLon, maxLon}{(numeric) Minimum and maximum longitude. Some datasets
have longitude values within -180 to 180, others use 0 to 360. If you
specify min and max Longitude within -180 to 180 (or 0 to 360), ERDDAP™ will
only find datasets that match the values you specify. Consider doing one
search: longitude -180 to 360, or two searches: longitude -180 to 180,
and 0 to 360.}

\item{minLat, maxLat}{(numeric) Minimum and maximum latitude, between -90
and 90}

\item{minTime, maxTime}{(numeric/character) Minimum and maximum time. Time
string with the format "yyyy-MM-ddTHH:mm:ssZ, (e.g., 2009-01-21T23:00:00Z).
If you specify something, you must include at least yyyy-MM-dd; you can
omit Z, :ss, :mm, :HH, and T. Always use UTC (GMT/Zulu) time. Or specify
the number of seconds since 1970-01-01T00:00:00Z.}

\item{url}{A URL for an ERDDAP™ server. Default:
https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET} (must be
named parameters)}
}
\description{
Advanced search for ERDDAP™ tabledep or griddap datasets
}
\examples{
\dontrun{
ed_search_adv(query = 'temperature')
ed_search_adv(query = 'temperature', protocol = "griddap")
ed_search_adv(query = 'temperature', protocol = "tabledap")
ed_search_adv(maxLat = 63, minLon = -107, maxLon = -87, minLat = 50,
  protocol = "griddap")
ed_search_adv(maxLat = 63, minLon = -107, maxLon = -87, minLat = 50,
  protocol = "tabledap")
ed_search_adv(minTime = "2010-01-01T00:00:00Z",
  maxTime="2010-02-01T00:00:00Z")
(out <- ed_search_adv(maxLat = 63, minLon = -107, maxLon = -87, minLat = 50,
             minTime = "2010-01-01T00:00:00Z",
             maxTime="2010-02-01T00:00:00Z"))
out$alldata[[1]]
ed_search_adv(variableName = 'upwelling')
ed_search_adv(query = 'upwelling', protocol = "tabledap")

# use a different URL
ed_search_adv(query = 'temperature', url = servers()$url[6])
}
}
\references{
https://upwell.pfeg.noaa.gov/erddap/index.html
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fipscounty.R
\name{fipscounty}
\alias{fipscounty}
\title{Convert a FIPS County Code to/from a County Name}
\usage{
fipscounty(county = NULL, code = NULL, url = eurl(), ...)
}
\arguments{
\item{county}{character; A county name.}

\item{code}{numeric; A FIPS code.}

\item{url}{A URL for an ERDDAP™ server. Default:
https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
Convert a FIPS County Code to/from a County Name
}
\examples{
 \dontrun{
fipscounty(code = "06053")
fipscounty(county = "CA, Monterey")
fipscounty(county = "OR, Multnomah")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{ed_search}
\alias{ed_search}
\alias{ed_datasets}
\title{Search for ERDDAP™ tabledep or griddap datasets}
\usage{
ed_search(
  query,
  page = NULL,
  page_size = NULL,
  which = "griddap",
  url = eurl(),
  ...
)

ed_datasets(which = "tabledap", url = eurl())
}
\arguments{
\item{query}{(character) Search terms}

\item{page}{(integer) Page number}

\item{page_size}{(integer) Results per page}

\item{which}{(character) One of tabledep or griddap.}

\item{url}{A URL for an ERDDAP™ server. Default:
https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET} (must be
named parameters)}
}
\description{
Search for ERDDAP™ tabledep or griddap datasets
}
\examples{
\dontrun{
(out <- ed_search(query='temperature'))
out$alldata[[1]]
(out <- ed_search(query='size'))
out$info

# List datasets
ed_datasets('table')
ed_datasets('grid')

# use a different ERDDAP™ server
## Marine Institute (Ireland)
ed_search("temperature", url = "http://erddap.marine.ie/erddap/")
}
}
\references{
https://upwell.pfeg.noaa.gov/erddap/index.html
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{disk}
\alias{disk}
\alias{memory}
\title{Options for saving ERDDAP™ datasets.}
\usage{
disk(path = NULL, overwrite = TRUE)

memory()
}
\arguments{
\item{path}{Path to store files in. A directory, not a file.
Default: the root cache path, see \code{\link{cache_setup}}}

\item{overwrite}{(logical) Overwrite an existing file of the same name?
Default: \code{TRUE}}
}
\description{
Options for saving ERDDAP™ datasets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/default-url.R
\name{eurl}
\alias{eurl}
\title{Default ERDDAP™ server URL}
\usage{
eurl()
}
\description{
Default ERDDAP™ server URL
}
\details{
default url is https://upwell.pfeg.noaa.gov/erddap/

You can set a default using an environment variable so you
don't have to pass anything to the URL parameter in your
function calls.

In your .Renviron file or similar set a URL for the environment
variable \code{RERDDAP_DEFAULT_URL}, like
\verb{RERDDAP_DEFAULT_URL=https://upwell.pfeg.noaa.gov/erddap/}

It's important that you include a trailing slash in your URL
}
\examples{
eurl()
Sys.setenv(RERDDAP_DEFAULT_URL = "https://google.com")
Sys.getenv("RERDDAP_DEFAULT_URL")
eurl()
Sys.unsetenv("RERDDAP_DEFAULT_URL")
eurl()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerddap-package.r
\docType{package}
\name{rerddap}
\alias{rerddap-package}
\alias{rerddap}
\title{rerddap}
\description{
General purpose R client for ERDDAP™ servers
}
\section{ERDDAP™ info}{

NOAA's ERDDAP™ service holds many datasets of interest. It's built on top of
OPenDAP. You can search for datasets via
\code{\link[=ed_search]{ed_search()}}, list datasets via \code{\link[=ed_datasets]{ed_datasets()}},
get information on a single dataset via \code{\link[=info]{info()}}, then get
data you want for either tabledap type via \code{\link[=tabledap]{tabledap()}}, or
for griddap type via \code{\link[=griddap]{griddap()}}
}

\section{tabledap/griddap}{

tabledap and griddap have different interfaces to query for data, so
\code{\link[=tabledap]{tabledap()}} and \code{\link[=griddap]{griddap()}} are separated out as
separate functions even though some of the internals are the same. In
particular, with tabledap you can query on/subset all variables, whereas
with gridddap, you can only query on/subset the dimension varibles (e.g.,
latitude, longitude, altitude).
}

\section{Data size}{

With griddap data via \code{\link[=griddap]{griddap()}} you can get a lot of
data quickly. Try small searches of a dataset to start to get a sense for
the data, then you can increase the amount of data you get. See
\code{\link[=griddap]{griddap()}} for more details.
}

\section{Caching}{

\pkg{rerddap} by default caches the requests you make, so that if you happen to
make the same request again, the data is restored from the cache, rather than
having to go out and retrieve it remotely.  For most applications, this is good,
as it can speed things up when doing a lot of request in a script, and works
because in most cases an ERDDAP™ request is "idempotent".  This means that the
the request will always return the same thing no matter what requests came
before - it doesn't depend on state. However this is not true if the
script uses either "last" in \code{\link[=griddap]{griddap()}} or "now" in \code{\link[=tabledap]{tabledap()}} as these
will return different values as time elapses and data are added to the
datasets.  While it is desirable to have ERDDAP™ purely idempotent,  the
"last" and "now" constructs are very helpful for people using ERDDAP™
in dashboards, webpages, regular input to models and the like, and the
benefits far outweigh the problems.  However, if you are using either "last"
or "now" in an \pkg{rerddap} based script, you want to be very careful to
clear the \pkg{rerddap} cache, otherwise the request will be viewed as the
same,  and the data from the last request, rather than the latest data,
will be returned.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/rerddap/}
  \item \url{https://github.com/ropensci/rerddap}
  \item Report bugs at \url{https://github.com/ropensci/rerddap/issues}
}

}
\author{
\strong{Maintainer}: Roy Mendelssohn \email{roy.mendelssohn@noaa.gov} [contributor]

Authors:
\itemize{
  \item Scott Chamberlain
}

Other contributors:
\itemize{
  \item Ben Tupper [contributor]
  \item Salvador Jesús Fernández Bejarano [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_units.R
\name{convert_units}
\alias{convert_units}
\title{Convert a CF Standard Name to/from a GCMD Science Keyword}
\usage{
convert_units(udunits = NULL, ucum = NULL, url = eurl(), ...)
}
\arguments{
\item{udunits}{character; A UDUNITS character string
https://www.unidata.ucar.edu/software/udunits/}

\item{ucum}{character; A UCUM character string
https://ucum.org/ucum.html}

\item{url}{Base URL of the ERDDAP server. See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
Convert a CF Standard Name to/from a GCMD Science Keyword
}
\examples{
 \dontrun{
convert_units(udunits = "degree_C meter-1")
convert_units(ucum = "Cel.m-1")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global_search.R
\name{global_search}
\alias{global_search}
\title{global_search}
\usage{
global_search(query, server_list, which_service)
}
\arguments{
\item{query}{(character) Search terms}

\item{server_list}{(list of character) List of ERDDAP™ servers to search}

\item{which_service}{(character) One of tabledep or griddap.}
}
\value{
If successful a dataframe wih columns:
\itemize{
\item title - the dataset title
\item dataset_id - the datasetid on that ERDDAP™ server
\item url - base url of dataset ERDDAP™ server
}
if urls are valid,  no match is found,  will return no match found
else returns error message
}
\description{
Search for ERDDAP™ tabledap or griddap datasets from a list
of ERDDAP™ servers based on search terms.
}
\details{
Uses the 'reddap' function ed_search() to search over
the list of servers
}
\examples{
# get list of servers know by
# https://irishmarineinstitute.github.io/awesome-erddap
# e_servers <- servers()$url
# select a couple to search
# e_servers <- e_servers[c(1, 40)]
# to meet CRAN time limits will only search 1 place
e_servers <- "https://coastwatch.pfeg.noaa.gov/erddap/"
test_query <- 'C-HARM v1 2-Day Forecast'
query_results <- global_search(test_query, e_servers, "griddap")
}
\seealso{
\code{\link[crul]{HttpClient}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerddap-package.r
\docType{data}
\name{longnames}
\alias{longnames}
\title{longnames}
\format{
A character vector
}
\description{
longnames
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache_setup.R
\name{cache_setup}
\alias{cache_setup}
\alias{cache_info}
\title{Setup cache path}
\usage{
cache_setup(full_path = NULL, temp_dir = FALSE)

cache_info()
}
\arguments{
\item{full_path}{(character) the full path to use for storing cached
files.}

\item{temp_dir}{(logical) if \code{TRUE} use a randomly assigned
\code{tempdir} (and \code{full_path} is ignored), if \code{FALSE}, you
can use \code{full_path}.}
}
\value{
the full cache path, a directory (character)
}
\description{
Setup cache path
}
\details{
On opening,  by default a temporary directory is created for caching
files.  To have files cached elsewhere,  give the full path of where to
cache files.  Adding \code{temp_dir = TRUE} will again use a temporary
dirctory for cacheing.
}
\examples{
\dontrun{
# default path
cache_setup()

# you can define your own path
cache_setup(path = "foobar")

# set a tempdir - better for programming with to avoid prompt
cache_setup(temp_dir = TRUE)

# cache info
cache_info()
}
}
\seealso{
Other cache:
\code{\link[=cache_delete]{cache_delete()}},
\code{\link[=cache_details]{cache_details()}},
\code{\link[=cache_list]{cache_list()}}
}
\concept{cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keywords.R
\name{key_words}
\alias{key_words}
\title{Convert a CF Standard Name to/from a GCMD Science Keyword}
\usage{
key_words(cf = NULL, gcmd = NULL, url = eurl(), ...)
}
\arguments{
\item{cf}{character; A cf standard name
http://cfconventions.org/Data/cf-standard-names/27/build/cf-standard-name-table.html}

\item{gcmd}{character; A GCMD science keyword
http://gcmd.gsfc.nasa.gov/learn/keyword_list.html}

\item{url}{A URL for an ERDDAP™ server. Default:
https://upwell.pfeg.noaa.gov/erddap/. See \code{\link[=eurl]{eurl()}} for
more information}

\item{...}{Curl options passed on to \link[crul:verb-GET]{crul::verb-GET}}
}
\description{
Convert a CF Standard Name to/from a GCMD Science Keyword
}
\examples{
 \dontrun{
key_words(cf = "air_pressure")
cat(key_words(cf = "air_pressure"))

# a different ERDDAP™ server
# key_words(cf = "air_pressure", url = servers()$url[6])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid.R
\name{griddap}
\alias{griddap}
\title{Get ERDDAP(TM) gridded data}
\usage{
griddap(
  datasetx,
  ...,
  fields = "all",
  stride = 1,
  fmt = "nc",
  url = eurl(),
  store = disk(),
  read = TRUE,
  callopts = list()
)
}
\arguments{
\item{datasetx}{Anything coercable to an object of class info. So the output of a
call to \code{\link{info}}, or a datasetid, which will internally be passed
through \code{\link{info}}}

\item{...}{Dimension arguments. See examples. Can be any 1 or more of the
dimensions for the particular dataset - and the dimensions vary by dataset.
For each dimension, pass in a vector of length two, with min and max value
desired. at least 1 required.}

\item{fields}{(character) Fields to return, in a character vector.}

\item{stride}{(integer) How many values to get. 1 = get every value, 2 = get
every other value, etc. Default: 1 (i.e., get every value)}

\item{fmt}{(character) One of nc (for netcdf), csv, or parquet. Default: nc}

\item{url}{A URL for an ERDDAP™ server. Default:
https://upwell.pfeg.noaa.gov/erddap/ - See \code{\link[=eurl]{eurl()}} for
more information}

\item{store}{One of \code{\link{disk}} (default) or \code{\link{memory}}. You
can pass options to \code{\link{disk}}. Beware: if you choose \code{fmt="nc"},
we force \code{store=disk()} because nc files have to be written to disk.}

\item{read}{(logical) Read data into memory or not. Does not apply when
\code{store} parameter is set to memory (which reads data into memory).
For large csv, or especially netcdf files, you may want to set this to
\code{FALSE}, which simply returns a summary of the dataset - and you can
read in data piecemeal later. Default: \code{TRUE}}

\item{callopts}{Curl options passed on to \code{\link[crul]{verb-GET}}}
}
\value{
An object of class \code{griddap_csv} if csv chosen or
\code{griddap_nc} if nc file format chosen.

\itemize{
\item \code{griddap_csv}: a data.frame created from the downloaded csv
data
\item \code{griddap_nc}: a list, with slots "summary" and "data". "summary"
is the unclassed output from \code{ncdf4::nc_open}, from which you can
do any netcdf operations you like. "data" is a data.frame created
from the netcdf data. the data.frame may be empty if there were problems
parsing the netcdf data
}

Both have the attributes: datasetid (the dataset id), path (the path on file
for the csv or nc file), url (the url requested to the ERDDAP server)

If \code{read=FALSE}, the data.frame for \code{griddap_csv}
and the data.frame in the "data" slot is empty for \code{griddap_nc}
}
\description{
Get ERDDAP(TM) gridded data
}
\details{
Details:

If you run into an error like "HTTP Status 500 - There was a (temporary?)
problem. Wait a minute, then try again.". it's likely they are hitting
up against a size limit, and they should reduce the amount of data they
are requesting either via space, time, or variables. Pass in
\code{config = verbose()} to the request, and paste the URL into your
browser to see if the output is garbled to examine if there's a problem
with servers or this package
}
\section{Dimensions and Variables}{

ERDDAP™ grid dap data has this concept of dimenions vs. variables. Dimensions
are things like time, latitude, longitude, altitude, and depth. Whereas
variables are the measured variables, e.g., temperature, salinity, air.

You can't separately adjust values for dimensions for different variables.
So, here's how it's gonna work:

Pass in lower and upper limits you want for each dimension as a vector
(e.g., \code{c(1,2)}), or leave to defaults (i.e., don't pass anything to
a dimension). Then pick which variables you want returned via the
\code{fields} parameter. If you don't pass in options to the \code{fields}
parameter, you get all variables back.

To get the dimensions and variables, along with other metadata for a
dataset, run \code{\link{info}}, and each will be shown, with their min
and max values, and some other metadata.
}

\section{Where does the data go?}{

You can choose where data is stored. Be careful though. You can easily get a
single file of hundreds of MB's (upper limit: 2 GB) in size with a single
request. To the \code{store} parameter, pass \code{\link{memory}} if you
want to store the data in memory (saved as a data.frame), or pass
\code{\link{disk}} if you want to store on disk in a file. Note that
\code{\link{memory}} and \code{\link{disk}} are not character strings, but
function calls. \code{\link{memory}} does not accept any inputs, while
\code{\link{disk}} does. Possibly will add other options, like
\dQuote{sql} for storing in a SQL database.
}

\section{Non-lat/lon grid data}{

Some gridded datasets have latitude/longitude components, but some do not.
When nc format gridded datasets have latitude and longitude we "melt" them into a
data.frame for easy downstream consumption. When nc format gridded datasets do
not have latitude and longitude components, we do not read in the data, throw
a warning saying so. You can readin the nc file yourself with the file path.
CSV format is not affected by this issue as CSV data is easily turned into
a data.frame regardless of whether latitude/longitude data are present.
}

\examples{
\dontrun{
# single variable dataset
## You can pass in the outpu of a call to info
(out <- info('erdVHNchlamday'))
## Or, pass in a dataset id
(res <- griddap('erdVHNchlamday',
 time = c('2015-04-01','2015-04-10'),
 latitude = c(18, 21),
 longitude = c(-120, -119)
))

# multi-variable dataset
(out <- info('erdQMekm14day'))
(res <- griddap(out,
 time = c('2015-12-28','2016-01-01'),
 latitude = c(24, 23),
 longitude = c(88, 90)
))
(res <- griddap(out, time = c('2015-12-28','2016-01-01'),
   latitude = c(24, 23), longitude = c(88, 90), fields = 'mod_current'))
(res <- griddap(out, time = c('2015-12-28','2016-01-01'),
   latitude = c(24, 23), longitude = c(88, 90), fields = 'mod_current',
   stride = c(1,2,1,2)))
(res <- griddap(out, time = c('2015-12-28','2016-01-01'),
   latitude = c(24, 23), longitude = c(88, 90),
   fields = c('mod_current','u_current')))


# Write to memory (within R), or to disk
(out <- info('erdQSwindmday'))
## disk, by default (to prevent bogging down system w/ large datasets)
## you can also pass in path and overwrite options to disk()
(res <- griddap(out,
 time = c('2006-07-11','2006-07-20'),
 longitude = c(166, 170),
 store = disk()
))
## the 2nd call is much faster as it's mostly just the time of reading in
## the table from disk
system.time( griddap(out,
 time = c('2006-07-11','2006-07-15'),
 longitude = c(10, 15),
 store = disk()
) )
system.time( griddap(out,
 time = c('2006-07-11','2006-07-15'),
 longitude = c(10, 15),
 store = disk()
) )

## memory - you have to choose fmt="csv" if you use memory
(res <- griddap("erdMBchla1day",
 time = c('2015-01-01','2015-01-03'),
 latitude = c(14, 15),
 longitude = c(125, 126),
 fmt = "csv", store = memory()
))

## Use ncdf4 package to parse data
info("erdMBchla1day")
(res <- griddap("erdMBchla1day",
 time = c('2015-01-01','2015-01-03'),
 latitude = c(14, 15),
 longitude = c(125, 126)
))

# Get data in csv format
## by default, we get netcdf format data
(res <- griddap('erdMBchla1day',
 time = c('2015-01-01','2015-01-03'),
 latitude = c(14, 15),
 longitude = c(125, 126),
 fmt = "csv"
))

# Use a different ERDDAP™ server url
## NOAA IOOS PacIOOS
url = "https://cwcgom.aoml.noaa.gov/erddap/"
out <- info("miamiacidification", url = url)
(res <- griddap(out,
 time = c('2019-11-01','2019-11-03'),
 latitude = c(15, 16),
 longitude = c(-90, -88)
))
## pass directly into griddap() - if you pass a datasetid string directly
## you must pass in the url or you'll be querying the default ERDDAP™ url,
## which isn't the one you want if you're not using the default ERDDAP™ url
griddap("miamiacidification", url = url,
 time = c('2019-11-01','2019-11-03'),
 latitude = c(15, 16),
 longitude = c(-90, -88)
)

# Using 'last'
## with time
griddap('erdVHNchlamday',
 time = c('last-5','last'),
 latitude = c(18, 21),
 longitude = c(-120, -119)
)
## with latitude
griddap('erdVHNchlamday',
  time = c('2015-04-01','2015-04-10'),
  latitude = c('last', 'last'),
  longitude = c(-120, -119)
)
## with longitude
griddap('erdVHNchlamday',
  time = c('2015-04-01','2015-04-10'),
  latitude = c(18, 21),
  longitude = c('last', 'last')
)

# datasets without lat/lon grid and with fmt=nc
# FIXME: this dataset is gone
# (x <- info('glos_tds_5912_ca66_3f41'))
# res <- griddap(x,
#   time = c('2018-04-01','2018-04-10'),
#   ny = c(1, 2),
#   nx = c(3, 5)
# )
## data.frame is empty
# res$data
## read in from the nc file path
# ncdf4::nc_open(res$summary$filename)
}
}
\references{
https://upwell.pfeg.noaa.gov/erddap/rest.html
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache_details.R
\name{cache_details}
\alias{cache_details}
\title{Get details of cached files}
\usage{
cache_details(x)
}
\arguments{
\item{x}{File names}
}
\description{
Get details of cached files
}
\details{
Can be used to list details for all files, both .nc and .csv
types, or details for just individual files of class \code{tabledap},
\code{griddap_nc}, and \code{griddap_csv}
}
\examples{
\dontrun{
# List details for all cached files
cache_details()
}
}
\seealso{
Other cache:
\code{\link[=cache_delete]{cache_delete()}},
\code{\link[=cache_list]{cache_list()}},
\code{\link[=cache_setup]{cache_setup()}}
}
\concept{cache}
